/*
 * PegasusN: Peta-Scale Graph Mining System (Pegasus v3.0)
 * Authors: Chiwan Park, Ha-Myung Park, U Kang
 *
 * This software is licensed under Apache License, Version 2.0 (the  "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 * -------------------------------------------------------------------------
 * File: PACC.java
 * - pacc. It finds connected components in a graph.
 * Version: 3.0
 */

package pegasus.mapreduce.cc;

import org.apache.hadoop.conf.Configuration;
import org.apache.hadoop.conf.Configured;
import org.apache.hadoop.fs.FileSystem;
import org.apache.hadoop.fs.Path;
import org.apache.hadoop.util.Tool;
import org.apache.hadoop.util.ToolRunner;
import org.apache.log4j.Logger;


public class PACC extends Configured implements Tool{

	private Logger logger = Logger.getLogger(getClass());

	/**
	 * the main entry point
	 * @param args
	 * [0]: input file path. [1]: output file path.
	 * Tool runner arguments:
	 * -D verbose if true, it prints log verbosely (default: false).
	 * -D mapred.reduce.tasks number of reduce tasks (default: 1).
	 * -D numPartitions the number of partitions (default: the number of reduce tasks).
	 * -D localThreshold pacc run a in-memory cc algorithm if the remaining number of edges is below
	 * this value (default: 1000000).
	 * @throws Exception by hadoop
	 */
	public static void main(String[] args) throws Exception{
		ToolRunner.run(new PACC(), args);
	}

	/**
	 * submit the hadoop job
	 * @param args
	 * [0]: input file path. [1]: output file path.
	 * Tool runner options:
	 * -D verbose if true, it prints log verbosely (default: false).
	 * -D mapred.reduce.tasks number of reduce tasks (default: 1).
	 * -D numPartitions the number of partitions (default: the number of reduce tasks).
	 * -D localThreshold pacc run an in-memory cc algorithm if the remaining number of edges is below
	 * this value (default: 1000000).
	 * @throws Exception by hadoop
	 */
	public int run(String[] args) throws Exception{
		
		Configuration conf = getConf();
		
		Path input = new Path(args[0]);
		Path output = new Path(args[1]);
		boolean verbose = conf.getBoolean("verbose", false);
		int numReduceTasks = conf.getInt("mapred.reduce.tasks", 1);
		int numPartitions = conf.getInt("numPartitions", numReduceTasks);
		long localThreshold = conf.getLong("localThreshold", 1000000);
		conf.setInt("numPartitions", numPartitions);
		conf.setLong("mapred.task.timeout", 0L);

		logger.info("Input                : " + input);
		logger.info("Output               : " + output);
		logger.info("Number of partitions : " + numPartitions);
		logger.info("Local threshold      : " + localThreshold);

		FileSystem fs = FileSystem.get(conf);
		
		fs.delete(output, true);
		
		long time = System.currentTimeMillis();
		long totalTime = time;



		Initialization init = new Initialization(input, output.suffix("_0/out"), verbose);
		
		ToolRunner.run(conf, init, null);

		logger.info("Round 0 (init) ends :\t" + ((System.currentTimeMillis() - time)/1000.0));

		PALargeStarOpt largeStar;
		PASmallStarOpt smallStar;
		
		long numEdges = init.outputSize;
		long numChanges;
		boolean converge;
		int i=0;
		
		do{

			if(numEdges > localThreshold){

				time = System.currentTimeMillis();

				largeStar = new PALargeStarOpt(output.suffix("_" + i + "/out"), output.suffix("_large_" + i), verbose); 
				ToolRunner.run(conf, largeStar, null);
				fs.delete(output.suffix("_" + i + "/out"), true);

				smallStar = new PASmallStarOpt(output.suffix("_large_" + i + "/out"), output.suffix("_" + (i + 1)), verbose);
				ToolRunner.run(conf, smallStar, null);
				fs.delete(output.suffix("_large_" + i + "/out"), true);

				logger.info(String.format("Round %d (star) ends :\tlout(%d)\tlcc(%d)\tlin(%d)\tsout(%d)\tsin(%d)\t%.2fs",
						i, largeStar.outSize, largeStar.ccSize, largeStar.inSize, smallStar.outSize, smallStar.inSize,
						((System.currentTimeMillis() - time) / 1000.0)));

				numChanges = largeStar.numChanges + smallStar.numChanges;
				numEdges = smallStar.outSize;
				
				converge = (numChanges == 0);
				
			}
			else{

				UnionFind lcc = new UnionFind(output.suffix("_" + i + "/out"), output.suffix("_" + (i+1) + "/out"));
				
				time = System.currentTimeMillis();
				
				ToolRunner.run(conf, lcc, null);

				logger.info(String.format("Round %d (local) ends :\tout(%d)\t%.2fs",
						i, lcc.outputSize, ((System.currentTimeMillis() - time) / 1000.0)));

				fs.delete(output.suffix("_" + i + "/out"), true);
				
				numEdges = lcc.outputSize;
				converge = true;
			}
			
			i++;
			
			
		}while(!converge && fs.exists(output.suffix("_" + i + "/out")));
		
		time = System.currentTimeMillis();
		
		Finalization fin = new Finalization(output, i, verbose);
		
		ToolRunner.run(conf, fin, null);

		logger.info(String.format("Round %d (final) ends :\t%.2fs",
				i, ((System.currentTimeMillis() - time) / 1000.0)));

		for(int r = 0; r <= i; r++){
			fs.delete(output.suffix("_"+r), true);
			fs.delete(output.suffix("_large_"+r), true);
		}
		
		System.out.print("[PACC-end]\t" + input + "\t" + output + "\t" + numPartitions + "\t" + numReduceTasks + "\t" + localThreshold + "\t" + (i+1) + "\t");
		System.out.print( ((System.currentTimeMillis() - totalTime)/1000.0) + "\t" );
		System.out.println("# input output numPartitions numReduceTasks localThreshold numRounds time(sec)");
		
		
		return 0;
	}
}
