/*
 * PegasusN: Peta-Scale Graph Mining System (Pegasus v3.0)
 * Authors: Chiwan Park, Ha-Myung Park, U Kang
 *
 * This software is licensed under Apache License, Version 2.0 (the  "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 * -------------------------------------------------------------------------
 * File: UnionFind.java
 * - the union-find algorithm for finding connected components in a graph.
 * Version: 3.0
 */

package pegasus.mapreduce.cc;

import gnu.trove.impl.Constants;
import gnu.trove.map.hash.TLongLongHashMap;
import org.apache.hadoop.conf.Configuration;
import org.apache.hadoop.conf.Configured;
import org.apache.hadoop.fs.FileStatus;
import org.apache.hadoop.fs.FileSystem;
import org.apache.hadoop.fs.Path;
import org.apache.hadoop.io.LongWritable;
import org.apache.hadoop.io.SequenceFile;
import org.apache.hadoop.util.Tool;

import java.io.IOException;

public class UnionFind extends Configured implements Tool {

	private TLongLongHashMap parent = new TLongLongHashMap(Constants.DEFAULT_CAPACITY,
			Constants.DEFAULT_LOAD_FACTOR, -1, -1);

	private final Path input;
	private final Path output;
	long outputSize = 0;

	/**
	 * constructor
	 * @param input file path
	 * @param output file path
	 */
	UnionFind(Path input, Path output) {
		this.input = input;
		this.output = output;
	}

	/**
	 * run union-find.
	 * @param args tool runner arguments
	 * @return not used
	 * @throws Exception by hadoop for sequence file I/O
	 */
	public int run(String[] args) throws Exception {

		Configuration conf = getConf();
		FileSystem fs = FileSystem.get(conf);

		fs.delete(output, true);

		LongWritable iKey = new LongWritable();
		LongWritable iValue = new LongWritable();

		FileStatus[] status = fs.listStatus(input, path -> path.getName().startsWith("part"));

		for (FileStatus statu : status) {
			try {
				SequenceFile.Reader sr = new SequenceFile.Reader(fs, statu.getPath(), conf);

				while (sr.next(iKey, iValue)) {
					long u = iKey.get();
					long v = iValue.get();
					if (u < 0) u = ~u;

					if (find(u) != find(v)) {
						union(u, v);
					}
				}
				sr.close();

			} catch (Exception ignored) {}
		}

		final SequenceFile.Writer out = new SequenceFile.Writer(fs, conf, output,
				LongWritable.class, LongWritable.class);

		final LongWritable ou = new LongWritable();
		final LongWritable ov = new LongWritable();

		parent.forEachKey(u -> {
            try{
                long v = find(u);
                if(u != v){
                    ou.set(u);
                    ov.set(v);
                    out.append(ou, ov);
                    outputSize++;
                }
            } catch (IOException ignored){}
            return true;
        });

		out.close();

		return 0;
	}

	/**
	 * the find function of the union-find algorithm.
	 * parent table is updated as a side effect
	 * @param x node to process
	 * @return component id (smallest node reachable to node x)
	 */
	private long find(long x) {

		long p = parent.get(x);
		if(p != -1){
			long new_p = find(p);
			parent.put(x, new_p);
			return new_p;
		}
		else{
			return x;
		}

	}

	/**
	 * the union function of the union-find algorithm.
	 * It updates parent table to union the two
	 * @param a a node
	 * @param b another node
	 */
	private void union(long a, long b) {
		long r1 = find(a);
		long r2 = find(b);

		if(r1 > r2){
			parent.put(r1, r2);
		}
		else if(r1 < r2){
			parent.put(r2, r1);
		}
	}
}
