package pegasus.mapreduce.cc.old;

import org.apache.hadoop.conf.Configuration;
import org.apache.hadoop.conf.Configured;
import org.apache.hadoop.fs.FileSystem;
import org.apache.hadoop.fs.Path;
import org.apache.hadoop.io.LongWritable;
import org.apache.hadoop.mapreduce.Job;
import org.apache.hadoop.mapreduce.Mapper;
import org.apache.hadoop.mapreduce.Reducer;
import org.apache.hadoop.mapreduce.TaskCounter;
import org.apache.hadoop.mapreduce.lib.input.FileInputFormat;
import org.apache.hadoop.mapreduce.lib.input.SequenceFileInputFormat;
import org.apache.hadoop.mapreduce.lib.output.FileOutputFormat;
import org.apache.hadoop.mapreduce.lib.output.LazyOutputFormat;
import org.apache.hadoop.mapreduce.lib.output.MultipleOutputs;
import org.apache.hadoop.mapreduce.lib.output.SequenceFileOutputFormat;
import org.apache.hadoop.util.Tool;
import org.apache.hadoop.util.ToolRunner;

import java.io.IOException;
import java.util.ArrayList;
import java.util.List;

public class PASmallStarOpt extends Configured implements Tool{
	
	private final Path input;
	private final Path output;
	private final String title;
	private final boolean verbose;
	public long numChanges;
	public long inputSize;
	public long outSize;
	public long inSize;
	
	
	public PASmallStarOpt(Path input, Path output, boolean verbose){
		this.input = input;
		this.output = output;
		this.verbose = verbose;
		this.title = String.format("[%s]%s", this.getClass().getSimpleName(), input.getName());
	}
	
	public static void main(String[] args) throws Exception{
		
		Path input = new Path(args[0]);
		Path output = new Path(args[1]);
		
		ToolRunner.run(new PASmallStarOpt(input, output, true), args);
	}

	public int run(String[] args) throws Exception{
		
		Configuration conf = getConf(); 
		
		Job job = new Job(conf, title);
		
		job.setJarByClass(this.getClass());
		
		job.setMapOutputKeyClass(LongWritable.class);
		job.setMapOutputValueClass(LongWritable.class);
		job.setOutputKeyClass(LongWritable.class);
		job.setOutputValueClass(LongWritable.class);
		
		job.setMapperClass(ColorSmallStarMapper.class);
		job.setReducerClass(ColorSmallStarReducer.class);
		job.setCombinerClass(ColorSmallStarCombiner.class);
		job.setInputFormatClass(SequenceFileInputFormat.class);
		LazyOutputFormat.setOutputFormatClass(job, SequenceFileOutputFormat.class);
		
		FileInputFormat.addInputPath(job, input);
		FileOutputFormat.setOutputPath(job, output);
		
		FileSystem fs = FileSystem.get(conf);
		
		
		fs.delete(output, true);
		
		if(fs.exists(input)){
			job.waitForCompletion(verbose);
			this.numChanges = job.getCounters().findCounter(Counters.NUM_CHANGES).getValue();
			this.inputSize = job.getCounters().findCounter(TaskCounter.MAP_INPUT_RECORDS).getValue();
			this.outSize = job.getCounters().findCounter(Counters.OUT_SIZE).getValue();
			this.inSize = job.getCounters().findCounter(Counters.IN_SIZE).getValue();
		}
		
		return 0;
	}

	static public class ColorSmallStarMapper extends Mapper<LongWritable, LongWritable, LongWritable, LongWritable>{

		@Override
		protected void map(LongWritable u, LongWritable v,
				Context context)
				throws IOException, InterruptedException{
				context.write(u, v);
				context.write(v, u);
		}
	}
	
	static public class ColorSmallStarCombiner extends Reducer<LongWritable, LongWritable, LongWritable, LongWritable>{

		
		@Override
		protected void reduce(LongWritable key, Iterable<LongWritable> values,
				Context context)
				throws IOException, InterruptedException{
			
			long u = key.get();
			
			long m = u;
			
			for(LongWritable _v : values){
				long v = _v.get();
				if(v < u){
					context.write(key, _v);
				}
				else if(v > m) m = v;
			}
			
			if(u != m){
				context.write(key, new LongWritable(m));
			}
			
		}

	}
	
	static public class ColorSmallStarReducer extends Reducer<LongWritable, LongWritable, LongWritable, LongWritable>{

		@Override
		protected void cleanup(
				Context context)
				throws IOException, InterruptedException{
			
			mout.close();
		}

		private int numPartitions;
		List<List<Long>> vLists;
		ArrayList<Long> mList;
		MultipleOutputs<LongWritable, LongWritable> mout;
		
		@Override
		protected void setup(
				Context context)
				throws IOException, InterruptedException{

			numPartitions = context.getConfiguration().getInt("numPartitions", 1);
			vLists = new ArrayList<List<Long>>(numPartitions);
			mList = new ArrayList<Long>(numPartitions);
			
			for(int i = 0; i < numPartitions; i++){
				vLists.add(new ArrayList<Long>());
				mList.add(Long.MAX_VALUE);
			}
			
			mout = new MultipleOutputs<LongWritable, LongWritable>(context);
			
		}
		
		LongWritable om = new LongWritable();
		LongWritable ov = new LongWritable();
		LongWritable ou = new LongWritable();
		
		@Override
		protected void reduce(LongWritable key, Iterable<LongWritable> values,
				Context context)
				throws IOException, InterruptedException{
			
			long u = key.get();
			int uPartition = (int) (u % numPartitions);
			long min = u;
			long max = u;
			long numChanges = 0;
			long outSize = 0;
			long inSize = 0;
			
			//initialize
			for(int i = 0; i < numPartitions; i++){
				vLists.get(i).clear();
				mList.set(i, Long.MAX_VALUE);
			}
			
			//read values and store them partitions.
			for(LongWritable _v : values){
				long v = _v.get();

				if(v < u){
					int vPartition = (int) (v % numPartitions);

					mList.set(vPartition, Math.min(mList.get(vPartition), v));

					vLists.get(vPartition).add(v);
					if(v < min) min = v;
				}
				else{
					if(v > max) max = v;
				}
				
			}
			
			//If u has no ascending neighbor, quit.
			if(min == u) return;
			
			long um = mList.get(uPartition);
			
			boolean uIsLocalRoot = um >= u;
			boolean uIsLeaf = max == u;
			
			
			
			ou.set(u);
			
			if(uIsLocalRoot){
				om.set(min);
				if(uIsLeaf){
					ou.set(-u-1);
					mout.write(ou, om, "out/part");
					outSize++;
				}
				else{
					mout.write(ou, om, "out/part");
					outSize++;
				}
			}
			else{
				om.set(um);
				if(uIsLeaf){
					mout.write(ou, om, "in/part");
					inSize++;
				}
				else{
					mout.write(ou, om, "out/part");
					outSize++;
				}
			}
			
			
			for(int i = 0; i < numPartitions; i++){
				
				List<Long> vList = vLists.get(i);
				
				if(vList.isEmpty()) continue;
				
				long vMin = mList.get(i);
				
				om.set(vMin);
				
				for(long v : vList){
					if(v != vMin){
						ov.set(v);
						mout.write(ov, om, "out/part");
						numChanges++;
						outSize++;
					}
				}
				
			}

			om.set(min);
			
			for(int i = 0; i < numPartitions; i++){
				
				long vMin = mList.get(i);
				
				if(vMin != Long.MAX_VALUE && vMin != min){
					ov.set(vMin);
					mout.write(ov, om, "out/part");
					numChanges++;
					outSize++;
				}
			}
			
			context.getCounter(Counters.NUM_CHANGES).increment(numChanges);
			context.getCounter(Counters.OUT_SIZE).increment(outSize);
			context.getCounter(Counters.IN_SIZE).increment(inSize);
		}
		
	}
	
}
