/*
 * PegasusN: Peta-Scale Graph Mining System v3.0
 * Authors: Ha-Myung Park, Chiwan Park, U Kang
 *
 * This software is licensed under Apache License, Version 2.0 (the  "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 * -------------------------------------------------------------------------
 * File: BitStringsVectorCache.java
 * - Represents a vector block contains bitstring values.
 * Version: 3.0
 */
package pegasus.mapreduce.common.cache;

import pegasus.mapreduce.common.Constants;
import pegasus.mapreduce.common.types.BitStrings;
import pegasus.mapreduce.common.utils.IDUtils;

import java.util.Arrays;

/**
 * Represents a vector block contains bitstring values.
 */
public final class BitStringsVectorCache extends BaseVectorCache<BitStrings> {
  private final long[][] values;
  private final long[] zeroValue;
  private final BitStrings value = new BitStrings();

  public BitStringsVectorCache(
      final int numBlocks, final int blockId, final long numVertices, final BitStrings zeroValue) {
    super(numBlocks, blockId, numVertices);

    this.values = new long[Constants.DefaultNumBitstrings][size];
    this.zeroValue = new long[Constants.DefaultNumBitstrings];
    System.arraycopy(zeroValue.getBits(), 0, this.zeroValue, 0, Constants.DefaultNumBitstrings);
    clear();
  }

  @Override
  public final BitStrings get(final int i) {
    for (int k = 0; k < Constants.DefaultNumBitstrings; ++k) {
      value.setBit(k, values[k][i]);
    }
    return value;
  }

  @Override
  public final void put(final int i, final BitStrings value) {
    super.put(i, value);
    for (int k = 0; k < Constants.DefaultNumBitstrings; ++k) {
      values[k][i] = value.getBit(k);
    }
  }

  @Override
  public final void addTo(final int i, final BitStrings value) {
    super.put(i, value);
    for (int k = 0; k < Constants.DefaultNumBitstrings; ++k) {
      values[k][i] |= value.getBit(k);
    }
  }

  @Override
  public final void clear() {
    super.clear();
    for (int i = 0; i < Constants.DefaultNumBitstrings; ++i) {
      Arrays.fill(values[i], zeroValue[i]);
    }
  }

  @Override
  public VectorIterator<BitStrings> iterator(final boolean onlyNonZero, final BitSet filterSet) {
    return new VectorIterator<BitStrings>() {
      private int pos = 0;
      private int id;
      private final BitStrings value = new BitStrings();

      @Override
      public final boolean next() {
        while (pos < size && IDUtils.localToGlobal(numBlocks, blockId, pos) < numVertices) {
          if (contains(pos) &&
              (filterSet == null || filterSet.contains(pos)) &&
              (!onlyNonZero || !isEmpty())) {
            id = pos;
            for (int i = 0; i < Constants.DefaultNumBitstrings; ++i) {
              value.setBit(i, values[i][pos]);
            }
            ++pos;
            return true;
          }
          ++pos;
        }
        return false;
      }

      @Override
      public final int currentId() {
        return id;
      }

      @Override
      public final BitStrings currentValue() {
        return value;
      }

      private boolean isEmpty() {
        for (int i = 0; i < Constants.DefaultNumBitstrings; ++i) {
          if (values[i][pos] != 0) {
            return false;
          }
        }
        return true;
      }
    };
  }
}

