/*
 * PegasusN: Peta-Scale Graph Mining System v3.0
 * Authors: Chiwan Park, Ha-Myung Park, U Kang
 *
 * This software is licensed under Apache License, Version 2.0 (the  "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 * -------------------------------------------------------------------------
 * File: FileUtils.java
 * - Provides some utility methods manipulating files.
 * Version: 3.0
 */
package pegasus.mapreduce.common.cache;

import org.apache.hadoop.conf.Configuration;
import org.apache.hadoop.fs.FileSystem;
import org.apache.hadoop.fs.Path;
import org.apache.hadoop.io.compress.CompressionCodec;
import org.apache.hadoop.io.compress.DefaultCodec;
import org.apache.hadoop.io.compress.SnappyCodec;
import org.apache.hadoop.mapreduce.TaskAttemptContext;
import org.apache.hadoop.util.ReflectionUtils;

import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;

/**
 * An utility class to manipulate files.
 */
public class FileUtils {
  private static CompressionCodec preferredCodec = null;

  /**
   * Returns a temporary directory path.
   *
   * @param name    The name of directory
   * @param context Hadoop context to create new directory
   * @return A temporary directory path
   */
  public static Path getLocalTemporaryPath(final String name, final TaskAttemptContext context) {
    return new Path(System.getProperty("java.io.tmpdir"), new Path(context.getJobID().toString(), name));
  }

  /**
   * Creates an input stream from the given path of a file. This method assumes that the file is
   * compressed by the preferred compression codec.
   *
   * @param path    The path of a file
   * @param context Hadoop context to open a file
   * @return An input stream from the given file
   * @throws IOException
   */
  public static DataInputStream getTemporaryFileInputStream(final Path path, final TaskAttemptContext context)
      throws IOException {
    final FileSystem fs = FileSystem.getLocal(context.getConfiguration());
    final CompressionCodec codec = getPreferredCompressionCodec(context);

    return new DataInputStream(codec.createInputStream(fs.open(path)));
  }

  /**
   * Creates an output stream to the given path of a file. This method compresses the values to be stored.
   *
   * @param path    The path of a file
   * @param context Hadoop context to create a file
   * @return An output stream to the given file
   * @throws IOException
   */
  public static DataOutputStream getTemporaryFileOutputStream(final Path path, final TaskAttemptContext context)
      throws IOException {
    final FileSystem fs = FileSystem.getLocal(context.getConfiguration());
    final CompressionCodec codec = getPreferredCompressionCodec(context);

    return new DataOutputStream(codec.createOutputStream(fs.create(path, (short) 1)));
  }

  /**
   * Returns the preferred compression codec.
   * <p>
   * This method returns SnappyCodec object if it is available, otherwise DefaultCodec object is returned.
   *
   * @param conf Hadoop configuration object
   * @return A compression codec object
   */
  public static CompressionCodec getPreferredCompressionCodec(final Configuration conf) {
    if (preferredCodec != null) {
      return preferredCodec;
    }
    Class<? extends CompressionCodec> codecClass;
    if (SnappyCodec.isNativeCodeLoaded()) {
      codecClass = SnappyCodec.class;
    } else {
      codecClass = DefaultCodec.class;
    }

    preferredCodec = ReflectionUtils.newInstance(codecClass, conf);
    return preferredCodec;
  }

  /**
   * Returns the preferred compression codec.
   *
   * See {getPreferredCompressionCodec(Configuration)}.
   *
   * @param context Hadoop context object
   * @return A compression codec object
   */
  public static CompressionCodec getPreferredCompressionCodec(final TaskAttemptContext context) {
    return getPreferredCompressionCodec(context.getConfiguration());
  }
}
