/*
 * PegasusG: Peta-Scale Graph Mining System v3.0
 * Authors: Chiwan Park, Ha-Myung Park, U Kang
 *
 * This software is licensed under Apache License, Version 2.0 (the  "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 * -------------------------------------------------------------------------
 * File: ZookeeperCounter.java
 * - Implementation of counter based on zookeeper
 * Version: 3.0
 */
package pegasus.mapreduce.common.coordination;

import org.apache.zookeeper.ZooKeeper;
import pegasus.mapreduce.common.Constants;
import pegasus.mapreduce.common.utils.ZookeeperUtils;

import java.util.Comparator;
import java.util.function.BinaryOperator;
import java.util.stream.Stream;

public abstract class ZookeeperCounter<T extends Comparable<T>> implements Counter<T> {
  private final ZooKeeper zk;
  private final String counterPath;
  private final String myCounterPath;

  public ZookeeperCounter(final ZooKeeper zk, final String name, final int workerId) {
    this.zk = zk;
    counterPath = String.format("%s/%s", Constants.CounterRoot, name);
    myCounterPath = String.format("%s/%d", counterPath, workerId);

    if (!ZookeeperUtils.exists(zk, counterPath)) {
      ZookeeperUtils.createPersistentNode(zk, counterPath);
    }
  }

  @Override
  public void update(final T value) {
    ZookeeperUtils.createOrUpdatePersistentNode(zk, myCounterPath, valueToString(value));
  }

  @Override
  public T getMinimum() {
    return getCounters().min(Comparator.naturalOrder()).orElse(positiveInfinite());
  }

  @Override
  public T getMaximum() {
    return getCounters().max(Comparator.naturalOrder()).orElse(negativeInfinite());
  }

  @Override
  public T getSum() {
    return getCounters().reduce(zero(), sumOp());
  }

  protected abstract T negativeInfinite();

  protected abstract T positiveInfinite();

  protected abstract T zero();

  protected abstract BinaryOperator<T> sumOp();

  protected abstract T valueFromString(final String string);

  protected abstract String valueToString(final T value);

  private Stream<T> getCounters() {
    return ZookeeperUtils.getChildren(zk, counterPath).parallelStream().map(
        path -> valueFromString(ZookeeperUtils.getNodeData(zk, String.format("%s/%s", counterPath, path))));
  }
}
