/*
 * PegasusN: Peta-Scale Graph Mining System v3.0
 * Authors: Ha-Myung Park, Chiwan Park, U Kang
 *
 * This software is licensed under Apache License, Version 2.0 (the  "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 * -------------------------------------------------------------------------
 * File: LongIntPairMapWritable.java
 * - Represents a map contains (long, int) pair values
 * Version: 3.0
 */
package pegasus.mapreduce.common.types;

import it.unimi.dsi.fastutil.longs.Long2IntMap;
import it.unimi.dsi.fastutil.longs.Long2IntOpenHashMap;
import org.apache.hadoop.io.Writable;
import pegasus.mapreduce.common.Constants;

import java.io.DataInput;
import java.io.DataOutput;
import java.io.IOException;
import java.util.Iterator;

/**
 * Represents a map contains (long, int) pair values.
 */
public class LongIntPairMapWritable implements Writable {
  private Long2IntOpenHashMap map;

  public LongIntPairMapWritable() {
    this(Constants.InitArrayWritableSize);
  }

  public LongIntPairMapWritable(final int size) {
    map = new Long2IntOpenHashMap(size, 1.0f);
    map.defaultReturnValue(0);
  }

  public LongIntPairMapWritable(final LongIntPairMapWritable other) {
    final int size = Math.max(other.size(), 4);
    map = new Long2IntOpenHashMap(size);
    merge(other);
  }

  public int get(final long id) {
    return map.get(id);
  }

  public long getMaximumByValue() {
    long maxKey = -1L;
    int maxValue = Integer.MIN_VALUE;

    final Iterator<Long2IntMap.Entry> iter = map.long2IntEntrySet().fastIterator();
    while (iter.hasNext()) {
      final Long2IntMap.Entry entry = iter.next();
      if (entry.getIntValue() > maxValue || (entry.getIntValue() == maxValue && entry.getLongKey() < maxKey)) {
        maxKey = entry.getLongKey();
        maxValue = entry.getIntValue();
      }
    }

    return maxKey;
  }

  public void addTo(final long id, final int value) {
    map.addTo(id, value);
  }

  public void merge(final LongIntPairMapWritable other) {
    final Iterator<Long2IntMap.Entry> iter = other.map.long2IntEntrySet().fastIterator();
    while (iter.hasNext()) {
      final Long2IntMap.Entry entry = iter.next();
      map.addTo(entry.getLongKey(), entry.getIntValue());
    }
  }

  public void clear() {
    map.clear();
  }

  public int size() {
    return map.size();
  }

  @Override
  public void write(final DataOutput out) throws IOException {
    out.writeInt(map.size());

    final Iterator<Long2IntMap.Entry> iter = map.long2IntEntrySet().fastIterator();
    while (iter.hasNext()) {
      final Long2IntMap.Entry entry = iter.next();
      out.writeLong(entry.getLongKey());
      out.writeInt(entry.getIntValue());
    }
  }

  @Override
  public void readFields(final DataInput in) throws IOException {
    final int size = in.readInt();
    for (int i = 0; i < size; ++i) {
      map.put(in.readLong(), in.readInt());
    }
  }
}
