/*
 * PegasusN: Peta-Scale Graph Mining System (Pegasus v3.0)
 * Authors: Ha-Myung Park, Chiwan Park, U Kang
 *
 * This software is licensed under Apache License, Version 2.0 (the  "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 * -------------------------------------------------------------------------
 * File: IOUtils.java
 * - Provides IO utility methods to manipulate files on secondary or distributed storage.
 * Version: 3.0
 */
package pegasus.mapreduce.common.utils;

import org.apache.hadoop.conf.Configuration;
import org.apache.hadoop.fs.FSDataInputStream;
import org.apache.hadoop.fs.FileSystem;
import org.apache.hadoop.fs.Path;
import org.apache.hadoop.io.IntWritable;
import org.apache.hadoop.io.Writable;
import org.apache.hadoop.io.compress.CompressionCodec;
import org.apache.hadoop.io.compress.DefaultCodec;
import org.apache.hadoop.io.compress.Lz4Codec;
import org.apache.hadoop.io.compress.SnappyCodec;
import org.apache.hadoop.mapreduce.RecordReader;
import org.apache.hadoop.mapreduce.RecordWriter;
import org.apache.hadoop.mapreduce.TaskAttemptContext;
import org.apache.hadoop.mapreduce.lib.output.FileOutputFormat;
import org.apache.hadoop.util.ReflectionUtils;
import pegasus.mapreduce.common.io.VectorRecordReader;
import pegasus.mapreduce.common.io.VectorRecordWriter;

import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;
import java.io.OutputStream;

public class IOUtils {
  private static CompressionCodec preferredCodec = null;

  /**
   * Returns a path of intermediate vector file.
   *
   * @param context The current execution context
   * @param prefix  A prefix string of vector file name
   * @param blockId The id of vector block
   * @return The path of intermediate vector file
   * @throws IOException
   */
  public static Path getIntermediateVectorPath(
      final TaskAttemptContext context, final String prefix, final int blockId) throws IOException {
    final Path rootPath = new Path(FileOutputFormat.getOutputPath(context), "intermediate");
    return new Path(rootPath, String.format("%s-%05d", prefix, blockId));
  }

  /**
   * Opens an intermediate vector file and returns a writer object to write values to the file.
   *
   * @param context The current execution context
   * @param prefix  A prefix string of vector file name
   * @param blockId The id of vector block
   * @param <T>     The type parameter of the values to be written
   * @return A writer object to write values to the file.
   * @throws IOException
   */
  public static <T extends Writable> RecordWriter<IntWritable, T> getIntermediateVectorWriter(
      final TaskAttemptContext context, final String prefix, final int blockId)
      throws IOException {
    return getIntermediateVectorWriter(context, getIntermediateVectorPath(context, prefix, blockId));
  }

  /**
   * Opens an intermediate vector file and returns a writer object to write values to the file.
   *
   * @param context The current execution context
   * @param path    The path of intermediate vector file
   * @param <T>     The type parameter of the values to be written
   * @return A writer object to write values to the file.
   * @throws IOException
   */
  @SuppressWarnings("unchecked")
  public static <T extends Writable> RecordWriter<IntWritable, T> getIntermediateVectorWriter(
      final TaskAttemptContext context, final Path path) throws IOException {
    final Configuration conf = context.getConfiguration();
    final int bufSize = conf.getInt("io.file.buffer.size", 4096);
    final FileSystem fs = path.getFileSystem(conf);

    final OutputStream out = fs.create(path, true, bufSize, (short) 1, fs.getDefaultBlockSize(path), context);
    return new VectorRecordWriter<>(out, context);
  }

  /**
   * Opens the intermediate vector file and returns a reader object to read values from the file.
   *
   * @param context    The current execution context
   * @param prefix     A prefix string of vector file name
   * @param blockId    The id of vector block
   * @param valueClass A class of the values to be read
   * @param <T>        The type parameter of the values to be read
   * @return A reader object to read values from the file
   * @throws IOException
   */
  public static <T extends Writable> RecordReader<IntWritable, T> getIntermediateVectorReader(
      final TaskAttemptContext context, final String prefix, final int blockId, final Class<T> valueClass)
      throws IOException {
    return getIntermediateVectorReader(context, getIntermediateVectorPath(context, prefix, blockId), valueClass);
  }

  /**
   * Opens the intermediate vector file and returns a reader object to read values from the file.
   *
   * @param context    The current execution context
   * @param path       The path of intermediate vector file
   * @param valueClass A class of the values to be read
   * @param <T>        The type parameter of the values to be read
   * @return A reader object to read values from the file
   * @throws IOException
   */
  @SuppressWarnings("unchecked")
  public static <T extends Writable> RecordReader<IntWritable, T> getIntermediateVectorReader(
      final TaskAttemptContext context, final Path path, final Class<T> valueClass)
      throws IOException {
    final Configuration conf = context.getConfiguration();
    final FileSystem fs = path.getFileSystem(conf);

    if (!fs.exists(path) || fs.getFileStatus(path).getLen() == 0) return null;
    final FSDataInputStream in = fs.open(path);

    return new VectorRecordReader<>(in, context, valueClass);
  }

  /**
   * Returns a path of local temporary file.
   *
   * @param name    The name of temporary file
   * @param context The current execution context
   * @return A path of local temporary file
   */
  public static Path getLocalTemporaryPath(final String name, final TaskAttemptContext context) {
    return getLocalTemporaryPath(name, context.getJobID().getId());
  }

  /**
   * Returns a path of local temporary file.
   *
   * @param name The name of temporary file
   * @param id   A id of temporary file
   * @return A path of local temporary file
   */
  public static Path getLocalTemporaryPath(final String name, final int id) {
    return getLocalTemporaryPath(String.format("%s-%d", name, id));
  }

  /**
   * Returns a path of local temporary file.
   *
   * @param name A name of temporary file
   * @return A path of the local temporary file
   */
  public static Path getLocalTemporaryPath(final String name) {
    return new Path(System.getProperty("java.io.tmpdir"), name);
  }

  /**
   * Opens the given local temporary file and returns an input stream object to read values from the file.
   *
   * @param path    The path of local temporary file
   * @param context The current execution context
   * @return An input stream object to read values from the file
   * @throws IOException
   */
  public static DataInputStream getTemporaryFileInputStream(final Path path, final TaskAttemptContext context)
      throws IOException {
    final FileSystem fs = FileSystem.getLocal(context.getConfiguration());
    final CompressionCodec codec = getPreferredCompressionCodec(context);

    return new DataInputStream(codec.createInputStream(fs.open(path)));
  }

  /**
   * Creates a local temporary file and returns an output stream object to write values to the file.
   *
   * @param path    A path of local temporary file
   * @param context The current execution context
   * @return An output stream object to write values to the file
   * @throws IOException
   */
  public static DataOutputStream getTemporaryFileOutputStream(final Path path, final TaskAttemptContext context)
      throws IOException {
    final FileSystem fs = FileSystem.getLocal(context.getConfiguration());
    final CompressionCodec codec = getPreferredCompressionCodec(context);

    return new DataOutputStream(codec.createOutputStream(fs.create(path, (short) 1)));
  }

  /**
   * Returns a codec object which is preferred.
   *
   * @param conf The configuration of current execution
   * @return The codec object which is preferred
   */
  public static CompressionCodec getPreferredCompressionCodec(final Configuration conf) {
    if (preferredCodec != null) {
      return preferredCodec;
    }
    Class<? extends CompressionCodec> codecClass;
    if (Lz4Codec.isNativeCodeLoaded()) {
      codecClass = Lz4Codec.class;
    } else if (SnappyCodec.isNativeCodeLoaded()) {
      codecClass = SnappyCodec.class;
    } else {
      codecClass = DefaultCodec.class;
    }

    preferredCodec = ReflectionUtils.newInstance(codecClass, conf);
    return preferredCodec;
  }

  /**
   * Returns a codec object which is preferred.
   *
   * @param context The current execution context
   * @return The codec object which is preferred
   */
  public static CompressionCodec getPreferredCompressionCodec(final TaskAttemptContext context) {
    return getPreferredCompressionCodec(context.getConfiguration());
  }
}
