/*
 * PegasusN: Peta-Scale Graph Mining System (Pegasus v3.0)
 * Authors: ByungSoo Jeon, Inah Jeon, Chiwan Park, Ha-Myung Park, U Kang
 *
 * This software is licensed under Apache License, Version 2.0 (the  "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 * -------------------------------------------------------------------------
 * File: KroneckerGenerator.java
 * - generate a kronecker graph
 * Version: 3.0
 */

package pegasus.mapreduce.graphgen;

import java.io.*;
import java.util.*;

import org.apache.hadoop.conf.Configuration;
import org.apache.hadoop.conf.Configured;
import org.apache.hadoop.fs.FileSystem;
import org.apache.hadoop.fs.Path;
import org.apache.hadoop.io.IntWritable;
import org.apache.hadoop.io.LongWritable;
import org.apache.hadoop.io.SequenceFile;
import org.apache.hadoop.io.Writable;
import org.apache.hadoop.mapreduce.Job;
import org.apache.hadoop.mapreduce.Mapper;
import org.apache.hadoop.mapreduce.lib.input.FileInputFormat;
import org.apache.hadoop.mapreduce.lib.input.SequenceFileInputFormat;
import org.apache.hadoop.mapreduce.lib.output.FileOutputFormat;
import org.apache.hadoop.mapreduce.lib.output.TextOutputFormat;
import org.apache.hadoop.util.Tool;
import org.apache.hadoop.util.ToolRunner;
import org.apache.log4j.Logger;

public class KroneckerGenerator extends Configured implements Tool {

	private static Logger logger = Logger.getLogger(KroneckerGenerator.class);

	/**
	 * The main entry point
	 * @param args [0]: level of recursion. [1]: degree of parallelism (the number of tasks to run). [2]: seed
	 *             graph file path. [3]: output file path.
	 * @throws Exception
	 */
	public static void main(final String[] args) throws Exception {
		ToolRunner.run(new KroneckerGenerator(), args);
	}

	/**
	 * submit the hadoop job of the Kronecker graph generator.
	 * @param args [0]: level of recursion. [1]: degree of parallelism (the number of tasks to run). [2]: seed
	 *             graph file path. [3]: output file path.
	 * @return not used
	 * @throws Exception by hadoop
	 */
	public int run(String[] args) throws Exception {

		int level = Integer.parseInt(args[0]);
		int numTasks = Integer.parseInt(args[1]);
		String seedGraphFile = args[2];
		Path outputPath = new Path(args[3]);
		Path inputPath = outputPath.suffix(".input");

		logger.info(String.format("lv: %d, numTasks: %d, seedGraphFile: %s", level, numTasks, seedGraphFile));

		long[] seedGraph = parseSeedGraph(seedGraphFile);
		long numEdges = (long) Math.pow(seedGraph.length, level);

		PriorityQueue<Region> regionQueue = new PriorityQueue<>();
		PriorityQueue<RegionSlot> regionSlotQueue = new PriorityQueue<>();

		for(int i = 0; i < numTasks; i++){
			long numEdgeAssigned = (numEdges * (i + 1) / numTasks) - (numEdges * i / numTasks);
			regionSlotQueue.add(new RegionSlot(numEdgeAssigned));
		}

		regionQueue.add(new Region(0,0,level-1, numEdges));

		while(!regionQueue.isEmpty()){
			Region region = regionQueue.poll();
			RegionSlot slot = regionSlotQueue.poll();

			if(region.level == 0 || region.numEdges <= slot.edgeCapacity){
				slot.add(region);
			}
			else{
				regionQueue.addAll(Arrays.asList(splitRegion(region, seedGraph)));
			}

			regionSlotQueue.add(slot);
		}

		//delete previous input file.
		FileSystem.get(getConf()).delete(inputPath, true);

		generateInputFile(regionSlotQueue, inputPath);

		Configuration conf = new Configuration();

		conf.set("seedGraph", serializeSeedGraph(seedGraph));

		Job job = Job.getInstance(conf, "[Kronecker]" + outputPath);
		job.setJarByClass(getClass());
		job.setOutputKeyClass(LongWritable.class);
		job.setOutputValueClass(LongWritable.class);
		job.setInputFormatClass(SequenceFileInputFormat.class);
		job.setOutputFormatClass(TextOutputFormat.class);
		job.setMapperClass(KroneckerGeneratorMapper.class);
		job.setNumReduceTasks(0);

		FileInputFormat.addInputPath(job, inputPath);
		FileOutputFormat.setOutputPath(job, outputPath);

		job.waitForCompletion(false);

		return 0;

	}

	/**
	 * Serialize the seed graph
	 * @param seedGraph seed graph
	 * @return a serialized string
	 */
	static String serializeSeedGraph(long[] seedGraph) {
		StringBuilder sb = new StringBuilder();

		for(long edge : seedGraph){
			sb.append(edge);
			sb.append('\t');
		}

		return sb.toString();
	}

	/**
	 * Deserialize a serialized seed graph
	 * @param serialSeedGraph a serialized seed graph
	 * @return seed graph
	 */
	static long[] deserializeSeedGraph(String serialSeedGraph){

		List<Long> bf = new ArrayList<>();

		StringTokenizer st = new StringTokenizer(serialSeedGraph);
		while(st.hasMoreTokens()) bf.add(Long.parseLong(st.nextToken()));

		long[] seedGraph = new long[bf.size()];
		for(int i=0; i<seedGraph.length; i++) seedGraph[i] = bf.get(i);

		return seedGraph;
	}

	/**
	 * Make the input file for the hadoop job of the Kronecker graph generator.
	 * @param regionSlotQueue priority queue of region slots
	 * @param input file path
	 * @throws IOException by hadoop
	 */
	private void generateInputFile(PriorityQueue<RegionSlot> regionSlotQueue, Path input) throws IOException {

		FileSystem fs = FileSystem.get(getConf());
		fs.delete(input, true);

		SequenceFile.Writer.Option optKey = SequenceFile.Writer.keyClass(IntWritable.class);
		SequenceFile.Writer.Option optValue = SequenceFile.Writer.valueClass(RegionSlot.class);

		int i = 0;
		IntWritable okey = new IntWritable();
		for(RegionSlot rs : regionSlotQueue){
			SequenceFile.Writer sw = SequenceFile.createWriter(new Configuration(),
					SequenceFile.Writer.file(input.suffix("/part-" + i)),
					optKey, optValue);

			okey.set(i++);
			sw.append(okey, rs);

			sw.close();
		}
	}

	/**
	 * read seed graph.
	 *
	 * @param seedGraphPath seed graph file path
	 * @return the array of edges in the seed graph. A long value represent an
	 * edge (u, v); 32-high bits are for u, 32-low bits are for v
	 */
	private long[] parseSeedGraph(String seedGraphPath) throws IOException {

		BufferedReader br = new BufferedReader(new FileReader(seedGraphPath));

		long[] seed = br.lines().mapToLong(line -> {
			StringTokenizer st = new StringTokenizer(line);
			return ((long) Integer.parseInt(st.nextToken()) << 32) | Integer.parseInt(st.nextToken());
		}).toArray();

		br.close();

		return seed;

	}

	public static class KroneckerGeneratorMapper extends Mapper<IntWritable, RegionSlot, LongWritable, LongWritable> {

		Context context;
		long[] seedGraph;

		LongWritable ou = new LongWritable();
		LongWritable ov = new LongWritable();

		/**
		 * Setup before execution. read seed graph.
		 * @param context of hadoop
		 * @throws IOException by hadoop
		 * @throws InterruptedException by hadoop
		 */
		@Override
		protected void setup(Context context) throws IOException, InterruptedException {
			this.context = context;
			this.seedGraph = deserializeSeedGraph(context.getConfiguration().get("seedGraph"));
		}

		/**
		 * The map function of the RMAT graph generator.
		 * @param key task id
		 * @param rs a region slot that contains regions to process here.
		 * @param context of hadoop
		 * @throws IOException by hadoop
		 * @throws InterruptedException by hadoop
		 */
		@Override
		protected void map(IntWritable key, RegionSlot rs, Context context) throws IOException, InterruptedException {
			for(Region r : rs.regions) generate(r);
		}

		/**
		 * generate edges in a given region.
		 * @param region to process
		 * @throws IOException by hadoop
		 * @throws InterruptedException by hadoop
		 */
		private void generate(Region region) throws IOException, InterruptedException {

			if(region.level == 0){
				for(long edge : seedGraph){
					int u = (int) (edge >>> 32); // high 32 bit
					int v = (int) edge; // low 32 bit

					ou.set(region.srcMin + u);
					ov.set(region.dstMin + v);

					context.write(ou, ov);
				}
			}
			else for(Region r : splitRegion(region, seedGraph)) generate(r);

		}
	}

	static class Region implements Comparable<Region>, Writable {

		long srcMin;
		long dstMin;
		int level;
		long numEdges;

		public Region(){};

		/**
		 * A region.
		 * @param srcMin the minimum id of source nodes.
		 * @param dstMin the minimum id of destination nodes.
		 * @param level the level
		 * @param numEdges the number of edges
		 */
		public Region(long srcMin, long dstMin, int level, long numEdges){
			this.srcMin = srcMin;
			this.dstMin = dstMin;
			this.level = level;
			this.numEdges = numEdges;
		}

		@Override
		@SuppressWarnings("NullableProblems")
		public int compareTo(Region that) {
			if(this.numEdges < that.numEdges) return 1;
			else if(this.numEdges > that.numEdges) return -1;
			else return 0;
		}

		/**
		 * write this region out
		 * @param out to write
		 * @throws IOException by hadoop
		 */
		@Override
		public void write(DataOutput out) throws IOException {
			out.writeLong(srcMin);
			out.writeLong(dstMin);
			out.writeInt(level);
			out.writeLong(numEdges);
		}

		/**
		 * read a region
		 * @param in read from it
		 * @throws IOException by hadoop
		 */
		@Override
		public void readFields(DataInput in) throws IOException {
			srcMin = in.readLong();
			dstMin = in.readLong();
			level = in.readInt();
			numEdges = in.readLong();
		}
	}

	static class RegionSlot implements Comparable<RegionSlot>, Writable{

		long edgeCapacity;
		List<Region> regions;

		public RegionSlot(){};

		/**
		 * A container of regions.
		 * @param edgeCapacity rough number of edges can accommodate. Real edge number can exceed this value.
		 */
		public RegionSlot(long edgeCapacity){
			this.edgeCapacity = edgeCapacity;
			regions = new ArrayList<>();
		}

		/**
		 * add a region
		 * @param r region
		 */
		public void add(Region r){
			regions.add(r);
			edgeCapacity -= r.numEdges;
		}

		@Override
		@SuppressWarnings("NullableProblems")
		public int compareTo(RegionSlot that) {
			if(this.edgeCapacity < that.edgeCapacity) return 1;
			else if(this.edgeCapacity > that.edgeCapacity) return -1;
			else return 0;
		}

		/**
		 * write this region slot out
		 * @param out to write
		 * @throws IOException by hadoop
		 */
		@Override
		public void write(DataOutput out) throws IOException {
			out.writeLong(edgeCapacity);
			out.writeInt(regions.size());

			for(Region r : regions) r.write(out);
		}

		/**
		 * read a region slot
		 * @param in read from it
		 * @throws IOException by hadoop
		 */
		@Override
		public void readFields(DataInput in) throws IOException {
			edgeCapacity = in.readLong();
			int numRegions = in.readInt();

			regions = new ArrayList<>();

			for(int i = 0; i < numRegions; i++){
				Region r = new Region();
				r.readFields(in);
				regions.add(r);
			}
		}
	}

	/**
	 * split a region
	 * @param region to process
	 * @return sub-regions
	 */
	static Region[] splitRegion(Region region, long[] seedGraph){

		int numSeedNodes = 0;

		for (long edge : seedGraph) {
			int u = (int) (edge >>> 32);
			int v = (int) edge;
			numSeedNodes = Math.max(numSeedNodes, Math.max(u, v));
		}
		numSeedNodes += 1;

		long unit = (long) Math.pow(numSeedNodes, region.level);

		Region[] res = new Region[seedGraph.length];

		for(int i = 0; i < seedGraph.length; i++) {

			int u = (int) (seedGraph[i] >>> 32);
			int v = (int) seedGraph[i];

			res[i] = new Region(region.srcMin + u * unit, region.dstMin + v * unit,
					region.level - 1, region.numEdges / seedGraph.length);
		}

		return res;
	}

}
