/*
 * PegasusN: Peta-Scale Graph Mining System (Pegasus v3.0)
 * Authors: Chiwan Park, Ha-Myung Park, U Kang
 *
 * This software is licensed under Apache License, Version 2.0 (the  "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 * -------------------------------------------------------------------------
 * File: RMATGenerator.java
 * - generate a rmat graph
 * Version: 3.0
 */

package pegasus.mapreduce.graphgen;

import org.apache.commons.math3.distribution.MultivariateNormalDistribution;
import org.apache.hadoop.conf.Configuration;
import org.apache.hadoop.conf.Configured;

import org.apache.hadoop.fs.FileSystem;
import org.apache.hadoop.fs.Path;
import org.apache.hadoop.io.IntWritable;
import org.apache.hadoop.io.LongWritable;
import org.apache.hadoop.io.SequenceFile;
import org.apache.hadoop.io.Writable;
import org.apache.hadoop.mapreduce.Job;
import org.apache.hadoop.mapreduce.Mapper;
import org.apache.hadoop.mapreduce.lib.input.FileInputFormat;
import org.apache.hadoop.mapreduce.lib.input.SequenceFileInputFormat;
import org.apache.hadoop.mapreduce.lib.output.FileOutputFormat;
import org.apache.hadoop.mapreduce.lib.output.TextOutputFormat;
import org.apache.hadoop.util.Tool;
import org.apache.hadoop.util.ToolRunner;
import org.apache.log4j.Logger;

import java.io.DataInput;
import java.io.DataOutput;
import java.io.IOException;
import java.util.*;


public class RMATGenerator extends Configured implements Tool {

	private static Logger logger = Logger.getLogger(RMATGenerator.class);

	long thresholdRegionEdgeNum = 10000;

    /**
     * The main entry point
     * @param args [0]: level of recursion. [1]: the number of edges. [2]: degree of parallelism (the
     *             number of tasks to run). [3], [4], [5], [6]: probabilities of quarters. [7]: noise
     *             for smoothing. [8]: output file path
     * @throws Exception
     */
	public static void main(final String[] args) throws Exception {
		ToolRunner.run(new RMATGenerator(), args);
	}

    /**
     * Submit the hadoop job of RMAT generator.
     * @param args [0]: level of recursion. [1]: the number of edges. [2]: degree of parallelism (the
     *             number of tasks to run). [3], [4], [5], [6]: probabilities of quarters. [7]: noise
     *             for smoothing. [8]: output file path
     * @return not used
     * @throws Exception by hadoop
     */
	@Override
	public int run(String[] args) throws Exception {

		int level = Integer.parseInt(args[0]);
		int numEdges = Integer.parseInt(args[1]);
		int numTasks = Integer.parseInt(args[2]);
		double[] abcd = new double[4];
		abcd[0] = Double.parseDouble(args[3]);
		abcd[1] = Double.parseDouble(args[4]);
		abcd[2] = Double.parseDouble(args[5]);
		abcd[3] = Double.parseDouble(args[6]);
		double noise = Double.parseDouble(args[7]);
        Path outputPath = new Path(args[8]);

		Path inputPath = outputPath.suffix(".input");

		logger.info(String.format("lv: %d, numEdges: $d, numTasks: %d, abcd: (%s)",
				level, numEdges, numTasks, Arrays.toString(abcd)));

		PriorityQueue<Region> regionQueue = new PriorityQueue<>();
		PriorityQueue<RegionSlot> regionSlotQueue = new PriorityQueue<>();

		for(int i = 0; i < numTasks; i++){
			long numEdgeAssigned = (numEdges * (i + 1) / numTasks) - (numEdges * i / numTasks);
			regionSlotQueue.add(new RegionSlot(numEdgeAssigned));
		}

		long numNodes = 1L << level;

		regionQueue.add(new Region(0,0, numNodes - 1, numNodes - 1, level-1, numEdges));

		double[] noises = getNoises(level,noise);

		RMAT rmat = new RMAT(abcd, noises);

		while(!regionQueue.isEmpty()){
			Region region = regionQueue.poll();
			RegionSlot slot = regionSlotQueue.poll();

			if(region.numEdges <= thresholdRegionEdgeNum || region.numEdges <= slot.edgeCapacity) slot.add(region);
			else regionQueue.addAll(Arrays.asList(rmat.splitRegion(region)));

			regionSlotQueue.add(slot);
		}

		//delete previous input file.
		FileSystem.get(getConf()).delete(inputPath, true);

		generateInputFile(regionSlotQueue, inputPath);

		Configuration conf = getConf();

		conf.set("noises", serializeDoubleArray(noises));
		conf.setDouble("abcd.a", abcd[0]);
		conf.setDouble("abcd.b", abcd[1]);
		conf.setDouble("abcd.c", abcd[2]);
		conf.setDouble("abcd.d", abcd[3]);


		Job job = Job.getInstance(conf, "[RMAT]" + outputPath);
		job.setJarByClass(getClass());
		job.setOutputKeyClass(LongWritable.class);
		job.setOutputValueClass(LongWritable.class);
		job.setInputFormatClass(SequenceFileInputFormat.class);
		job.setOutputFormatClass(TextOutputFormat.class);
		job.setMapperClass(RMATGeneratorMapper.class);
		job.setNumReduceTasks(0);

		FileInputFormat.addInputPath(job, inputPath);
		FileOutputFormat.setOutputPath(job, outputPath);

		job.waitForCompletion(false);

		return 0;

	}

    public static class RMATGeneratorMapper extends Mapper<IntWritable, RegionSlot, LongWritable, LongWritable> {

        Context context;
        double[] noises;
        double[] abcd;

        LongWritable ou = new LongWritable();
        LongWritable ov = new LongWritable();

        /**
         * Setup before execution.
         * Read noises and probabilities.
         * @param context of hadoop
         * @throws IOException by hadoop
         * @throws InterruptedException by hadoop
         */
        @Override
        protected void setup(Context context) throws IOException, InterruptedException {
            this.context = context;
            Configuration conf = context.getConfiguration();
            this.noises = deserializeDoubleArray(conf.get("noises"));
            this.abcd = new double[4];
            abcd[0] = conf.getDouble("abcd.a", 0);
            abcd[1] = conf.getDouble("abcd.b", 0);
            abcd[2] = conf.getDouble("abcd.c", 0);
            abcd[3] = conf.getDouble("abcd.d", 0);
        }

        /**
         * The map function of the RMAT graph generator.
         * @param key task id
         * @param rs a region slot that contains regions to process here.
         * @param context of hadoop
         * @throws IOException by hadoop
         * @throws InterruptedException by hadoop
         */
        @Override
        protected void map(IntWritable key, RegionSlot rs, Context context) throws IOException, InterruptedException {

            RMAT rmat = new RMAT(abcd, noises);

            rmat.setContext(context);

            for(Region r : rs.regions){
                rmat.generate(r);
            }
        }
    }

    /**
     * Make the input file for the hadoop job of the RMAT graph generator.
     * @param regionSlotQueue priority queue of region slots
     * @param input file path
     * @throws IOException by hadoop
     */
	private void generateInputFile(PriorityQueue<RegionSlot> regionSlotQueue, Path input) throws IOException {

		FileSystem fs = FileSystem.get(getConf());
		fs.delete(input, true);

		SequenceFile.Writer.Option optKey = SequenceFile.Writer.keyClass(IntWritable.class);
		SequenceFile.Writer.Option optValue = SequenceFile.Writer.valueClass(RegionSlot.class);

		int i = 0;
		IntWritable okey = new IntWritable();
		for(RegionSlot rs : regionSlotQueue){
			SequenceFile.Writer sw = SequenceFile.createWriter(new Configuration(),
					SequenceFile.Writer.file(input.suffix("/part-" + i)),
					optKey, optValue);

			okey.set(i++);
			sw.append(okey, rs);

			sw.close();
		}
	}

    /**
     * make noise for each level
     * @param level the number of noises to make
     * @param noise base noise
     * @return noises
     */
	private static double[] getNoises(int level, double noise){

		double[] noises = new double[level];

		for(int i = 0; i < level; i++)
			noises[i] = Math.random() * 2 * noise - noise;

		return noises;
	}

    /**
     * serialize a double array
     * @param arr to serialize
     * @return the serialized string
     */
	private static String serializeDoubleArray(double[] arr) {

		StringBuilder sb = new StringBuilder();

		for(int i = 0; i < arr.length; i++){
			sb.append(Double.doubleToRawLongBits(arr[i]));
			sb.append('\t');
		}

		return sb.toString();
	}

    /**
     * deserialize a serialized string for double array
     * @param serializedDoubleArray string to deserialize
     * @return a double array
     */
	private static double[] deserializeDoubleArray(String serializedDoubleArray){

		List<Double> bf = new ArrayList<>();

		StringTokenizer st = new StringTokenizer(serializedDoubleArray);
		while(st.hasMoreTokens()) bf.add(Double.longBitsToDouble(Long.parseLong(st.nextToken())));

		double[] noises = new double[bf.size()];
		for(int i=0; i<noises.length; i++) noises[i] = bf.get(i);

		return noises;
	}


	static class RMAT{

		long thresholdRegionEdgeNum = 10000;

		double[][] probs;

		Mapper<IntWritable, RegionSlot, LongWritable, LongWritable>.Context context;
		LongWritable ou = new LongWritable();
		LongWritable ov = new LongWritable();

        /**
         * In-memory RMAT generator
         * @param abcd probabilities
         * @param noises for smoothing
         */
		RMAT(double[] abcd, double[] noises){

			probs = new double[noises.length][4];

			for(int i = 0; i < noises.length; i++) {

				double mu = noises[i];

				probs[i][0] = abcd[0] - 2 * mu * abcd[0] / (abcd[0] + abcd[3]);
				probs[i][1] = abcd[1] + mu;
				probs[i][2] = abcd[2] + mu;
				probs[i][3] = abcd[3] - 2 * mu * abcd[3] / (abcd[0] + abcd[3]);

			}

		}

        /**
         * set context
         * @param context to set
         */
		void setContext(Mapper<IntWritable, RegionSlot, LongWritable, LongWritable>.Context context){
			this.context = context;
		}

        /**
         * generate edges in a given region.
         * @param region to process
         * @throws IOException by hadoop
         * @throws InterruptedException by hadoop
         */
		void generate(Region region) throws IOException, InterruptedException {

			if(region.numEdges < thresholdRegionEdgeNum && region.level < 32){

				long x = region.srcMin;
				long y = region.dstMin;

				long[] localEdges = new long[(int) region.numEdges];

				for(int i=0; i<region.numEdges; i++){
					localEdges[i] = makeOneLocalEdge(region);
				}

				Arrays.sort(localEdges);

				// emit unique edges.
				long prev = Long.MIN_VALUE;
				for (long edge : localEdges) {
					if (edge != prev) {
						ou.set(x + (edge >>> 32));
						ov.set(y + (edge & 0x00000000FFFFFFFFL));

						context.write(ou, ov);
						prev = edge;
					}
				}
			}
			else for(Region r : splitRegion(region)) generate(r);

		}

        /**
         * make an edge in a region without offset
         * @param region to process
         * @return an edge in the region without offset
         * @throws IOException by hadoop
         * @throws InterruptedException by hadoop
         */
		private long makeOneLocalEdge(Region region) throws IOException, InterruptedException {

			long u = 0;
			long v = 0;

			double[] t = probs[region.level];

			for(int l = region.level; l > 0; l--){
				double p = Math.random();
				if(p > t[0] + t[1] + t[2]){
					u += (1L << l);
					v += (1L << l);
				}
				else if (p > t[0] + t[1]) u += (1L << l);
				else if (p > t[0]) v += (1L << l);
			}

			return (u << 32) | v;

		}

        /**
         * split a region into 4 sub-regions.
         * @param region to process
         * @return 4 sub-regions.
         */
		Region[] splitRegion(Region region){

			double[] t = probs[region.level];

			long[] splitEdgeNums = multinomial(region.numEdges, Arrays.copyOf(t, 3));
			long lastSplitEdgeNum = region.numEdges - splitEdgeNums[0] - splitEdgeNums[1] - splitEdgeNums[2];

			long srcMid = (region.srcMin + region.srcMax) >> 1;
			long dstMid = (region.dstMin + region.dstMax) >> 1;

			long maxSplitNumNodes = (region.srcMax - srcMid) * (region.srcMax - srcMid);

			List<Region> res = new ArrayList<>();

			if (splitEdgeNums[0] > 0)
				res.add(new Region(region.srcMin, region.dstMin, srcMid, dstMid,
						region.level - 1, splitEdgeNums[0]));

			if (splitEdgeNums[1] > 0)
				res.add(new Region(region.srcMin, dstMid + 1, srcMid, region.dstMax,
						region.level - 1, splitEdgeNums[1]));

			if (splitEdgeNums[2] > 0)
				res.add(new Region(srcMid + 1, region.dstMin, region.srcMax, dstMid,
						region.level - 1, splitEdgeNums[2]));

			if (lastSplitEdgeNum > 0)
				res.add(new Region(srcMid + 1, dstMid + 1, region.srcMax, region.dstMax,
						region.level - 1, lastSplitEdgeNum));


			Region[] ret = new Region[res.size()];
			for(int i=0; i<ret.length; i++) ret[i] = res.get(i);

			return ret;
		}

        /**
         * get a random sample following a multinomial distribution.
         * @param n the size of sample (a.k.a. the number of trials).
         * @param p probabilities
         * @return a random sample.
         */
		private long[] multinomial(long n, double[] p) {

			double[][] nM = new double[p.length][p.length];
			double[] np = new double[p.length];

			for (int i = 0; i < np.length; i++) {
				np[i] = p[i] * n;
			}

			for(int i=0; i<p.length; i++){
				for(int j=0; j<p.length; j++){
					nM[i][j] = i == j ? p[i] : 0;
					nM[i][j] -= p[i] * p[j];
					nM[i][j] *= n;
				}
			}

			MultivariateNormalDistribution distribution = new MultivariateNormalDistribution(np, nM);

			boolean negativeExist;
			double[] _sample;

			do{
				negativeExist = false;
				_sample = distribution.sample();

				for(int i=0; i<_sample.length; i++)
					if(_sample[i] < 0) negativeExist = true;
			}while(negativeExist);

			long[] sample = new long[_sample.length];

			for(int i=0; i<sample.length; i++)
				sample[i] = Math.round(_sample[i]);

			return sample;
		}


	}

	static class Region implements Comparable<Region>, Writable {

		long srcMin;
		long dstMin;
		long srcMax;
		long dstMax;
		int level;
		long numEdges;

		public Region(){};

        /**
         * A region.
         * @param srcMin the minimum id of source nodes.
         * @param dstMin the minimum id of destination nodes.
         * @param srcMax the maximum id of source nodes.
         * @param dstMax the maximum id of destination nodes.
         * @param level the level
         * @param numEdges the number of edges
         */
		public Region(long srcMin, long dstMin, long srcMax, long dstMax, int level, long numEdges){
			this.srcMin = srcMin;
			this.dstMin = dstMin;
			this.srcMax = srcMax;
			this.dstMax = dstMax;
			this.level = level;
			this.numEdges = numEdges;
		}

		@Override
		@SuppressWarnings("NullableProblems")
		public int compareTo(Region that) {
			if(this.numEdges < that.numEdges) return 1;
			else if(this.numEdges > that.numEdges) return -1;
			else return 0;
		}

        /**
         * write this region out
         * @param out to write
         * @throws IOException by hadoop
         */
		@Override
		public void write(DataOutput out) throws IOException {
			out.writeLong(srcMin);
			out.writeLong(dstMin);
			out.writeLong(srcMax);
			out.writeLong(dstMax);
			out.writeInt(level);
			out.writeLong(numEdges);
		}

        /**
         * read a region
         * @param in read from it
         * @throws IOException by hadoop
         */
		@Override
		public void readFields(DataInput in) throws IOException {
			srcMin = in.readLong();
			dstMin = in.readLong();
			srcMax = in.readLong();
			dstMax = in.readLong();
			level = in.readInt();
			numEdges = in.readLong();
		}
	}

	static class RegionSlot implements Comparable<RegionSlot>, Writable{

		long edgeCapacity;
		List<Region> regions;

		public RegionSlot(){};

        /**
         * A container of regions.
         * @param edgeCapacity rough number of edges can accommodate. Real edge number can exceed this value.
         */
		public RegionSlot(long edgeCapacity){
			this.edgeCapacity = edgeCapacity;
			regions = new ArrayList<>();
		}

        /**
         * add a region
         * @param r region
         */
		public void add(Region r){
			regions.add(r);
			edgeCapacity -= r.numEdges;
		}

		@Override
		@SuppressWarnings("NullableProblems")
		public int compareTo(RegionSlot that) {
			if(this.edgeCapacity < that.edgeCapacity) return 1;
			else if(this.edgeCapacity > that.edgeCapacity) return -1;
			else return 0;
		}

        /**
         * write this region slot out
         * @param out to write
         * @throws IOException by hadoop
         */
		@Override
		public void write(DataOutput out) throws IOException {
			out.writeLong(edgeCapacity);
			out.writeInt(regions.size());

			for(Region r : regions) r.write(out);
		}

        /**
         * read a region slot
         * @param in read from it
         * @throws IOException by hadoop
         */
		@Override
		public void readFields(DataInput in) throws IOException {
			edgeCapacity = in.readLong();
			int numRegions = in.readInt();

			regions = new ArrayList<>();

			for(int i = 0; i < numRegions; i++){
				Region r = new Region();
				r.readFields(in);
				regions.add(r);
			}
		}
	}



}
