/**
 * PegasusN: Peta-Scale Graph Mining System (Pegasus v3.0)
 * Authors: Chiwan Park, Ha-Myung Park, U Kang
 *
 * This software is licensed under Apache License, Version 2.0 (the  "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 * -------------------------------------------------------------------------
 * File: RandomGenerator.java
 * - generate a random graph.
 * Version: 3.0
 */

package pegasus.mapreduce.graphgen;

import java.io.*;
import java.util.*;

import org.apache.commons.math3.distribution.MultivariateNormalDistribution;
import org.apache.hadoop.conf.Configuration;
import org.apache.hadoop.conf.Configured;
import org.apache.hadoop.fs.FileSystem;
import org.apache.hadoop.fs.Path;
import org.apache.hadoop.io.IntWritable;
import org.apache.hadoop.io.LongWritable;
import org.apache.hadoop.io.SequenceFile;
import org.apache.hadoop.io.Writable;
import org.apache.hadoop.mapreduce.Job;
import org.apache.hadoop.mapreduce.Mapper;
import org.apache.hadoop.mapreduce.lib.input.FileInputFormat;
import org.apache.hadoop.mapreduce.lib.input.SequenceFileInputFormat;
import org.apache.hadoop.mapreduce.lib.output.FileOutputFormat;
import org.apache.hadoop.mapreduce.lib.output.TextOutputFormat;
import org.apache.hadoop.util.Tool;
import org.apache.hadoop.util.ToolRunner;
import org.apache.log4j.Logger;

public class RandomGenerator extends Configured implements Tool {

	private static Logger logger = Logger.getLogger(RandomGenerator.class);

	long thresholdRegionEdgeNum = 10000;

    /**
     * The main entry point
     * @param args [0]: the number of nodes. [1]: the number of edges. [2]: the number of tasks to run.
     *             [3]: output file path
     * @throws Exception by hadoop
     */
	public static void main(String[] args) throws Exception {
	    ToolRunner.run(new RandomGenerator(), args);
    }

    /**
     * Submit the hadoop job of generating a random graph.
     * @param args [0]: the number of nodes. [1]: the number of edges. [2]: the number of tasks to run.
     *             [3]: output file path
     * @return not used
     * @throws Exception by hadoop
     */
    @Override
    public int run(String[] args) throws Exception {

        if(args.length < 4){
            System.err.println("parameters: numNodes numEdges numTasks outputPath");
            return -1;
        }

	    long numNodes = Long.parseLong(args[0]);
	    long numEdges = Long.parseLong(args[1]);
	    long numTasks = Integer.parseInt(args[2]);
	    Path outputPath = new Path(args[3]);

	    Path inputPath = outputPath.suffix(".input");

	    logger.info(String.format("numNodes: %d, numEdges: %d, numTasks: %d", numNodes, numEdges, numTasks));

        PriorityQueue<Region> regionQueue = new PriorityQueue<>();
        PriorityQueue<RegionSlot> regionSlotQueue = new PriorityQueue<>();

        for(int i = 0; i < numTasks; i++){
            long numEdgeAssigned = (numEdges * (i + 1) / numTasks) - (numEdges * i / numTasks);
            regionSlotQueue.add(new RegionSlot(numEdgeAssigned));
        }

        regionQueue.add(new Region(0, 0, numNodes - 1, numNodes - 1, numEdges));

        while(!regionQueue.isEmpty()){
            Region region = regionQueue.poll();
            RegionSlot slot = regionSlotQueue.poll();

            if(region.numEdges <= thresholdRegionEdgeNum || region.numEdges <= slot.edgeCapacity){
                slot.add(region);
            }
            else{
                regionQueue.addAll(Arrays.asList(splitRegion(region)));
            }

            regionSlotQueue.add(slot);
        }

        //delete previous input file.
        FileSystem.get(getConf()).delete(inputPath, true);

        generateInputFile(regionSlotQueue, inputPath);

        Configuration conf = getConf();

        Job job = Job.getInstance(conf, "[Random]" + outputPath);
        job.setJarByClass(getClass());
        job.setOutputKeyClass(LongWritable.class);
        job.setOutputValueClass(LongWritable.class);
        job.setInputFormatClass(SequenceFileInputFormat.class);
        job.setOutputFormatClass(TextOutputFormat.class);
        job.setMapperClass(RandomGeneratorMapper.class);
        job.setNumReduceTasks(0);

        FileInputFormat.addInputPath(job, inputPath);
        FileOutputFormat.setOutputPath(job, outputPath);

        job.waitForCompletion(false);

        return 0;
    }

    public static class RandomGeneratorMapper extends Mapper<IntWritable, RegionSlot, LongWritable, LongWritable> {

	    Context context;
        long thresholdRegionEdgeNum = 10000;

        LongWritable ou = new LongWritable();
        LongWritable ov = new LongWritable();

        /**
         * Setup before execution
         * @param context of hadoop
         * @throws IOException by hadoop
         * @throws InterruptedException by hadoop
         */
        @Override
        protected void setup(Context context) throws IOException, InterruptedException {
            this.context = context;
        }

        /**
         * The map function of the random graph generator.
         * @param key task id.
         * @param rs a region slot that contains regions to process here.
         * @param context of hadoop
         * @throws IOException by hadoop
         * @throws InterruptedException by hadoop
         */
        @Override
        protected void map(IntWritable key, RegionSlot rs, Context context) throws IOException, InterruptedException {

            for(Region r : rs.regions){
                generate(r);
            }
        }

        /**
         * generate edges in a given region.
         * @param region to process
         * @throws IOException by hadoop
         * @throws InterruptedException by hadoop
         */
        void generate(Region region) throws IOException, InterruptedException {

            long srcDiff = region.srcMax - region.srcMin + 1;
            long dstDiff = region.dstMax - region.dstMin + 1;

            long x = region.srcMin;
            long y = region.dstMin;

            if(region.numEdges < thresholdRegionEdgeNum &&
                    srcDiff < Integer.MAX_VALUE &&
                    dstDiff < Integer.MAX_VALUE) {

                long[] localEdges = new long[(int) region.numEdges];

                for (int i = 0; i < region.numEdges; i++) {
                    long u = (long) (Math.random() * srcDiff);
                    long v = (long) (Math.random() * dstDiff);

                    localEdges[i] = u << 32 | v;
                }

                Arrays.sort(localEdges);

                // emit unique edges.
                long prev = Long.MIN_VALUE;
                for (long edge : localEdges) {
                    if (edge != prev) {
                        ou.set(x + (edge >>> 32));
                        ov.set(y + (edge & 0x00000000FFFFFFFFL));

                        context.write(ou, ov);
                        prev = edge;
                    }
                }
            }
            else for(Region r : splitRegion(region)) generate(r);
        }
    }

    /**
     * Make the input file for the hadoop job of the random graph generator.
     * @param regionSlotQueue priority queue of region slots
     * @param input file path
     * @throws IOException by hadoop
     */
    private void generateInputFile(PriorityQueue<RegionSlot> regionSlotQueue, Path input) throws IOException {

        FileSystem fs = FileSystem.get(getConf());
        fs.delete(input, true);

        SequenceFile.Writer.Option optKey = SequenceFile.Writer.keyClass(IntWritable.class);
        SequenceFile.Writer.Option optValue = SequenceFile.Writer.valueClass(RegionSlot.class);

        int i = 0;
        IntWritable okey = new IntWritable();
        for(RegionSlot rs : regionSlotQueue){
            SequenceFile.Writer sw = SequenceFile.createWriter(new Configuration(),
                    SequenceFile.Writer.file(input.suffix("/part-" + i)),
                    optKey, optValue);

            okey.set(i++);
            sw.append(okey, rs);

            sw.close();
        }
    }

    /**
     * split a region into 4 sub-regions.
     * @param region to process
     * @return 4 sub-regions.
     */
    static Region[] splitRegion(Region region){

	    long[] splitEdgeNums = multinomial(region.numEdges, new double[]{0.25, 0.25, 0.25});
	    long lastSplitEdgeNum = region.numEdges - splitEdgeNums[0] - splitEdgeNums[1] - splitEdgeNums[2];

	    long srcMid = (region.srcMin + region.srcMax) >> 1;
	    long dstMid = (region.dstMin + region.dstMax) >> 1;

	    Region[] res = new Region[4];

	    res[0] = new Region(region.srcMin, region.dstMin, srcMid, dstMid, splitEdgeNums[0]);
        res[1] = new Region(region.srcMin, dstMid + 1, srcMid, region.dstMax, splitEdgeNums[1]);
        res[2] = new Region(srcMid + 1, region.dstMin, region.srcMax, dstMid, splitEdgeNums[2]);
        res[3] = new Region(srcMid + 1, dstMid + 1, region.srcMax, region.dstMax, lastSplitEdgeNum);

	    return res;
    }

    /**
     * get a random sample following a multinomial distribution.
     * @param n the size of sample (a.k.a. the number of trials).
     * @param p probabilities
     * @return a random sample.
     */
    private static long[] multinomial(long n, double[] p) {

        double[][] nM = new double[p.length][p.length];
        double[] np = new double[p.length];

        for (int i = 0; i < np.length; i++) {
            np[i] = p[i] * n;
        }

        for(int i=0; i<p.length; i++){
            for(int j=0; j<p.length; j++){
                nM[i][j] = i == j ? p[i] : 0;
                nM[i][j] -= p[i] * p[j];
                nM[i][j] *= n;
            }
        }

        MultivariateNormalDistribution distribution = new MultivariateNormalDistribution(np, nM);

        boolean negativeExist;
        double[] _sample;

        do{
            negativeExist = false;
            _sample = distribution.sample();

            for(int i=0; i<_sample.length; i++)
                if(_sample[i] < 0) negativeExist = true;
        }while(negativeExist);

        long[] sample = new long[_sample.length];

        for(int i=0; i<sample.length; i++)
            sample[i] = Math.round(_sample[i]);

        return sample;
    }

    static class Region implements Comparable<Region>, Writable {

        long srcMin;
        long dstMin;
        long srcMax;
        long dstMax;
        long numEdges;

        public Region(){};

        /**
         * A region.
         * @param srcMin the minimum id of source nodes.
         * @param dstMin the minimum id of destination nodes.
         * @param srcMax the maximum id of source nodes.
         * @param dstMax the maximum id of destination nodes.
         * @param numEdges the number of edges.
         */
        public Region(long srcMin, long dstMin, long srcMax, long dstMax, long numEdges){
            this.srcMin = srcMin;
            this.dstMin = dstMin;
            this.srcMax = srcMax;
            this.dstMax = dstMax;
            this.numEdges = numEdges;
        }

        @Override
        @SuppressWarnings("NullableProblems")
        public int compareTo(Region that) {
            if(this.numEdges < that.numEdges) return 1;
            else if(this.numEdges > that.numEdges) return -1;
            else return 0;
        }

        /**
         * write this region out
         * @param out to write
         * @throws IOException by hadoop
         */
        @Override
        public void write(DataOutput out) throws IOException {
            out.writeLong(srcMin);
            out.writeLong(dstMin);
            out.writeLong(srcMax);
            out.writeLong(dstMax);
            out.writeLong(numEdges);
        }

        /**
         * read a region
         * @param in read from it
         * @throws IOException by hadoop
         */
        @Override
        public void readFields(DataInput in) throws IOException {
            srcMin = in.readLong();
            dstMin = in.readLong();
            srcMax = in.readLong();
            dstMax = in.readLong();
            numEdges = in.readLong();
        }
    }

    static class RegionSlot implements Comparable<RegionSlot>, Writable{

        long edgeCapacity;
        List<Region> regions;

        public RegionSlot(){};

        /**
         * A container of regions.
         * @param edgeCapacity rough number of edges can accommodate. Real edge number can exceed this value.
         */
        public RegionSlot(long edgeCapacity){
            this.edgeCapacity = edgeCapacity;
            regions = new ArrayList<>();
        }

        /**
         * add a region
         * @param r region
         */
        public void add(Region r){
            regions.add(r);
            edgeCapacity -= r.numEdges;
        }

        @Override
        @SuppressWarnings("NullableProblems")
        public int compareTo(RegionSlot that) {
            if(this.edgeCapacity < that.edgeCapacity) return 1;
            else if(this.edgeCapacity > that.edgeCapacity) return -1;
            else return 0;
        }

        /**
         * write this region slot out
         * @param out to write
         * @throws IOException by hadoop
         */
        @Override
        public void write(DataOutput out) throws IOException {
            out.writeLong(edgeCapacity);
            out.writeInt(regions.size());

            for(Region r : regions) r.write(out);
        }

        /**
         * read a region slot
         * @param in read from it
         * @throws IOException by hadoop
         */
        @Override
        public void readFields(DataInput in) throws IOException {
            edgeCapacity = in.readLong();
            int numRegions = in.readInt();

            regions = new ArrayList<>();

            for(int i = 0; i < numRegions; i++){
                Region r = new Region();
                r.readFields(in);
                regions.add(r);
            }
        }
    }
}
