/*
 * PegasusG: Peta-Scale Graph Mining System v3.0
 * Authors: Chiwan Park, Ha-Myung Park, U Kang
 *
 * This software is licensed under Apache License, Version 2.0 (the  "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 * -------------------------------------------------------------------------
 * File: NetRaySpy.java
 * - Draw a spy plot of the given matrix
 * Version: 3.0
 */
package pegasus.mapreduce.netray;

import org.apache.hadoop.conf.Configuration;
import org.apache.hadoop.conf.Configured;
import org.apache.hadoop.fs.FileSystem;
import org.apache.hadoop.fs.Path;
import org.apache.hadoop.io.LongWritable;
import org.apache.hadoop.io.Text;
import org.apache.hadoop.mapreduce.Job;
import org.apache.hadoop.mapreduce.Mapper;
import org.apache.hadoop.mapreduce.Reducer;
import org.apache.hadoop.mapreduce.lib.input.FileInputFormat;
import org.apache.hadoop.mapreduce.lib.output.FileOutputFormat;
import org.apache.hadoop.util.Tool;
import org.apache.hadoop.util.ToolRunner;

import java.io.IOException;

public class NetRaySpy extends Configured implements Tool {
  // Digitize data
  public static class MapStage1 extends Mapper<LongWritable, Text, Text, LongWritable> {
    double s_over_n = 0;
    double sm1_over_logn = 0;
    int is_linear = 0;      // linear or log scale

    @Override
    protected void setup(Context context) throws IOException, InterruptedException {
      final Configuration conf = context.getConfiguration();
      s_over_n = conf.getDouble("s_over_n", Double.MAX_VALUE);
      sm1_over_logn = conf.getDouble("sm1_over_logn", Double.MAX_VALUE);
      is_linear = conf.getInt("is_linear", 0);

      System.out.println("MapStage1: s_over_n = " + s_over_n + ", s_over_n = " + s_over_n + ", is_linear = " + is_linear);
    }


    @Override
    protected void map(final LongWritable key, final Text value, final Context context)
        throws IOException, InterruptedException {
      final String line_text = value.toString();
      if (line_text.startsWith("#")) return;

      final String[] line = line_text.split("\t");
      if (line.length < 2) return;

      final int row_id = Integer.parseInt(line[0]);
      final int col_id = Integer.parseInt(line[1]);
      int block_rowid = 0, block_colid = 0;

      if (is_linear == 1) {  // linear scale
        block_rowid = (int) Math.ceil((double) (row_id + 1) * s_over_n);
        block_colid = (int) Math.ceil((double) (col_id + 1) * s_over_n);
      } else {      // log scale
        block_rowid = (int) Math.ceil(Math.log10(row_id + 1) * sm1_over_logn + 0.5);
        block_colid = (int) Math.ceil(Math.log10(col_id + 1) * sm1_over_logn + 0.5);
      }

      long weight = 1;
      if (line.length >= 3) {
        weight = Long.parseLong(line[2]);
      }

      context.write(new Text("" + block_rowid + "\t" + block_colid), new LongWritable(weight));

    }
  }

  public static class RedStage1 extends Reducer<Text, LongWritable, Text, LongWritable> {
    @Override
    protected void reduce(Text key, Iterable<LongWritable> values, Context context) throws IOException, InterruptedException {
      long sum = 0;
      for (LongWritable value : values) {
        sum += value.get();
      }

      context.write(key, new LongWritable(sum));
    }
  }

  //////////////////////////////////////////////////////////////////////
  // command line interface
  //////////////////////////////////////////////////////////////////////
  protected Path edge_path = null;
  protected Path output_path = null;
  protected int orig_nnodes = 0;
  protected int target_nnodes = 0;
  protected double s_over_n = 0;
  protected double sm1_over_logn = 0;

  protected int nreducer = 1;
  protected String output_prefix;
  protected int makesym = 0;
  protected int is_linear = 1;

  // Main entry point.
  public static void main(final String[] args) throws Exception {
    final int result = ToolRunner.run(new Configuration(), new NetRaySpy(), args);
    System.exit(result);
  }

  // Print the command-line usage text.
  protected static int printUsage() {
    System.out.println("NetRaySpy <edge_path> <outputedge_path> <original number of nodes> <target number of nodes> <lin or log> <# of reducer> ");

    ToolRunner.printGenericCommandUsage(System.out);

    return -1;
  }

  // submit the map/reduce job.
  public int run(final String[] args) throws Exception {
    if (args.length != 6) {
      return printUsage();
    }

    edge_path = new Path(args[0]);        // input edge HDFS path
    output_path = new Path(args[1]);      // output HDFS path
    orig_nnodes = Integer.parseInt(args[2]);  // original matrix width
    target_nnodes = Integer.parseInt(args[3]);  // target matrix width
    if (args[4].startsWith("log"))
      is_linear = 0;
    nreducer = Integer.parseInt(args[5]);

    s_over_n = (double) (target_nnodes) / (double) (orig_nnodes);
    sm1_over_logn = (double) (target_nnodes - 1) / (double) Math.log10(orig_nnodes);

    System.out.println("\n-----===[PEGASUS: A Peta-Scale Graph Mining System]===-----\n");
    System.out.println("[PEGASUS] Drawing a spy plot. original nodes = " + orig_nnodes + ", target nnodes = " + target_nnodes + ", s_over_n = " + s_over_n + ", is_linear = " + is_linear + "\n");

    // run job
    final Job stage1Job = configStage1();
    stage1Job.waitForCompletion(true);

    System.out.println("\n[PEGASUS] Spyplot data generated.");
    System.out.println("[PEGASUS] Output is saved in the HDFS " + args[1] + "\n");

    return 0;
  }

  // Configure pass1
  protected Job configStage1() throws Exception {
    final Job job = Job.getInstance(getConf(), "NetRaySpy_Stage1");
    final Configuration conf = job.getConfiguration();
    conf.setDouble("s_over_n", s_over_n);
    conf.setDouble("sm1_over_logn", sm1_over_logn);
    conf.setInt("is_linear", is_linear);

    job.setJarByClass(NetRaySpy.class);

    job.setMapperClass(MapStage1.class);
    job.setReducerClass(RedStage1.class);
    job.setCombinerClass(RedStage1.class);

    final FileSystem fs = FileSystem.get(getConf());
    fs.delete(output_path, true);

    FileInputFormat.setInputPaths(job, edge_path);
    FileOutputFormat.setOutputPath(job, output_path);

    int num_reduce_tasks = nreducer;
    job.setNumReduceTasks(num_reduce_tasks);

    job.setOutputKeyClass(Text.class);
    job.setOutputValueClass(LongWritable.class);

    return job;
  }
}
