/*
 * PegasusN: Peta-Scale Graph Mining System (Pegasus v3.0)
 * Authors: Chiwan Park, Ha-Myung Park, U Kang
 *
 * This software is licensed under Apache License, Version 2.0 (the  "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 * -------------------------------------------------------------------------
 * File: PSE.java
 * - PSE for subgraph enumeration
 * Version: 3.0
 */

package pegasus.mapreduce.subgraph;

import java.io.FileNotFoundException;
import java.io.IOException;
import java.lang.management.ManagementFactory;
import java.util.Arrays;
import java.util.Collections;
import java.util.Comparator;
import java.util.List;
import java.util.StringTokenizer;

import org.apache.hadoop.conf.Configuration;
import org.apache.hadoop.conf.Configured;
import org.apache.hadoop.fs.FSDataOutputStream;
import org.apache.hadoop.fs.FileSystem;
import org.apache.hadoop.fs.Path;
import org.apache.hadoop.io.NullWritable;
import org.apache.hadoop.io.Text;
import org.apache.hadoop.mapreduce.Job;
import org.apache.hadoop.mapreduce.Mapper;
import org.apache.hadoop.mapreduce.lib.input.NLineInputFormat;
import org.apache.hadoop.mapreduce.lib.output.FileOutputFormat;
import org.apache.hadoop.mapreduce.lib.output.NullOutputFormat;
import org.apache.hadoop.mapreduce.lib.output.TextOutputFormat;
import org.apache.hadoop.util.Tool;
import org.apache.hadoop.util.ToolRunner;
import org.apache.log4j.Logger;
import org.xerial.snappy.SnappyInputStream;

import com.esotericsoftware.kryo.KryoException;
import com.esotericsoftware.kryo.io.Input;

public class PSE extends Configured implements Tool{

	long SHUFFLED_BYTES = 0;
	long HDFS_BYTES_READ = 0;
	long NUM_SUBGRAPHS = 0;

	/**
	 * The main entry point
	 * @param args [0]: input file path. [1]: patternString. [2]: output file path.
	 *             Tool runner arguments:
	 *          -D numColors the number of colors
	 * @throws Exception by hadoop
	 */
	public static void main(String args[]) throws Exception{
		ToolRunner.run(new PSE(), args);
	}

	/**
	 * Submit the hadoop job of PSE
	 * @param args [0]: input file path. [1]: patternString. [2]: output file path.
	 *             Tool runner arguments:
	 *          -D numColors the number of colors
	 * @return not used
	 * @throws Exception by hadoop
	 */
	public int run(String[] args) throws Exception{
		
		long pt, et;

		Configuration conf = getConf();
		int numColors = conf.getInt("numColors", 0);

		if(args.length < 2 || numColors == 0){
			System.out.println("parameters: [inputFilePath] [patternString] " +
					"[outputFilePath (counting mode if not given)] -DnumColors [numColors]");
			return 0;
		}

		String input = args[0];
		String patternString = args[1];
		String output = null;
		if(args.length < 3) System.err.println("counting mode.");
		else{
			output = args[2];
			System.err.println("listing mode. The output file is stored in " + output);
		}

		int patternSize; for (patternSize = 1; patternSize*(patternSize-1)/2 < patternString.length(); patternSize++);
		
		String part = input + ".part";
		String seed = input + ".seed";
		
		pt = System.currentTimeMillis();
		
		PrePartition pp = new PrePartition();
		pp.setConf(getConf());
		pp.partitionGraph(input, part, numColors);
		int nodeSize = pp.nodeSize;
		System.err.println("nodeSize: " + nodeSize);
		
		pt = (System.currentTimeMillis() - pt)/1000;

		createSeedFile(seed, numColors, patternSize);
		
		et = System.currentTimeMillis();
		enumerateSubgraphs(numColors, nodeSize, patternString, part, seed, output);
		et = (System.currentTimeMillis() - et)/1000;
		
		System.out.println("partitionTime\tEnumerationTime\tShuffle\tNetwork\tTriangles");
		System.out.println("[" + this.getClass().getSimpleName() + "]" + input + "\t" + pt + "\t"
				+ et + "\t" + SHUFFLED_BYTES + "\t" + HDFS_BYTES_READ + "\t" + NUM_SUBGRAPHS);
//		
		FileSystem.get(getConf()).delete(new Path(part), true);
		FileSystem.get(getConf()).delete(new Path(seed), true);
//		
		return 0;
	}

	/**
	 * Find subgraphs that match a query graph
	 * @param numColors the number of colors
	 * @param nodeSize the number of nodes in the input graph
	 * @param patternString the binary string of a query graph
	 * @param part file path
	 * @param seed file path
	 * @param output file path
	 * @throws IOException by hadoop
	 * @throws InterruptedException by hadoop
	 * @throws ClassNotFoundException by hadoop
	 */
	private void enumerateSubgraphs(int numColors, int nodeSize, String patternString,
									String part, String seed, String output) throws IOException,
			InterruptedException, ClassNotFoundException{

		Job job;

		if(output != null){
			job = Job.getInstance(getConf(), "[" + this.getClass().getSimpleName() + "_listing]" + seed);
			job.setOutputFormatClass(TextOutputFormat.class);
			job.setOutputKeyClass(Text.class);
			job.setOutputValueClass(NullWritable.class);
			FileOutputFormat.setOutputPath(job, new Path(output));
			job.getConfiguration().setBoolean("listingMode", true);
		}
		else{
			job = Job.getInstance(getConf(), "[" + this.getClass().getSimpleName() + "_counting]" + seed);
			job.setOutputFormatClass(NullOutputFormat.class);
		}

		Configuration conf = job.getConfiguration();
		conf.set("mapreduce.map.java.opts", conf.get("mapreduce.map.java.opts", "") +
				" -XX:+UseG1GC -XX:+UseLargePagesInMetaspace");

		job.setMapperClass(SubgraphEnumerationMapper.class);
		job.setJarByClass(getClass());
		job.setNumReduceTasks(0);
		conf.setInt("mapred.line.input.format.linespermap", 1);
		conf.set("part", part);
		conf.setInt("numColors", numColors);
		conf.setInt("nodeSize", nodeSize);
		conf.set("patternString", patternString);
		job.setInputFormatClass(NLineInputFormat.class);
		NLineInputFormat.setInputPaths(job, new Path(seed));

		job.getConfiguration().setLong("mapred.task.timeout", 0L);
		job.getConfiguration().setLong("mapred.min.split.size", Long.MAX_VALUE);
		
		job.waitForCompletion(false);
		
		HDFS_BYTES_READ = job.getCounters().findCounter("FileSystemCounters", "HDFS_BYTES_READ").getValue();
		NUM_SUBGRAPHS = job.getCounters().findCounter("PTE", "subgraphs").getValue();

	}

	/**
	 * create seed file for subgraph enumeration of PSE
	 * @param seedFile seed file path
	 * @param p the number of colors
	 * @param n the number of pattern nodes
	 * @throws IOException by hadoop
	 */
	private void createSeedFile(String seedFile, int p, int n) throws IOException{

		List<String> seedList = SeedGenerator.dual(p, n);
		
		FSDataOutputStream seedOut = FileSystem.get(getConf()).create(new Path(seedFile));
		
		Collections.shuffle(seedList);
		
		System.out.printf("[Info] %d subproblem(s) to process\n", seedList.size());
		for(String seedString : seedList){
			seedOut.writeBytes(seedString + "\n");
		}
		
		seedOut.close();
	}

	public static class SubgraphEnumerationMapper extends Mapper<Object, Text, Text, NullWritable>{

		private final Logger LOGGER = Logger.getLogger(this.getClass().getName());
		
		int numColors;
		int nodeSize_edgeSet;
		Configuration conf;
		FileSystem fs;
		Path part;
		
		String patternString;
		int patternSize;

		boolean listingMode = false;

		/**
		 * setup before execution.
		 * @param context of hadoop
		 * @throws IOException by hadoop
		 * @throws InterruptedException by hadoop
		 */
		@Override
		protected void setup(Context context)
				throws IOException, InterruptedException{
			conf = context.getConfiguration();
			numColors = conf.getInt("numColors", 0);
			int nodeSize = context.getConfiguration().getInt("nodeSize", 0);
			nodeSize_edgeSet = nodeSize / numColors + (nodeSize % numColors == 0 ? 0 : 1);
			patternString = context.getConfiguration().get("patternString");
			for (patternSize = 1; patternSize*(patternSize-1)/2 < patternString.length(); patternSize++);
			
			fs = FileSystem.get(conf);
			part = new Path(conf.get("part"));

			listingMode = conf.getBoolean("listingMode", false);

			ManagementFactory.getGarbageCollectorMXBeans().forEach(x -> LOGGER.info("garbage collector: " + x.getName()));
			
		}

		/**
		 * The map function of PSE
		 * @param key meaning less
		 * @param value problem
		 * @param context of hadoop
		 * @throws IOException by hadoop
		 * @throws InterruptedException by hadoop
		 */
		@Override
		protected void map(Object key, Text value,
				Context context)
				throws IOException, InterruptedException{

			StringTokenizer st = new StringTokenizer(value.toString());

			int numProblems = st.countTokens();
			long[] problems = new long[numProblems];

			for(int i = 0; i < numProblems; i++){
				problems[i] = Long.parseLong(st.nextToken(), 2);
			}

			long problemGroup = problems[0];


			int[] colors = new int[Long.bitCount(problemGroup)];
			for(int i = 0, j = 0; i < numColors; i++)
				if((problemGroup & (1 << i)) != 0) colors[j++] = i;

			LOGGER.info("Colors of this problem:" + Arrays.toString(colors));

			int nodeSize_problemGroup = nodeSize_edgeSet * colors.length;
			int[][] adj = new int[nodeSize_problemGroup][];

			for(int i = 0; i < colors.length; i++){
				for(int j = 0; j < colors.length; j++){
					if(i == j && colors.length == patternSize) continue;

					getAdj(part.suffix("/edgeset-" + (colors[i] * numColors + colors[j])), adj, i, j, colors.length);

					LOGGER.info(String.format("edgeset-(%d, %d) is loaded (free-mem: %fMB)", i, j, availableMemoryBytes(false)/1000000.0));

				}
			}

			LOGGER.info("Data loading complete.");


			for(int i=0; i<adj.length; i++){

				if(adj[i] == null) continue;

				System.arraycopy(Arrays.stream(adj[i]).boxed().sorted(new Comparator<Integer>() {
					@Override
					public int compare(Integer o1, Integer o2) {

						long x = ((long) adj[o1].length << 32) | o1.intValue();
						long y = ((long) adj[o2].length << 32) | o2.intValue();

						return Long.compare(x, y);
					}}).mapToInt(x -> x).toArray(), 0, adj[i], 0, adj[i].length);
			}

			LOGGER.info(String.format("Adj sorting complete. (free-mem %.2fMB)", availableMemoryBytes(true)/1000000.0));


			long[] filter = Arrays.stream(problems).map(problem -> {

				long colorbit = 0;
				for(int i=0; i<colors.length; i++){
					colorbit |= ((problem >>> colors[i]) & 1) << i;
				}

				return colorbit;
			}).toArray();
			
			
			VF2_degree vf2 = new VF2_degree();

			if(listingMode) vf2.setContext(context);

			long count = vf2.countSubgraphs(adj, patternString, filter, colors, numColors);

			LOGGER.info("Subgraph enumerate complete.");
			context.getCounter("PTE", "subgraphs").increment(count);
			
		}

		/**
		 * load an edge set and attach it to 'adj', an adjacency list.
		 * @param graphPath graph file path
		 * @param adj loaded adj is attached here
		 * @param localColor_src source node color in this local problem
		 * @param localColor_dst destination node color in this local problem
		 * @param numLocalColors the number of colors in this local problem
		 * @throws IOException by hadoop
		 */
		private void getAdj(Path graphPath, int[][] adj, int localColor_src, int localColor_dst, int numLocalColors) throws IOException{
			long t = System.currentTimeMillis();
			
			try{
				Input in = new Input(new SnappyInputStream(fs.open(graphPath).getWrappedStream()));
				
				
				try{
					while(true){
						int u_local = in.readInt(true);
						int uSize = in.readInt(true);
						int u = u_local * numLocalColors + localColor_src;
						
						int[] adj_u;
						int length_prev = 0;
						
						if(adj[u] != null){
							length_prev = adj[u].length;
							adj_u = new int[length_prev + uSize];
							System.arraycopy(adj[u], 0, adj_u, 0, length_prev);
						}
						else{
							adj_u = new int[uSize];
						}
						
						adj[u] = adj_u;
						
						for(int i=length_prev; i<adj_u.length; i++){
							adj_u[i] = in.readInt(true) * numLocalColors + localColor_dst;
						}
					}
				}catch(KryoException e){}
				
				in.close();
				
			} catch(FileNotFoundException e){}
		}
		
		public static long availableMemoryBytes(boolean gc){
			
			if(gc){
				System.gc(); System.gc();
			}
			
			return Runtime.getRuntime().maxMemory()
			- (Runtime.getRuntime().totalMemory() - Runtime.getRuntime().freeMemory());
		}

	}
	
}
