/*
 * PegasusN: Peta-Scale Graph Mining System (Pegasus v3.0)
 * Authors: Chiwan Park, Ha-Myung Park, U Kang
 *
 * This software is licensed under Apache License, Version 2.0 (the  "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 * -------------------------------------------------------------------------
 * File: PrePartition.java
 * - Prepartitioning step of PSE
 * - partition the graph according to the number of colors
 * Version: 3.0
 */

package pegasus.mapreduce.subgraph;

import java.io.IOException;
import java.util.StringTokenizer;

import pegasus.mapreduce.common.io.LocalPaths;
import org.apache.hadoop.conf.Configuration;
import org.apache.hadoop.conf.Configured;
import org.apache.hadoop.fs.FileSystem;
import org.apache.hadoop.fs.Path;
import org.apache.hadoop.io.LongWritable;
import org.apache.hadoop.io.NullWritable;
import org.apache.hadoop.io.ShortWritable;
import org.apache.hadoop.io.Text;
import org.apache.hadoop.mapreduce.Job;
import org.apache.hadoop.mapreduce.Mapper;
import org.apache.hadoop.mapreduce.Reducer;
import org.apache.hadoop.mapreduce.lib.input.FileInputFormat;
import org.apache.hadoop.mapreduce.lib.output.FileOutputFormat;
import org.apache.hadoop.mapreduce.lib.output.NullOutputFormat;
import org.apache.hadoop.util.Tool;
import org.apache.hadoop.util.ToolRunner;
import org.xerial.snappy.SnappyOutputStream;

import com.esotericsoftware.kryo.io.Output;

import gnu.trove.list.array.TLongArrayList;

public class PrePartition extends Configured implements Tool{

	public int nodeSize;

	/**
	 * The main entry point
	 * @param args [0]: input file path. [1]: output file path. [2]: the number of colors.
	 * @throws Exception by hadoop
	 */
	public static void main(String args[]) throws Exception{
		ToolRunner.run(new PrePartition(), args);
	}

	/**
	 * submit the hadoop job for pre-partitioning
	 * @param args [0]: input file path. [1]: output file path. [2]: the number of colors.
	 * @return not used
	 * @throws Exception by hadoop
	 */
	public int run(String[] args) throws Exception{
		
		if(args.length < 2){
			System.out.println("parameters: inputFilePath outputFilePath numColors");
			return 0;
		}
		
		String input = args[0];
		String output = args[1];
		
		int numColors = Integer.parseInt(args[2]);
		
		partitionGraph(input, output, numColors);
		
		return 0;
	}


	/**
	 * partition the graph
	 * @param input file path
	 * @param output file path
	 * @param numColors the number of colors
	 * @throws IOException by hadoop
	 * @throws InterruptedException by hadoop
	 * @throws ClassNotFoundException by hadoop
	 */
	void partitionGraph(String input, String output, int numColors)
			throws IOException, InterruptedException, ClassNotFoundException {

		FileSystem.get(getConf()).delete(new Path(output), true);

		Job job = Job.getInstance(getConf(), "[" + this.getClass().getSimpleName() + "_partitionGraph]" + input);
		Configuration conf = job.getConfiguration();
		conf.setInt("numColors", numColors);
		
		job.setJarByClass(this.getClass());
		job.setMapperClass(PGMapper.class);
		job.setReducerClass(PGReducer.class);
		job.setNumReduceTasks(numColors * numColors);
		job.setMapOutputKeyClass(ShortWritable.class);
		job.setMapOutputValueClass(LongWritable.class);
		job.setOutputFormatClass(NullOutputFormat.class);
		FileInputFormat.setInputPaths(job, new Path(input));
		FileOutputFormat.setOutputPath(job, new Path(output));

		job.waitForCompletion(false);
		
		nodeSize = (int) job.getCounters().findCounter("PTE", "nodeSize").getValue();
	}
	
	
	public static class PGMapper extends Mapper<Object, Text, ShortWritable, LongWritable>{

		private short numColors;
		
		ShortWritable ok = new ShortWritable();
		LongWritable ov = new LongWritable();
		
		private int maxNodeId = 0;

		/**
		 * setup before execution.
		 * set the number of colors.
		 * @param context of hadoop
		 * @throws IOException by hadoop
		 * @throws InterruptedException by hadoop
		 */
		@Override
		protected void setup(Context context)
				throws IOException, InterruptedException{

			numColors = (short) context.getConfiguration().getInt("numColors", 0);

		}

		/**
		 * The map function of pre-partitioning
		 * @param key not used
		 * @param value an edge as a text
		 * @param context of hadoop
		 * @throws IOException by hadoop
		 * @throws InterruptedException by hadoop
		 */
		@Override
		protected void map(Object key, Text value,
				Context context)
				throws IOException, InterruptedException{

			StringTokenizer st = new StringTokenizer(value.toString());
			int u = Integer.parseInt(st.nextToken());
			int v = Integer.parseInt(st.nextToken());
			
			if(u == v) return;
			
			int u_local = u / numColors;
			int v_local = v / numColors;
			short u_color = (short) (u - u_local * numColors);
			short v_color = (short) (v - v_local * numColors);
			
			ok.set((short) ((u_color * numColors) + (v_color)));
			ov.set(((long) (u_local) << 32) | v_local);
			context.write(ok, ov);
			
			ok.set((short) ((v_color * numColors) + (u_color)));
			ov.set(((long) (v_local) << 32) | u_local);
			context.write(ok, ov);
			
			maxNodeId = maxNodeId < u ? u : maxNodeId;
			maxNodeId = maxNodeId < v ? v : maxNodeId;
			
			
		}

		/**
		 * Cleanup after execution.
		 * To find the global maximum vertex id, it outputs maximum vertex id in each partition.
		 * @param context of hadoop
		 * @throws IOException by hadoop
		 * @throws InterruptedException by hadoop
		 */
		@Override
		protected void cleanup(Context context)
				throws IOException, InterruptedException {
			
			context.write(new ShortWritable((short)-1), new LongWritable(maxNodeId));
			
		}
		
		

	}
	
	public static class PGReducer extends Reducer<ShortWritable, LongWritable, NullWritable, NullWritable>{

		private Path outputPath;

		/**
		 * setup before execution.
		 * @param context of hadoop
		 * @throws IOException by hadoop
		 * @throws InterruptedException by hadoop
		 */
		@Override
		protected void setup(Context context)
				throws IOException, InterruptedException {
			
			String[] tmpDirs = context.getConfiguration().getTrimmedStrings("yarn.nodemanager.local-dirs");
			
			outputPath = FileOutputFormat.getOutputPath(context);
			
			LocalPaths.setBasePaths(tmpDirs);
			
		}

		/**
		 * The reduce function of pre-partitioning
		 * @param edgeColor the color of edges (u, v). (color(u) * numColors + color(v))
		 * @param edges edges of the edge color
		 * @param context of hadoop
		 * @throws IOException by hadoop
		 * @throws InterruptedException by hadoop
		 */
		@Override
		protected void reduce(ShortWritable edgeColor, Iterable<LongWritable> edges,
				Context context)
				throws IOException, InterruptedException {
			
			
			//find nodeSize. local maximum values are in `edges`.
			if(edgeColor.get() == -1){
				long maxNodeId = 0;
				
				for(LongWritable _value : edges){
					long value = _value.get();
					maxNodeId = maxNodeId < value ? value : maxNodeId;
				}
				
				int numColors = context.getConfiguration().getInt("numColors", 0);
				
				maxNodeId = 1 + maxNodeId;
				
				context.getCounter("PTE", "nodeSize").setValue(maxNodeId);
				return;
			}
			
			int maxNodeId = 0;
			
			TLongArrayList _edgeList = new TLongArrayList();
			
			for(LongWritable _edge : edges){
				_edgeList.add( _edge.get());
			}
			
			_edgeList.sort();
			long[] edgeList = _edgeList.toArray();
			
			FileSystem fs = FileSystem.get(context.getConfiguration());
			Output out = new Output(new SnappyOutputStream(fs.create(outputPath.suffix("/edgeset-" + edgeColor)).getWrappedStream()));
			
			int edgeLen = edgeList.length;
			for(int uStart = 0, uEnd = 0; uStart < edgeLen; uStart = uEnd){
				
				long edge = edgeList[uStart];
				int u = (int) (edge >>> 32);
				long limit = (long) (u + 1) << 32;
				
				maxNodeId = maxNodeId < u ? u : maxNodeId; 
				
				int count_uniq = 0;
				try{
					long prev = -1;
					long cur = 0;
					for(;(cur = edgeList[uEnd]) < limit ; uEnd++){
						if(prev != cur){
							count_uniq++;
							prev = cur;
						}
					}
				}catch(ArrayIndexOutOfBoundsException e){}
				
				out.writeInt(u, true);
				out.writeInt(count_uniq, true);
				
				long prev = -1;
				long cur = 0;
				for(int i=uStart; i<uEnd; i++){
					cur = edgeList[i];
					if(prev != cur){
						out.writeInt((int) cur, true);
						prev = cur;
					}
				}
			}
			
			out.close();
			
		}

	}
	
}
