/**
  * PegasusN: Peta-Scale Graph Mining System (Pegasus v3.0)
  * Authors: Chiwan Park, Ha-Myung Park, U Kang
  *
  * This software is licensed under Apache License, Version 2.0 (the  "License");
  * you may not use this file except in compliance with the License.
  * You may obtain a copy of the License at
  *
  * http://www.apache.org/licenses/LICENSE-2.0
  *
  * Unless required by applicable law or agreed to in writing, software
  * distributed under the License is distributed on an "AS IS" BASIS,
  * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  * See the License for the specific language governing permissions and
  * limitations under the License.
  * -------------------------------------------------------------------------
  * File: PACCExternal.java
  * - Find Connected components.
  * - This version minimizes memory usage by an external sort to avoid
  *   an out-of-memory error.
  * Version: 3.0
  */

package pegasus.spark.cc

import org.apache.spark.rdd.RDD
import org.apache.spark.storage.StorageLevel
import pegasus.spark.cc.utils.FilteringOps._
import pegasus.spark.cc.utils.StarGroupOps._
import pegasus.spark.cc.utils.{LongExternalSorter, SerializableConfiguration}

class PACCExternal extends PACC{

  override val APP_NAME: String = "pacc-external"
  override val EXEC_NAME: String = "paccext"

  /**
    * PA-Large-Star Operation.
    * For each node n, this operation links each large neighbor v to the minimum node mcu(p(v))
    * in the same partition p(v) that contains the neighbor v.
    *
    * @param inputRDD the input rdd
    * @param numPartitions the number of partitions
    * @param round current round number
    * @param tmpPath the temporary path to save the intermediate results
    * @return (RDD for next round input, # changed edges, # of 'out' edges,
    *         # filtered 'cc' edges, # filtered 'in' edges)
    */
  override def largeStar(inputRDD: RDD[(Long, Long)], numPartitions: Int,
                round: Int, tmpPath: String): (RDD[(Long, Long)], Long, Long, Long, Long) = {

    val sc = inputRDD.sparkContext

    val NUM_CHANGES = sc.longAccumulator
    val LCC_SIZE = sc.longAccumulator
    val LIN_SIZE = sc.longAccumulator
    val LOUT_SIZE = sc.longAccumulator

    val groupedRDD = inputRDD.flatMap{
      case (u, v) =>
        if (u < 0) Seq((v, u))
        else Seq((u, v), (v, u))
    }.starGrouped()

    val tmpPaths = sc.hadoopConfiguration.getTrimmedStrings("yarn.nodemanager.local-dirs")

    val res_all = groupedRDD.mapPartitions{ it =>

      val longExternalSorter = new LongExternalSorter(tmpPaths)

      def processNode(x: (Long, Iterator[Long])): Iterator[(Boolean, Long, Long)] ={
        val (u, uN) = x

        val mcu = Array.fill[Long](numPartitions)(Long.MaxValue)
        mcu((u % numPartitions).toInt) = u

        var isStar = true

        val _uN_large = uN.filter { v_raw =>
          val v = if (v_raw >= 0) {
            isStar = false
            v_raw
          } else ~v_raw

          val vp = (v % numPartitions).toInt
          mcu(vp) = Math.min(v, mcu(vp))

          v > u
        }

        val uN_large = longExternalSorter.sort(_uN_large)

        val mu = mcu.min

        if (isStar) uN_large.map { v_raw => LCC_SIZE.add(1); (false, ~v_raw, u) }
        else{
          uN_large.map{ v_raw =>
            val vIsLeaf = v_raw < 0
            val v: Long = if(vIsLeaf) ~v_raw else v_raw

            val vp = (v % numPartitions).toInt
            val mcu_vp = mcu(vp)

            if(v != mcu_vp) {

              if(mcu_vp != u) NUM_CHANGES.add(1)

              if(vIsLeaf){
                LIN_SIZE.add(1)
                (false, v, mcu_vp)
              }
              else{
                LOUT_SIZE.add(1)
                (true, v, mcu_vp)
              }
            }
            else{// v is a local minimum but not the global minimum because 'it' has only large neighbors.
              LOUT_SIZE.add(1)
              (true, v, mu)
            }
          }
        }
      }

      it.flatMap{processNode}


    }

    val hconf = new SerializableConfiguration(sc.hadoopConfiguration)
    val lout = res_all.filtered(tmpPath, f"large-$round%05d", hconf).persist(StorageLevel.DISK_ONLY)
    lout.count()

    (lout, NUM_CHANGES.value, LOUT_SIZE.value, LCC_SIZE.value, LIN_SIZE.value)

  }

  /**
    * PA-Small-Star Operation.
    * For each node n, this operation links each small neighbor v to the minimum node mcu(p(v))
    * in the same partition p(v) that contains the neighbor v.
    *
    * @param inputRDD the input rdd
    * @param numPartitions the number of partitions
    * @param round current round number
    * @param tmpPath the temporary path to save the intermediate results
    * @return (RDD for next round input, # changed edges, # of 'out' edges,
    *         # filtered 'in' edges)
    */
  override def smallStar(inputRDD: RDD[(Long, Long)], numPartitions: Int,
                round: Int, tmpPath: String): (RDD[(Long, Long)], Long, Long, Long) = {

    val sc = inputRDD.sparkContext

    val NUM_CHANGES = sc.longAccumulator
    val SIN_SIZE = sc.longAccumulator
    val SOUT_SIZE = sc.longAccumulator

    val groupedRDD = inputRDD.flatMap{
      case (u, v) => Seq((u, v), (v, u))
    }.starGrouped()

    val tmpPaths = sc.hadoopConfiguration.getTrimmedStrings("yarn.nodemanager.local-dirs")

    val res_all = groupedRDD.mapPartitions{ it =>

      val longExternalSorter = new LongExternalSorter(tmpPaths)

      def processNode(x: (Long, Iterator[Long])): Iterator[(Boolean, Long, Long)] ={
        val (u, uN) = x

        val mcu = Array.fill[Long](numPartitions)(Long.MaxValue)
        val up = (u % numPartitions).toInt
        mcu(up) = u

        var isLeaf = true

        val _uN_small = uN.filter { v =>

          if(v > u) isLeaf = false
          val vp = (v % numPartitions).toInt
          mcu(vp) = Math.min(v, mcu(vp))

          v < u
        }

        val uN_small = longExternalSorter.sort(_uN_small)

        val mu = mcu.min

        val sout = uN_small filter {_ != mu} map { v =>

          val vp = (v % numPartitions).toInt

          NUM_CHANGES.add(1)
          SOUT_SIZE.add(1)

          if (v != mcu(vp))
            (true, v, mcu(vp))
          else // v is mcu but not mu
            (true, v, mu)

        }

        if (u != mcu(up)) {
          if (isLeaf) {
            SIN_SIZE.add(1)
            sout ++ Iterator((false, u, mcu(up)))
          }
          else {
            SOUT_SIZE.add(1)
            sout ++ Iterator((true, u, mcu(up)))
          }
        }
        else if (u != mu) {
          SOUT_SIZE.add(1)
          sout ++ Iterator((true, if (isLeaf) ~u else u, mu))
        }
        else sout

      }

      it.flatMap{processNode}

    }

    val hconf = new SerializableConfiguration(sc.hadoopConfiguration)

    val sout = res_all.filtered(tmpPath, f"small-$round%05d", hconf).persist(StorageLevel.DISK_ONLY)

    sout.count()

    (sout, NUM_CHANGES.value, SOUT_SIZE.value, SIN_SIZE.value)

  }

}

object PACCExternal{
  def main(args: Array[String]): Unit = new PACCExternal().runByCLI(args)
}

