/**
  * PegasusN: Peta-Scale Graph Mining System (Pegasus v3.0)
  * Authors: Chiwan Park, Ha-Myung Park, U Kang
  *
  * This software is licensed under Apache License, Version 2.0 (the  "License");
  * you may not use this file except in compliance with the License.
  * You may obtain a copy of the License at
  *
  * http://www.apache.org/licenses/LICENSE-2.0
  *
  * Unless required by applicable law or agreed to in writing, software
  * distributed under the License is distributed on an "AS IS" BASIS,
  * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  * See the License for the specific language governing permissions and
  * limitations under the License.
  * -------------------------------------------------------------------------
  * File: UnionFind.java
  * - Find Connected components.
  * - Single machine in-memory algorithm.
  * Version: 3.0
  */

package pegasus.spark.cc

import org.apache.spark.rdd.RDD
import org.apache.spark.sql.Dataset

import scala.collection.mutable

object UnionFind{

  /**
    * Find connected components.
    * @param edges the edges of an input graph
    * @return iterator of connected components
    */
  def run(edges: Iterator[(Long, Long)]): Iterator[(Long, Long)] ={

    val parent = new mutable.LongMap[Long]().withDefaultValue(-1)

    // find the minimum node rechable from node x.
    def find(x: Long): Long = {
      val res = parent.get(x) match {
        case None => x
        case Some(p) =>
          val new_p = find(p)
          parent(x) = new_p
          new_p
      }

      res
    }

    // union two connected components concerning the nodes in edge (x, y)
    def union(x: Long, y: Long): Unit ={
      val r1 = find(x)
      val r2 = find(y)

      if(r1 > r2)
        parent(r1) = r2
      else if(r1 < r2)
        parent(r2) = r1
    }

    for( (u, v) <- edges){
      if(find(u) != find(v)){
        union(u, v)
      }
    }

    parent.map{case (x,y) => (x, find(x))}.iterator

  }

  // interface fo Dataset
  def run(out: Dataset[(Long, Long)]): Dataset[(Long, Long)] = {
    import out.sparkSession.implicits._
    val arr = out.collect().iterator

    run(arr).toSeq.toDS()
  }

  // interface for RDD
  def run(out: RDD[(Long, Long)]): RDD[(Long, Long)] = {
    val arr = out.collect().iterator
    out.sparkContext.parallelize[(Long, Long)](run(arr).toStream)
  }


}
