package pegasus.spark.cc.deprecated

import java.util.StringTokenizer

import org.apache.log4j.{Level, Logger}
import org.apache.spark.broadcast.Broadcast
import org.apache.spark.rdd.RDD
import org.apache.spark.{Partitioner, SparkConf, SparkContext}
import pegasus.spark.cc.UnionFind

/**
  * Created by hmpark on 17. 3. 16.
  */
object PACC {

//  Logger.getLogger("org").setLevel(Level.WARN)
//  Logger.getLogger("akka").setLevel(Level.WARN)

  val logger = Logger.getLogger(getClass)
    logger.setLevel(Level.INFO)

  case class Config(localThreshold: Int = 100000, numPartitions: Int = 80,
                    inputPath: String = "", outputPath: String = "")

  def main(args: Array[String]): Unit = {

    val parser = new scopt.OptionParser[Config]("pacc"){
      head("pacc", "1.0")

      opt[Int]('t', "localThreshold")
        .action( (x, c) => c.copy(localThreshold = x))
        .text("if the number of remaining edges are lower than this value, " +
          "pacc run a single machine algorithm (LocalCC). (default: 100000)")

      opt[Int]('p', "numPartition").required()
        .action((x, c) => c.copy(numPartitions = x))
        .text("the number of partitions. (default: 80)")

      arg[String]("input")
        .action((x, c) => c.copy(inputPath = x))
        .text("input file path.")

      arg[String]("output")
        .action((x,c) => c.copy(outputPath = x))
        .text("output file path.")

    }

    parser.parse(args, Config()) match {
      case Some(opts) =>

        logger.info(f"inputPath: ${opts.inputPath}, output: ${opts.outputPath}, " +
          f"localTrheshold: ${opts.localThreshold}, numPartitions: ${opts.numPartitions}")

        val conf = new SparkConf().setAppName("[PACC]" + opts.inputPath)
          .set("spark.serializer", "org.apache.spark.serializer.KryoSerializer")
          .set("spark.kryo.registrator", "pegasus.spark.subgraph.MyRegistrator")
          .set("spark.kryoserializer.buffer", "24m")

        val sc = new SparkContext(conf)

        run(opts.inputPath, opts.localThreshold, opts.numPartitions, sc)
          .map{ case (u, v) => u + "\t" + v }
          .saveAsTextFile(opts.outputPath)

      case None =>
    }


  }

  var numPartitions: Broadcast[Int] = _


  def run(inputPath: String, localThreshold: Int, numPartitions: Int, sc: SparkContext): RDD[(Long, Long)] = {

    this.numPartitions = sc.broadcast(numPartitions)

    //initialize
    var out = sc.textFile(inputPath).map{ line =>
      val st = new StringTokenizer(line)
      val u = st.nextToken().toLong
      val v = st.nextToken().toLong
      (u, v)
    }

    var numEdges = out.count()

    var prevRDD = out
    var converge = false

    var cc: RDD[(Long, Long)] = sc.emptyRDD[(Long, Long)]
    var in: RDD[(Long, Long)] = sc.emptyRDD[(Long, Long)]

    do{

      if(numEdges > localThreshold) {
        val (lout, lcc, l_change, lout_size, lcc_size) = largeStar(out)

        cc = cc ++ lcc

        val (sout, sin, s_change, sout_size, sin_size) = smallStar(lout)
        out = sout
        in = in ++ sin

        logger.info(f"lout: $lout_size, lcc: $lcc_size, sout: $sout_size, sin: $sin_size")

        converge = l_change == 0 && s_change == 0
        numEdges = sout_size

      }
      else{
        out = UnionFind.run(out.map{ case (u, v) => (if (u < 0) ~u else u, v) })
        converge = true
      }
    }while(!converge)

    ccComputation(out ++ cc ++ in)
  }

  def largeStar(inputRDD: RDD[(Long, Long)]): (RDD[(Long, Long)], RDD[(Long, Long)], Long, Long, Long) = {

    val sc = inputRDD.sparkContext

    val NUM_CHANGES = sc.longAccumulator
    val LCC_SIZE = sc.longAccumulator
    val LOUT_SIZE = sc.longAccumulator

    val groupedRDD = inputRDD.flatMap{
      case (u, v) =>
        if (u < 0) Seq((v, u))
        else Seq((u, v), (v, u))
    }.groupByKey()

    val LCC = true
    val LOUT = false

    val res_all = groupedRDD.map {
      case (u, _uN) =>

        val numPartitions = this.numPartitions.value

        val mcu = Array.fill[Long](numPartitions)(Long.MaxValue)
        mcu((u % numPartitions).toInt) = u





        var isStar = true

        val uN_large = _uN.map{ v_raw =>
          if (v_raw >= 0) {
            isStar = false
            v_raw
          } else ~v_raw
        }.filter { v =>

          val vp = (v % numPartitions).toInt

          mcu(vp) = Math.min(v, mcu(vp))

          v > u
        }.toArray

        val mu = mcu.min

        if(mu != u) isStar = false

        if(isStar){
          val lcc = uN_large.map{ v =>
            (v, u)
          }

          LCC_SIZE.add(lcc.length)
          (LCC, lcc)
        }
        else{
          val lout = uN_large.flatMap{ v =>
            val vp = (v % numPartitions).toInt
            val mcu_vp = mcu(vp)
            if(v != mcu_vp) {
              if(mcu_vp != u) NUM_CHANGES.add(1)
              Seq((v, mcu_vp))
            }
            else if(v != mu) {
              if(mu != u) NUM_CHANGES.add(1)
              Seq((v, mu))
            }
            else Seq[(Long, Long)]()


          }

          LOUT_SIZE.add(lout.length)
          (LOUT, lout)
        }

    }

    val lcc = res_all.filter(_._1 == LCC).flatMap(_._2)
    val lout = res_all.filter(_._1 == LOUT).flatMap(_._2)

    res_all.count()

    (lout, lcc, NUM_CHANGES.value, LOUT_SIZE.value, LCC_SIZE.value)

  }

  def smallStar(inputRDD: RDD[(Long, Long)]): (RDD[(Long, Long)], RDD[(Long, Long)], Long, Long, Long) = {

    val sc = inputRDD.sparkContext

    val NUM_CHANGES = sc.longAccumulator
    val SIN_SIZE = sc.longAccumulator
    val SOUT_SIZE = sc.longAccumulator

    val groupedRDD = inputRDD.flatMap{
      case (u, v) => Seq((u, v), (v, u))
    }.groupByKey()

    val SIN = true
    val SOUT = false

    val res_all = groupedRDD.flatMap{
      case (u, uN) =>

        val numPartitions = this.numPartitions.value

        val mcu = Array.fill[Long](numPartitions)(Long.MaxValue)
        val up = (u % numPartitions).toInt
        mcu(up) = u




        var isLeaf = true

        val uN_small = uN.filter { v =>

          if(v > u) isLeaf = false

          val vp = (v % numPartitions).toInt

          mcu(vp) = Math.min(v, mcu(vp))

          v < u
        }.toArray

        val mu = mcu.min


//        if(u == mu) return


        val res_u = if (u != mcu(up)) //u is not local minimum.
          if (isLeaf) {
            SIN_SIZE.add(1)
            (SIN, Array((u, mcu(up))))
          }
          else {
            SOUT_SIZE.add(1)
            (SOUT, Array((u, mcu(up))))
          }
        else if (u != mu) { //u is the minimum in local but not in global
          SOUT_SIZE.add(1)
          (SOUT, Array((if (isLeaf) ~u else u, mu)))
        }
        else //u is the global minimal
          (SIN, Array[(Long, Long)]())



        val sout = uN_small.flatMap{ v =>

          val vp = (v % numPartitions).toInt

          if (v != mcu(vp)) {
            Seq((v, mcu(vp)))
          }
          else if (v != mu) {
            Seq((v, mu))
          }
          else {
            Seq[(Long, Long)]()
          }
        }

        SOUT_SIZE.add(sout.length)
        NUM_CHANGES.add(sout.length)

        Seq((SOUT, sout), res_u)

    }

    val sin = res_all.filter(_._1 == SIN).flatMap(_._2)
    val sout = res_all.filter(_._1 == SOUT).flatMap(_._2)

    res_all.count()

    (sout, sin, NUM_CHANGES.value, SOUT_SIZE.value, SIN_SIZE.value)

  }

  def ccComputation(remains: RDD[(Long, Long)]): RDD[(Long, Long)] = {

    val numPartitions = this.numPartitions.value

    remains.map{ case (u, v) => (if(u < 0) ~u else u, v) }
      .partitionBy(new CCPartitioner(numPartitions))
      .mapPartitions(edges => UnionFind.run(edges))

  }

  class CCPartitioner(numParts: Int) extends Partitioner {

    def getPartition(key: Any): Int = {
      (key.asInstanceOf[Long] % numParts).toInt
    }

    override def numPartitions: Int = numParts
  }

}


