package pegasus.spark.cc.deprecated.alpha

import java.io.{ObjectInputStream, ObjectOutputStream}
import java.util.StringTokenizer

import org.apache.hadoop.conf.Configuration
import org.apache.hadoop.fs.{FileSystem, Path}
import org.apache.hadoop.io.{LongWritable, SequenceFile}
import org.apache.hadoop.mapred.JobConf
import org.apache.log4j.{Level, Logger}
import org.apache.spark.rdd.RDD
import org.apache.spark.storage.StorageLevel
import org.apache.spark.{HashPartitioner, SparkConf, SparkContext}
import pegasus.spark.cc.deprecated.PACC.CCPartitioner
import pegasus.spark.cc.UnionFind
import pegasus.spark.common.io.ExternalSorter

/**
  * Created by hmpark on 17. 3. 16.
  */
object PACCDual {

  private val logger = Logger.getLogger(getClass)
  logger.setLevel(Level.INFO)

  case class Config(localThreshold: Int = 100000, numPartitions: Int = 80,
                    inputPath: String = "", outputPath: String = "")

  def main(args: Array[String]): Unit = {

    val parser = new scopt.OptionParser[Config]("pacc"){
      head("pacc", "1.0")

      opt[Int]('t', "localThreshold")
        .action( (x, c) => c.copy(localThreshold = x))
        .text("if the number of remaining edges are lower than this value, " +
          "pacc run a single machine algorithm (LocalCC). (default: 100000)")

      opt[Int]('p', "numPartition").required()
        .action((x, c) => c.copy(numPartitions = x))
        .text("the number of partitions. (default: 80)")

      arg[String]("input")
        .action((x, c) => c.copy(inputPath = x))
        .text("input file path.")

      arg[String]("output")
        .action((x,c) => c.copy(outputPath = x))
        .text("output file path.")
    }

    parser.parse(args, Config()) match {
      case Some(opts) =>

        logger.info(f"inputPath: ${opts.inputPath}, output: ${opts.outputPath}, " +
          f"localTrheshold: ${opts.localThreshold}, numPartitions: ${opts.numPartitions}")

        val conf = new SparkConf().setAppName("[PACCDual]" + opts.inputPath)
          .set("spark.serializer", "org.apache.spark.serializer.KryoSerializer")
          .set("spark.kryo.registrator", "pegasus.spark.subgraph.MyRegistrator")
          .set("spark.kryoserializer.buffer", "24m")
          .set("spark.shuffle.blockTransferService", "nio")

        val sc = new SparkContext(conf)

        FileSystem.get(sc.hadoopConfiguration).delete(new Path(opts.outputPath), true)

        run(opts.inputPath, opts.localThreshold, opts.numPartitions, sc)
          .map{ case (u, v) => u + "\t" + v }
          .saveAsTextFile(opts.outputPath)

        sc.stop()

      case None =>
    }
  }


  def run(inputPath: String, localThreshold: Int, numPartitions: Int, sc: SparkContext): RDD[(Long, Long)] = {

    val tmpPath = inputPath + ".pacc.tmp"

    //initialize
    var out = sc.textFile(inputPath).map{ line =>
      val st = new StringTokenizer(line)
      val u = st.nextToken().toLong
      val v = st.nextToken().toLong
      (u, v)
    }

    var numEdges = out.count()

    var converge = false

    var round = 0

    do{

      if(numEdges > localThreshold) {
        val (lout, l_change, lout_size, lcc_size, lin_size) = largeStar(out, numPartitions, round, tmpPath)
        out.unpersist()
        val (sout, s_change, sout_size, sin_size) = smallStar(lout, numPartitions, round, tmpPath)
        lout.unpersist()
        out = sout

        logger.info(f"lout: $lout_size, lcc: $lcc_size, lin: $lin_size sout: $sout_size, sin: $sin_size")

        converge = l_change == 0 && s_change == 0
        numEdges = sout_size

        round += 1
      }
      else{
        out = UnionFind.run(out.map{ case (u, v) => (if (u < 0) ~u else u, v) })
        converge = true
      }
    }while(!converge)

    val others = sc.sequenceFile[Long, Long](tmpPath)
    val res = ccComputation(out ++ others, numPartitions)

    FileSystem.get(sc.hadoopConfiguration).deleteOnExit(new Path(tmpPath))

    res
  }

  def largeStar(inputRDD: RDD[(Long, Long)], numPartitions: Int,
                round: Int, tmpPath: String): (RDD[(Long, Long)], Long, Long, Long, Long) = {

    val sc = inputRDD.sparkContext

    val NUM_CHANGES = sc.longAccumulator
    val LCC_SIZE = sc.longAccumulator
    val LIN_SIZE = sc.longAccumulator
    val LOUT_SIZE = sc.longAccumulator
    val tmpPaths = sc.hadoopConfiguration.getTrimmedStrings("yarn.nodemanager.local-dirs")

    val groupedRDD = inputRDD.flatMap{ case (u, v) =>
      if (u < 0) Seq((v, u))
      else Seq((u, v), (v, u))
    }.starGrouped()

    def processNode(x: (Long, Iterator[Long])): Iterator[(Boolean, Long, Long)] ={
      val (u, uN) = x

      val mcu = Array.fill[Long](numPartitions)(Long.MaxValue)
      mcu((u % numPartitions).toInt) = u

      var isStar = true
      var cnt: Long = 0

      val _uN_large = uN.filter { v_raw =>
        val v = if (v_raw >= 0) {
          isStar = false
          v_raw
        } else ~v_raw

        cnt += 1
        if(u == 12 && cnt % 10000 == 0){
          println(f"u($u)-v_raw($v_raw)-cnt($cnt)")
        }

        val vp = (v % numPartitions).toInt
        mcu(vp) = Math.min(v, mcu(vp))

        v > u
      }

      val uN_large = ExternalSorter.sort(_uN_large, tmpPaths)

      val mu = mcu.min

      if (isStar) uN_large.map { v_raw => LCC_SIZE.add(1); (false, ~v_raw, u) }
      else{
        uN_large.map{ v_raw =>
          val vIsLeaf = v_raw < 0
          val v: Long = if(vIsLeaf) ~v_raw else v_raw

          val vp = (v % numPartitions).toInt
          val mcu_vp = mcu(vp)

          if(v != mcu_vp) {

            if(mcu_vp != u) NUM_CHANGES.add(1)

            if(vIsLeaf){
              LIN_SIZE.add(1)
              (false, v, mcu_vp)
            }
            else{
              LOUT_SIZE.add(1)
              (true, v, mcu_vp)
            }
          }
          else{// v is a local minimum but not the global minimum because 'it' has only large neighbors.
            LOUT_SIZE.add(1)
            (true, v, mu)
          }
        }
      }
    }

    val res_all = groupedRDD.flatMap(x => processNode(x))

    val hconf = new SerializableJobConf(sc.hadoopConfiguration)

    val lout = res_all.filtered(tmpPath, f"large-$round%05d", hconf).persist(StorageLevel.DISK_ONLY)

    lout.count()

    (lout, NUM_CHANGES.value, LOUT_SIZE.value, LCC_SIZE.value, LIN_SIZE.value)

  }


  def smallStar(inputRDD: RDD[(Long, Long)], numPartitions: Int,
                round: Int, tmpPath: String): (RDD[(Long, Long)], Long, Long, Long) = {

    val sc = inputRDD.sparkContext

    val NUM_CHANGES = sc.longAccumulator
    val SIN_SIZE = sc.longAccumulator
    val SOUT_SIZE = sc.longAccumulator
    val tmpPaths = sc.hadoopConfiguration.getTrimmedStrings("yarn.nodemanager.local-dirs")

    val groupedRDD = inputRDD.flatMap{
      case (u, v) => Seq((u, v), (v, u))
    }.starGrouped()

    val res_all = groupedRDD.flatMap{
      case (u, uN) =>

        val mcu = Array.fill[Long](numPartitions)(Long.MaxValue)
        val up = (u % numPartitions).toInt
        mcu(up) = u

        var isLeaf = true

        val _uN_small = uN.filter { v =>

          if(v > u) isLeaf = false
          val vp = (v % numPartitions).toInt
          mcu(vp) = Math.min(v, mcu(vp))

          v < u
        }

        val uN_small = ExternalSorter.sort(_uN_small, tmpPaths)

        val mu = mcu.min

        val sout = uN_small filter {_ != mu} map { v =>

          val vp = (v % numPartitions).toInt

          NUM_CHANGES.add(1)
          SOUT_SIZE.add(1)

          if (v != mcu(vp))
            (true, v, mcu(vp))
          else // v is mcu but not mu
            (true, v, mu)

        }

        if (u != mcu(up)) {
          if (isLeaf) {
            SIN_SIZE.add(1)
            sout ++ Iterator((false, u, mcu(up)))
          }
          else {
            SOUT_SIZE.add(1)
            sout ++ Iterator((true, u, mcu(up)))
          }
        }
        else if (u != mu) {
          SOUT_SIZE.add(1)
          sout ++ Iterator((true, if (isLeaf) ~u else u, mu))
        }
        else sout
    }

    val hconf = new SerializableJobConf(sc.hadoopConfiguration)

    val sout = res_all.filtered(tmpPath, f"small-$round%05d", hconf).persist(StorageLevel.DISK_ONLY)

    sout.count()

    (sout, NUM_CHANGES.value, SOUT_SIZE.value, SIN_SIZE.value)

  }

  def ccComputation(remains: RDD[(Long, Long)], numPartitions: Int): RDD[(Long, Long)] = {
    remains.map{ case (u, v) => (if(u < 0) ~u else u, v) }
      .partitionBy(new CCPartitioner(numPartitions))
      .mapPartitions(edges => UnionFind.run(edges))
  }

  class SerializableJobConf(@transient var value: Configuration) extends Serializable {
    private def writeObject(out: ObjectOutputStream): Unit = {
      out.defaultWriteObject()
      value.write(out)
    }

    private def readObject(in: ObjectInputStream): Unit = {
      value = new JobConf(false)
      value.readFields(in)
    }
  }

  implicit class FilteredRDD(rdd: RDD[(Boolean, Long, Long)]) {
    def filtered(path: String, prefix: String, conf: SerializableJobConf): RDD[(Long,Long)] ={

      rdd.mapPartitionsWithIndex{
        case (partitionId, it) =>
          val filePath = f"$path/$prefix-$partitionId%05d"
          println(filePath)
          it.filtered(filePath, conf.value)
      }
    }
  }

  implicit class FilteredIterator(it: Iterator[(Boolean, Long, Long)]) {
    def filtered(path: String, conf: Configuration)
    = new Iterator[(Long,Long)] {

      val writer: SequenceFile.Writer = SequenceFile.createWriter(conf,
        SequenceFile.Writer.keyClass(classOf[LongWritable]),
        SequenceFile.Writer.valueClass(classOf[LongWritable]),
        SequenceFile.Writer.file(new Path(path))
      )

      val aw = new LongWritable
      val bw = new LongWritable

      def close(){
        writer.close()
      }

      def pp(elem: (Boolean, Long, Long)): Boolean = {
        if(!elem._1){
          aw.set(elem._2)
          bw.set(elem._3)
          writer.append(aw, bw)
        }
        elem._1
      }

      private var hd: (Boolean, Long, Long) = _
      private var hdDefined: Boolean = false

      def hasNext: Boolean = hdDefined || {
        do {
          if (!it.hasNext){
            close()
            return false
          }
          hd = it.next()
        } while (!pp(hd))
        hdDefined = true
        true
      }

      def next(): (Long, Long) = if (hasNext) { hdDefined = false; (hd._2, hd._3) } else Iterator.empty.next()
    }
  }

  implicit class StarRDDOp(rdd: RDD[(Long, Long)]){
    def starGrouped() = {
      rdd.repartitionAndSortWithinPartitions(new HashPartitioner(rdd.getNumPartitions)).mapPartitionsWithIndex{
        case (idx, it) =>
          it.starGrouped()
      }
    }
  }

  implicit class StarIteratorOp(it: Iterator[(Long, Long)]){

    def starGrouped()= new Iterator[(Long, Iterator[Long])] {
      var first: Option[(Long, Long)] = None
      var prev: GroupedIterator = _

      override def hasNext: Boolean = first.isDefined ||
        (prev != null && {
          first = prev.consumeAndGetHead
          first.isDefined
        }) ||
        (it.hasNext && {
          first = Some(it.next())
          first.isDefined
        })

      override def next(): (Long, Iterator[Long]) = {

        if(hasNext) {
          prev = new GroupedIterator(first, it)
          val res = (first.get._1, prev.map(_._2))
          first = None
          res
        }
        else Iterator.empty.next()

      }

      class GroupedIterator(first: Option[(Long, Long)], base: Iterator[(Long, Long)]) extends Iterator[(Long, Long)]{

        private var (hd, hdDefined): ((Long, Long), Boolean) = first match {
          case Some(x) => (x, true)
          case None => (null, false)
        }
        var tailConsumed: Boolean = false

        private var tail: Iterator[(Long, Long)] = base

        def hasNext = hdDefined || tail.hasNext && {

          val cur = tail.next()
          tailConsumed = true

          if(cur._1 == hd._1) hdDefined = true
          else tail = Iterator.empty

          hd = cur
          hdDefined

        }
        def next() = if (hasNext) { hdDefined = false; tailConsumed = false; hd } else Iterator.empty.next()

        def consumeAndGetHead: Option[(Long, Long)] = {
          while(hasNext) next

          if(tailConsumed) Some(hd)
          else None

        }
      }
    }
  }

}


