package pegasus.spark.cc.deprecated.beta

import java.util.StringTokenizer

import org.apache.hadoop.fs.{FileSystem, Path}
import org.apache.log4j.{Level, Logger}
import org.apache.spark
import org.apache.spark.rdd.{RDD, ShuffledRDD}
import org.apache.spark.util.LongAccumulator
import org.apache.spark.{Partitioner, SparkConf, SparkContext}
import pegasus.spark.cc.deprecated.PACC.CCPartitioner
import pegasus.spark.cc.UnionFind
import pegasus.spark.common.io.ExternalSorter

import scala.collection.{AbstractIterator, Iterator}
import scala.collection.Iterator.empty

/**
  * Created by hmpark on 17. 3. 16.
  */
object PACCFile {

  val logger = Logger.getLogger(getClass)
    logger.setLevel(Level.WARN)

  case class Config(localThreshold: Int = 100000, numPartitions: Int = 80,
                    inputPath: String = "", outputPath: String = "")

  def main(args: Array[String]): Unit = {

    val parser = new scopt.OptionParser[Config]("pacc"){
      head("pacc", "1.0")

      opt[Int]('t', "localThreshold")
        .action( (x, c) => c.copy(localThreshold = x))
        .text("if the number of remaining edges are lower than this value, " +
          "pacc run a single machine algorithm (LocalCC). (default: 100000)")

      opt[Int]('p', "numPartition").required()
        .action((x, c) => c.copy(numPartitions = x))
        .text("the number of partitions. (default: 80)")

      arg[String]("input")
        .action((x, c) => c.copy(inputPath = x))
        .text("input file path.")

      arg[String]("output")
        .action((x,c) => c.copy(outputPath = x))
        .text("output file path.")
    }

    parser.parse(args, Config()) match {
      case Some(opts) =>

        logger.info(f"inputPath: ${opts.inputPath}, output: ${opts.outputPath}, " +
          f"localTrheshold: ${opts.localThreshold}, numPartitions: ${opts.numPartitions}")

        val conf = new SparkConf().setAppName("[PACCDisk]" + opts.inputPath)
          .set("spark.serializer", "org.apache.spark.serializer.KryoSerializer")
          .set("spark.kryo.registrator", "pegasus.spark.subgraph.MyRegistrator")
          .set("spark.kryoserializer.buffer", "24m")
          .set("spark.network.timeout", "12000s")

        val sc = new SparkContext(conf)

        FileSystem.get(sc.hadoopConfiguration).delete(new Path(opts.outputPath), true)

        run(opts.inputPath, opts.localThreshold, opts.numPartitions, sc)
          .map{ case (u, v) => u + "\t" + v }
          .saveAsTextFile(opts.outputPath)

        sc.stop()

      case None =>
    }
  }

  def run(inputPath: String, localThreshold: Int, numPartitions: Int, sc: SparkContext): RDD[(Long, Long)] = {

    //initialize
    var out = sc.textFile(inputPath).map{ line =>
      val st = new StringTokenizer(line)
      val u = st.nextToken().toLong
      val v = st.nextToken().toLong
      (u, v)
    }

    var numEdges = out.count()

    var prevRDD = out
    var converge = false

    var others: RDD[(Long, Long)] = sc.emptyRDD[(Long, Long)]

    do{

      if(numEdges > localThreshold) {
        val (lout, lothers, l_change, lout_size, lcc_size, lin_size) = largeStar(out, numPartitions)

        others ++= lothers

        val (sout, sin, s_change, sout_size, sin_size) = smallStar(lout, numPartitions)
        out = sout
        others ++= sin

        logger.info(f"lout: $lout_size, lcc: $lcc_size, sout: $sout_size, sin: $sin_size")

        converge = l_change == 0 && s_change == 0
        numEdges = sout_size

      }
      else{
        out = UnionFind.run(out.map{ case (u, v) => (if (u < 0) ~u else u, v) })
        converge = true
      }
    }while(!converge)

    ccComputation(out ++ others, numPartitions)
  }

  implicit class MyRDD(rdd: RDD[(Long, Long)]) {

    val partitions = rdd.getNumPartitions

    val pairOrdering = new Ordering[(Long, Long)] {
      override def compare(x: (Long, Long), y: (Long, Long)): Int = {
        if (x._1 > y._1) 1
        else if (x._1 < y._1) -1
        else {
          if(x._2 < 0) 1
          else if(y._2 < 0) -1
          else if(x > y) 1
          else if(x < y) -1
          else 0
        }
      }
    }

    val keyPairPartitioner = new Partitioner {
      override def getPartition(key: Any): Int = {
        val keyPair = key.asInstanceOf[(Long, Long)]
        (keyPair._1 % partitions).toInt
      }
      override def numPartitions: Int = partitions
    }

    def groupByKeyWithSecondarySort(): RDD[(Long, Iterator[Long])] ={

      val allInKeyRDD = rdd.map(x => (x, null))

      allInKeyRDD.partitionBy(keyPairPartitioner)
        .asInstanceOf[ShuffledRDD[(Long, Long), Null, Null]].setKeyOrdering(pairOrdering)
        .map{case (x, _) => x}
        .mapPartitions { it =>
          new Iterator[(Long, Iterator[Long])]{

            var iter: Iterator[(Long, Long)] = it

            override def hasNext: Boolean = iter.hasNext
            override def next(): (Long, Iterator[Long]) = {
              val (u, v) = iter.next()
              val (i1, i2) = iter.span(x=> x._1 == u)
              iter = i2
              (u, Iterator(v) ++ i1.map(_._2))
            }
          }
        }
//      .map{case (x, n) => x}


    }

    def groupByKeyEx(): RDD[(Long, Iterator[Long])] = {

      rdd.partitionBy(new spark.HashPartitioner(rdd.getNumPartitions))
        .asInstanceOf[ShuffledRDD[Long, Long, Long]]
        .setKeyOrdering(Ordering.Long)
        .mapPartitions { it =>

          new Iterator[(Long, Iterator[Long])] {

            var first: (Long, Long) = _
            var firstDefined: Boolean = false

            override def hasNext: Boolean = firstDefined || it.hasNext && {
              first = it.next()
              firstDefined = true
              true
            }

            override def next(): (Long, Iterator[Long]) = if (hasNext) {

              val i1 = new Iterator[(Long, Long)] {
                var hd: (Long, Long) = _
                private var hdDefined: Boolean = false
                private var tail: Iterator[(Long, Long)] = it

                def hasNext: Boolean = hdDefined || tail.hasNext && {
                  hd = tail.next()
                  if (first._1 == hd._1) hdDefined = true
                  else {
                    firstDefined = true
                    first = hd
                    tail = Iterator.empty
                  }
                  hdDefined
                }

                def next(): (Long, Long) = if (hasNext) {
                  hdDefined = false
                  hd
                } else empty.next()
              }


              firstDefined = false
              (first._1, Iterator(first._2) ++ i1.map(_._2))

            } else empty.next()

          }
        }
    }
  }

  class GroupIterator


  def largeStar(inputRDD: RDD[(Long, Long)], numPartitions: Int): (RDD[(Long, Long)],
    RDD[(Long, Long)], Long, Long, Long, Long) = {

    val sc = inputRDD.sparkContext

    val NUM_CHANGES = sc.longAccumulator
    val LCC_SIZE = sc.longAccumulator
    val LIN_SIZE = sc.longAccumulator
    val LOUT_SIZE = sc.longAccumulator

    val groupedRDD = inputRDD.flatMap{
      case (u, v) =>
        if (u < 0) Seq((v, u))
        else Seq((u, v), (v, u))
    }.groupByKeyEx()

    val LOTHERS = true
    val LOUT = false
    val tmpPaths = sc.hadoopConfiguration.getTrimmedStrings("yarn.nodemanager.local-dirs")

    val res_all = groupedRDD.flatMap {
      case (u, uN) =>
        val mcu = Array.fill[Long](numPartitions)(Long.MaxValue)
        mcu((u % numPartitions).toInt) = u

        var isStar = true

        var cnt: Long = 0

        val _uN_large = uN.filter { v_raw =>
          val v = if (v_raw >= 0) {
            isStar = false
            v_raw
          } else ~v_raw

          val vp = (v % numPartitions).toInt
          mcu(vp) = Math.min(v, mcu(vp))

          cnt += 1
          if(cnt % 10000 == 0){
            println(f"u($u)-v_raw($v_raw)-cnt($cnt)")
          }

          v > u
        }

        val uN_large = ExternalSorter.sort(_uN_large, tmpPaths)

        val mu = mcu.min

        if (isStar) uN_large.map { v_raw => LCC_SIZE.add(1); ((LOTHERS, ~v_raw), u) }
        else{
          uN_large.map{ v_raw =>
            val vIsLeaf = v_raw < 0
            val v: Long = if(vIsLeaf) ~v_raw else v_raw

            val vp = (v % numPartitions).toInt
            val mcu_vp = mcu(vp)

            if(v != mcu_vp) {

              if(mcu_vp != u) NUM_CHANGES.add(1)

              if(vIsLeaf){
                LIN_SIZE.add(1)
                ((LOTHERS, v), mcu_vp)
              }
              else{
                LOUT_SIZE.add(1)
                ((LOUT, v), mcu_vp)
              }
            }
            else{// v is a local minimum but not the global minimum because 'it' has only large neighbors.
              LOUT_SIZE.add(1)
              ((LOUT, v), mu)
            }
          }
        }

    }

    val lother = res_all.filter(_._1._1 == LOTHERS).map{case ((_, u), m) => (u, m)}
    val lout = res_all.filter(_._1._1 == LOUT).map{case ((_, u), m) => (u, m)}

    res_all.count()

    (lout, lother, NUM_CHANGES.value, LOUT_SIZE.value, LCC_SIZE.value, LIN_SIZE.value)

  }



  def smallStar(inputRDD: RDD[(Long, Long)], numPartitions: Int): (RDD[(Long, Long)],
    RDD[(Long, Long)], Long, Long, Long) = {

    val sc = inputRDD.sparkContext

    val NUM_CHANGES = sc.longAccumulator
    val SIN_SIZE = sc.longAccumulator
    val SOUT_SIZE = sc.longAccumulator


    val groupedRDD = inputRDD.flatMap{
      case (u, v) => Seq((u, v), (v, u))
    }.groupByKeyEx()

    val SIN = true
    val SOUT = false

    val res_all = groupedRDD.flatMap{
      case (u, uN) =>

        val mcu = Array.fill[Long](numPartitions)(Long.MaxValue)
        val up = (u % numPartitions).toInt
        mcu(up) = u

        var isLeaf = true

        val uN_small = uN.filter { v =>

          if(v > u) isLeaf = false
          val vp = (v % numPartitions).toInt
          mcu(vp) = Math.min(v, mcu(vp))

          v < u
        }.toArray.toStream

        val mu = mcu.min

        val sout = uN_small filter {_ != mu} map { v =>

          val vp = (v % numPartitions).toInt

          NUM_CHANGES.add(1)
          SOUT_SIZE.add(1)

          if (v != mcu(vp))
            ((SOUT, v), mcu(vp))
          else // v is mcu but not mu
            ((SOUT, v), mu)

        }

        if (u != mcu(up)) {
          if (isLeaf) {
            SIN_SIZE.add(1)
            sout ++ Stream(((SIN, u), mcu(up)))
          }
          else {
            SOUT_SIZE.add(1)
            sout ++ Stream(((SOUT, u), mcu(up)))
          }
        }
        else if (u != mu) {
          SOUT_SIZE.add(1)
          sout ++ Stream(((SOUT, if (isLeaf) ~u else u), mu))
        }
        else sout
    }

    val sin = res_all.filter(_._1._1 == SIN).map{case ((_, u), m) => (u, m)}
    val sout = res_all.filter(_._1._1 == SOUT).map{case ((_, u), m) => (u, m)}

    res_all.count()

    (sout, sin, NUM_CHANGES.value, SOUT_SIZE.value, SIN_SIZE.value)

  }

  def ccComputation(remains: RDD[(Long, Long)], numPartitions: Int): RDD[(Long, Long)] = {

    remains.map{ case (u, v) => (if(u < 0) ~u else u, v) }
      .partitionBy(new CCPartitioner(numPartitions))
      .mapPartitions(edges => UnionFind.run(edges))

  }

  class CPartitioner(numPart: Int) extends Partitioner{
    override def numPartitions: Int = numPart
    val halfNumPartitions = numPartitions / 2

    override def getPartition(key: Any): Int = {
      val (other, v) = key.asInstanceOf[(Boolean, Long)]
      if(other) (v % halfNumPartitions + halfNumPartitions).toInt
      else (v % halfNumPartitions).toInt
    }
  }


}


