package pegasus.spark.cc.deprecated.beta


import java.util.StringTokenizer

import org.apache.hadoop.fs.{FileSystem, Path}
import org.apache.log4j.{Level, Logger}
import org.apache.spark.{Partitioner, SparkConf, SparkContext}
import org.apache.spark.sql.{Dataset, SparkSession}
import pegasus.spark.cc.UnionFind
import pegasus.spark.cc.deprecated.beta.PACCDisk.{Config, logger, run}
import pegasus.spark.common.io.ExternalSorter

/**
  * Created by hmpark on 17. 3. 16.
  */
object PACCTest {

  val logger = Logger.getLogger(getClass)
    logger.setLevel(Level.WARN)

  case class Config(localThreshold: Int = 100000, numPartitions: Int = 80,
                    inputPath: String = "", outputPath: String = "")

  def main(args: Array[String]): Unit = {

    val parser = new scopt.OptionParser[Config]("pacc"){
      head("pacc", "1.0")

      opt[Int]('t', "localThreshold")
        .action( (x, c) => c.copy(localThreshold = x))
        .text("if the number of remaining edges are lower than this value, " +
          "pacc run a single machine algorithm (LocalCC). (default: 100000)")

      opt[Int]('p', "numPartition").required()
        .action((x, c) => c.copy(numPartitions = x))
        .text("the number of partitions. (default: 80)")

      arg[String]("input")
        .action((x, c) => c.copy(inputPath = x))
        .text("input file path.")

      arg[String]("output")
        .action((x,c) => c.copy(outputPath = x))
        .text("output file path.")
    }

    parser.parse(args, Config()) match {
      case Some(opts) =>

        logger.info(f"inputPath: ${opts.inputPath}, output: ${opts.outputPath}, " +
          f"localTrheshold: ${opts.localThreshold}, numPartitions: ${opts.numPartitions}")

        val conf = new SparkConf().setAppName("[PACCDisk]" + opts.inputPath)
          .set("spark.serializer", "org.apache.spark.serializer.KryoSerializer")
          .set("spark.kryo.registrator", "pegasus.spark.subgraph.MyRegistrator")
          .set("spark.kryoserializer.buffer", "24m")
          .set("spark.network.timeout", "12000s")

        val ss = SparkSession.builder().appName("[PACCTest]" + opts.inputPath).getOrCreate()

        import ss.implicits._

        FileSystem.get(ss.sparkContext.hadoopConfiguration).delete(new Path(opts.outputPath), true)

        run(opts.inputPath, opts.localThreshold, opts.numPartitions, ss)
          .map{ case (u, v) => u + "\t" + v }
          .rdd
          .saveAsTextFile(opts.outputPath)

        ss.stop()

      case None =>
    }

  }

  def run(inputPath: String, localThreshold: Int, numPartitions: Int, ss: SparkSession): Dataset[(Long, Long)] = {

    import ss.implicits._

    //initialize
    var out: Dataset[(Long, Long)] = ss.read.textFile(inputPath).map{line =>
      val st = new StringTokenizer(line)
      val u = st.nextToken().toLong
      val v = st.nextToken().toLong
      (u, v)
    }

    var numEdges = out.count()

    var converge = false

    var others: Dataset[(Long, Long)] = ss.emptyDataset[(Long, Long)]

    do{

      if(numEdges > localThreshold) {
        val (lout, lothers, l_change, lout_size, lcc_size, lin_size) = largeStar(out, numPartitions)

        others = others union lothers

        val (sout, sin, s_change, sout_size, sin_size) = smallStar(lout, numPartitions)
        out = sout
        others = others union sin

        logger.info(f"lout: $lout_size, lcc: $lcc_size, sout: $sout_size, sin: $sin_size")

        converge = l_change == 0 && s_change == 0
        numEdges = sout_size

      }
      else{
        out = UnionFind.run(out.map{ case (u, v) => (if (u < 0) ~u else u, v) })
        converge = true
      }
    }while(!converge)

    ccComputation(out union others, numPartitions)
  }

  def largeStar(inDS: Dataset[(Long, Long)], numPartitions: Int): (Dataset[(Long, Long)],
    Dataset[(Long, Long)], Long, Long, Long, Long) = {

    import inDS.sparkSession.implicits._
    val sc = inDS.sparkSession.sparkContext



    val NUM_CHANGES = sc.longAccumulator
    val LCC_SIZE = sc.longAccumulator
    val LIN_SIZE = sc.longAccumulator
    val LOUT_SIZE = sc.longAccumulator

    val groupedDS = inDS.flatMap{
      case (u, v) =>
        if (u < 0) Seq((v, u))
        else Seq((u, v), (v, u))
    }.groupByKey{case (u, v) => u}.mapValues{case (u, v) => v}

    val LOTHERS = true
    val LOUT = false
    val tmpPaths = sc.hadoopConfiguration.getTrimmedStrings("yarn.nodemanager.local-dirs")

    val res_all : Dataset[(Boolean, Long, Long)] = groupedDS.flatMapGroups {
      case (u, uN) =>

        val mcu = Array.fill[Long](numPartitions)(Long.MaxValue)
        mcu((u % numPartitions).toInt) = u

        var isStar = true

        val _uN_large = uN.filter { v_raw =>
          val v = if (v_raw >= 0) {
            isStar = false
            v_raw
          } else ~v_raw

          val vp = (v % numPartitions).toInt
          mcu(vp) = Math.min(v, mcu(vp))

          v > u
        }

        val uN_large = ExternalSorter.sort(_uN_large, tmpPaths)

        val mu = mcu.min

        if (isStar) uN_large.map { v_raw => LCC_SIZE.add(1); (LOTHERS, ~v_raw, u) }
        else{
          uN_large.map{ v_raw =>
            val vIsLeaf = v_raw < 0
            val v: Long = if(vIsLeaf) ~v_raw else v_raw

            val vp = (v % numPartitions).toInt
            val mcu_vp = mcu(vp)

            if(v != mcu_vp) {

              if(mcu_vp != u) NUM_CHANGES.add(1)

              if(vIsLeaf){
                LIN_SIZE.add(1)
                (LOTHERS, v, mcu_vp)
              }
              else{
                LOUT_SIZE.add(1)
                (LOUT, v, mcu_vp)
              }
            }
            else{// v is a local minimum but not the global minimum because 'it' has only large neighbors.
              LOUT_SIZE.add(1)
              (LOUT, v, mu)
            }
          }
        }

    }

    val lother = res_all.where($"_1" === LOTHERS).select($"_2", $"_3").as[(Long, Long)]
    val lout = res_all.where($"_1" === LOUT).select($"_2", $"_3").as[(Long, Long)]

    res_all.foreach{_=>}

    (lout, lother, NUM_CHANGES.value, LOUT_SIZE.value, LCC_SIZE.value, LIN_SIZE.value)

  }


  def smallStar(inDS: Dataset[(Long, Long)], numPartitions: Int): (Dataset[(Long, Long)],
    Dataset[(Long, Long)], Long, Long, Long) = {

    import inDS.sparkSession.implicits._
    val sc = inDS.sparkSession.sparkContext

    val NUM_CHANGES = sc.longAccumulator
    val SIN_SIZE = sc.longAccumulator
    val SOUT_SIZE = sc.longAccumulator


    val groupedRDD = inDS.flatMap{
      case (u, v) => Seq((u, v), (v, u))
    }.groupByKey{case (u, v) => u}.mapValues{case (u, v) => v}

    val SIN = true
    val SOUT = false
    val tmpPaths = sc.hadoopConfiguration.getTrimmedStrings("yarn.nodemanager.local-dirs")

    val res_all : Dataset[(Boolean, Long, Long)] = groupedRDD.flatMapGroups{
      case (u, uN) =>

        val mcu = Array.fill[Long](numPartitions)(Long.MaxValue)
        val up = (u % numPartitions).toInt
        mcu(up) = u

        var isLeaf = true

        val _uN_small = uN.filter { v =>

          if(v > u) isLeaf = false
          val vp = (v % numPartitions).toInt
          mcu(vp) = Math.min(v, mcu(vp))

          v < u
        }

        val uN_small = ExternalSorter.sort(_uN_small, tmpPaths)

        val mu = mcu.min

        val sout = uN_small filter {_ != mu} map { v =>

          val vp = (v % numPartitions).toInt

          NUM_CHANGES.add(1)
          SOUT_SIZE.add(1)

          if (v != mcu(vp))
            (SOUT, v, mcu(vp))
          else // v is mcu but not mu
            (SOUT, v, mu)

        }

        if (u != mcu(up)) {
          if (isLeaf) {
            SIN_SIZE.add(1)
            sout ++ Iterator((SIN, u, mcu(up)))
          }
          else {
            SOUT_SIZE.add(1)
            sout ++ Iterator((SOUT, u, mcu(up)))
          }
        }
        else if (u != mu) {
          SOUT_SIZE.add(1)
          sout ++ Iterator((SOUT, if (isLeaf) ~u else u, mu))
        }
        else sout
    }


    val sin = res_all.where($"_1" === SIN).select($"_2", $"_3").as[(Long, Long)]
    val sout = res_all.where($"_1" === SOUT).select($"_2", $"_3").as[(Long, Long)]

    res_all.foreach{x=>}

    (sout, sin, NUM_CHANGES.value, SOUT_SIZE.value, SIN_SIZE.value)

  }

  def ccComputation(remains: Dataset[(Long, Long)], numPartitions: Int): Dataset[(Long, Long)] = {

    val ss = remains.sparkSession

    import ss.implicits._

    val _res = remains.rdd.map{ case (u, v) => (if(u < 0) ~u else u, v) }
      .partitionBy(new CCPartitioner(numPartitions))
      .mapPartitions(edges => UnionFind.run(edges))

    ss.createDataset(_res)
  }

  class CCPartitioner(numParts: Int) extends Partitioner {

    def getPartition(key: Any): Int = {
      (key.asInstanceOf[Long] % numParts).toInt
    }

    override def numPartitions: Int = numParts
  }


}


