/**
  * PegasusG: Peta-Scale Graph Mining System v3.0
  * Authors: Chiwan Park, Ha-Myung Park, U Kang
  *
  * This software is licensed under Apache License, Version 2.0 (the  "License");
  * you may not use this file except in compliance with the License.
  * You may obtain a copy of the License at
  *
  * http://www.apache.org/licenses/LICENSE-2.0
  *
  * Unless required by applicable law or agreed to in writing, software
  * distributed under the License is distributed on an "AS IS" BASIS,
  * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  * See the License for the specific language governing permissions and
  * limitations under the License.
  * -------------------------------------------------------------------------
  * File: FilteringOpts.java
  * - Implicit class container for a 'filtered' method.
  * - Given an rdd[(Boolean, Long, Long)], 'filtered' filter out items whose
  *   first element is false. The items filtered out are saved in a separate
  *   file.
  * Version: 3.0
  */

package pegasus.spark.cc.utils

import org.apache.hadoop.conf.Configuration
import org.apache.hadoop.fs.Path
import org.apache.hadoop.io.{LongWritable, SequenceFile}
import org.apache.spark.rdd.RDD

object FilteringOps {
  implicit class FilteredRDD(rdd: RDD[(Boolean, Long, Long)]) {

    /**
      *
      * @param path the path of a directory where output files will be placed
      * @param prefix the prefix of the name of an output file for items filtered out
      * @param conf serializable hadoop configuration
      * @return rdd without false
      */
    def filtered(path: String, prefix: String, conf: SerializableConfiguration): RDD[(Long,Long)] ={

      rdd.mapPartitionsWithIndex{
        case (partitionId, it) =>
          val filePath = f"$path/$prefix-$partitionId%05d"
          println(filePath)
          it.filtered(filePath, conf.value)
      }
    }
  }

  implicit class FilteredIterator(it: Iterator[(Boolean, Long, Long)]) {
    def filtered(path: String, conf: Configuration)
    = new Iterator[(Long,Long)] {

      val writer: SequenceFile.Writer = SequenceFile.createWriter(conf,
        SequenceFile.Writer.keyClass(classOf[LongWritable]),
        SequenceFile.Writer.valueClass(classOf[LongWritable]),
        SequenceFile.Writer.file(new Path(path))
      )

      val aw = new LongWritable
      val bw = new LongWritable

      def close(){
        writer.close()
      }

      def pp(elem: (Boolean, Long, Long)): Boolean = {
        if(!elem._1){
          aw.set(elem._2)
          bw.set(elem._3)
          writer.append(aw, bw)
        }
        elem._1
      }

      private var hd: (Boolean, Long, Long) = _
      private var hdDefined: Boolean = false

      def hasNext: Boolean = hdDefined || {
        do {
          if (!it.hasNext){
            close()
            return false
          }
          hd = it.next()
        } while (!pp(hd))
        hdDefined = true
        true
      }

      def next(): (Long, Long) = if (hasNext) { hdDefined = false; (hd._2, hd._3) } else Iterator.empty.next()
    }
  }
}
