package pegasus.spark.common

import java.io.{BufferedReader, InputStreamReader}
import java.util.function.IntFunction

import org.apache.hadoop.fs.Path
import org.apache.spark.SparkContext
import org.apache.spark.rdd.RDD
import org.apache.spark.storage.StorageLevel
import pegasus.spark.common.io.{UnweightedMatrixReader, WeightedMatrixReader}
import pegasus.spark.common.partitioner.TaskPartitioner
import pegasus.spark.common.types.{MatrixBlock, UnweightedMatrixBlock, WeightedMatrixBlock}

object MatrixLoader {
  private def parseStatsLine(line: String): (Int, Int) = {
    val token = line.split("=")
    (token(0).substring(token(0).length - 5).toInt, token(1).toInt)
  }

  def loadSubproblems(sc: SparkContext, path: String): (Long, RDD[(Int, Seq[SubProblem])]) = {
    val statsPath = new Path(path, Constants.StatsPath)
    val fs = statsPath.getFileSystem(sc.hadoopConfiguration)
    val reader = new BufferedReader(new InputStreamReader(fs.open(statsPath)))
    try {
      val lines = reader.lines().toArray(new IntFunction[Array[String]] {
        override def apply(size: Int): Array[String] = Array.ofDim(size)
      })

      val numVertices = lines.filter(_.startsWith("numVertices")).map(_.split("=")(1).toLong).head
      val numSparseCols = lines.filter(_.startsWith("numSparseCols")).map(parseStatsLine).toMap
      val numSparseRows = lines.filter(_.startsWith("numSparseRows")).map(parseStatsLine).toMap
      val numDenseCols = lines.filter(_.startsWith("numDenseCols")).map(parseStatsLine).toMap
      val numDenseRows = lines.filter(_.startsWith("numDenseRows")).map(parseStatsLine).toMap
      val numBlocks = math.sqrt(numSparseCols.size).toInt

      val problems = (0 until numBlocks).map(taskId => {
        val sparseProblems = (0 until numBlocks).map(subTaskId => {
          val blockId = taskId + subTaskId * numBlocks
          SubProblem(path, isDense = false, blockId, numSparseCols(blockId), numSparseRows(blockId))
        })
        val denseProblems = (0 until numBlocks).map(subTaskId => {
          val blockId = taskId * numBlocks + subTaskId
          SubProblem(path, isDense = true, blockId, numDenseCols(blockId), numDenseRows(blockId))
        })

        (taskId, (sparseProblems ++ denseProblems).asInstanceOf[Seq[SubProblem]])
      })

      (numVertices, sc.parallelize(problems).partitionBy(new TaskPartitioner(numBlocks)).cache())
    } finally {
      reader.close()
    }
  }

  private def readUnweightedMatrixBlock(path: String, blockId: Int, numCols: Int, numRows: Int): MatrixBlock[Int] = {
    val reader = new UnweightedMatrixReader(path)
    val block = new UnweightedMatrixBlock(blockId, numCols, numRows)
    try {
      while (reader.hasNext) {
        val (col, deg, rows) = reader.next()
        block += (col, deg, rows)
      }
      block
    } finally {
      reader.close()
    }
  }

  private def readWeightedMatrixBlock(
    path: String,
    blockId: Int,
    numCols: Int,
    numRows: Int
  ): MatrixBlock[(Int, Double)] = {
    val reader = new WeightedMatrixReader(path)
    val block = new WeightedMatrixBlock(blockId, numCols, numRows)
    try {
      while (reader.hasNext) {
        val (col, deg, rows) = reader.next()
        block += (col, deg, rows)
      }

      block
    } finally {
      reader.close()
    }
  }

  def loadUnweightedMatrix(
    problemRDD: RDD[(Int, Seq[SubProblem])],
    storageLevel: StorageLevel = StorageLevel.MEMORY_AND_DISK
  ): (RDD[(Int, MatrixBlock[Int])], RDD[(Int, MatrixBlock[Int])]) = {
    val sparseRDD = problemRDD
      .mapPartitions(iter => {
        iter.flatMap(p => {
          p._2.filter(!_.isDense).map(block => {
            (p._1, readUnweightedMatrixBlock(block.matBlockPath, block.matBlockId, block.numCols, block.numRows))
          })
        })
      }, preservesPartitioning = true)
      .persist(storageLevel)
    val denseRDD = problemRDD
      .mapPartitions(iter => {
        iter.flatMap(p => {
          p._2.filter(_.isDense).map(block => {
            (p._1, readUnweightedMatrixBlock(block.matBlockPath, block.matBlockId, block.numCols, block.numRows))
          })
        })
      }, preservesPartitioning = true)
      .persist(storageLevel)

    sparseRDD.foreachPartition(_ => ())
    denseRDD.foreachPartition(_ => ())

    (sparseRDD, denseRDD)
  }

  def loadWeightedMatrix(
    problemRDD: RDD[(Int, Seq[SubProblem])],
    storageLevel: StorageLevel = StorageLevel.MEMORY_AND_DISK
  ): (RDD[(Int, MatrixBlock[(Int, Double)])], RDD[(Int, MatrixBlock[(Int, Double)])]) = {
    val sparseRDD = problemRDD
      .mapPartitions(iter => {
        iter.flatMap(p => {
          p._2.filter(!_.isDense).map(block => {
            (p._1, readWeightedMatrixBlock(block.matBlockPath, block.matBlockId, block.numCols, block.numRows))
          })
        })
      }, preservesPartitioning = true)
      .persist(storageLevel)
    val denseRDD = problemRDD
      .mapPartitions(iter => {
        iter.flatMap(p => {
          p._2.filter(_.isDense).map(block => {
            (p._1, readWeightedMatrixBlock(block.matBlockPath, block.matBlockId, block.numCols, block.numRows))
          })
        })
      }, preservesPartitioning = true)
      .persist(storageLevel)

    sparseRDD.foreachPartition(_ => ())
    denseRDD.foreachPartition(_ => ())

    (sparseRDD, denseRDD)
  }
}
