package pegasus.spark.common

import org.apache.spark.rdd.RDD
import org.apache.spark.storage.StorageLevel
import pegasus.spark.common.io.UnitVectorReader
import pegasus.spark.common.types.{BitSet, MatrixBlock, VectorBlock}

object MultiplicationRunner {
  def run[M, V, PV](
    multiplication: Multiplication[M, V, PV],
    problemRDD: RDD[(Int, Seq[SubProblem])],
    sparseRDD: RDD[(Int, MatrixBlock[M])],
    denseRDD: RDD[(Int, MatrixBlock[M])],
    numVertices: Long,
    maxIterations: Int,
    tempPath: String
  ): RDD[(Long, V)] = {
    val sc = problemRDD.sparkContext
    val numBlocks = problemRDD.count().toInt
    var converged = false
    var nIter = 0

    // build vector RDD
    var vectorRDD = initializeVectorBlocks(multiplication, problemRDD, numVertices, tempPath)

    while ((maxIterations == -1 || nIter < maxIterations) && !converged) {
      val convergedAccum = sc.longAccumulator(f"${Constants.ConvergedAccumulator}-$nIter%05d")
      val context = new MultiplicationContext(nIter, numVertices, numBlocks, convergedAccum)

      // combine2 for sparse matrix blocks
      vectorRDD.zipPartitions(sparseRDD)((vecIter, matIter) => {
        val vecBlock = vecIter.next()._2
        val intermediateBlock = multiplication.createIntermediateVectorBlock(numVertices, numBlocks, 0, context)
        while (matIter.hasNext) {
          val (blockId, matBlock) = matIter.next()
          intermediateBlock.setBlockId(blockId / numBlocks)
          multiplication.combine2Op(matBlock, vecBlock, intermediateBlock, context)

          val intermediateIter = intermediateBlock.iterator(true)
          val vecWriter = multiplication.createIntermediateVectorWriter(tempPath, matBlock.blockId)
          try {
            while (intermediateIter.next()) {
              vecWriter.write(intermediateIter.currentId, intermediateIter.currentValue)
            }
          } finally {
            vecWriter.close()
          }
          intermediateBlock.clear()
        }

        Seq(0).iterator
      }).foreachPartition(_ => ())

      // combineAll for sparse matrix blocks
      val intermediateSparse = problemRDD.mapPartitions(iter => {
        val blockId = iter.next()._1
        val block = multiplication.createIntermediateVectorBlock(numVertices, numBlocks, blockId, context)
        (0 until numBlocks).foreach(i => {
          val reader = multiplication.createIntermediateVectorReader(tempPath, blockId * numBlocks + i)
          try {
            multiplication.combineAllOp(reader, block, context)
          } finally {
            reader.close()
          }
        })

        Seq((blockId, block)).iterator
      })

      // combine for dense matrix blocks
      val intermediateVec = intermediateSparse.zipPartitions(denseRDD, preservesPartitioning = true)(
        (pvecIter, matIter) => {
          val (vecBlockId, pvecBlock) = pvecIter.next()
          val denseVec = multiplication.createVectorBlock(0, numVertices, numBlocks, init = false, context)
          while (matIter.hasNext) {
            val (blockId, matBlock) = matIter.next()
            denseVec.clear()
            val reader = multiplication.createVectorReader(tempPath, blockId % numBlocks)
            try {
              while (reader.hasNext) {
                val (id, value) = reader.next()
                denseVec.put(id, value)
              }
            } finally {
              reader.close()
            }

            multiplication.combine2Op(matBlock, denseVec, pvecBlock, context)
          }

          Seq((vecBlockId, pvecBlock)).iterator
        }
      )

      // apply for vector blocks
      val newVectorRDD = vectorRDD.zipPartitions(intermediateVec, preservesPartitioning = true)((vecIter, iVecIter) => {
        val (blockId, vecBlock, denseSet) = vecIter.next()
        val (_, iVecBlock) = iVecIter.next()
        multiplication.applyOp(vecBlock, iVecBlock, context)
        Seq((blockId, vecBlock, denseSet)).iterator
      }).persist(StorageLevel.MEMORY_AND_DISK)

      newVectorRDD.foreachPartition(_ => ())
      vectorRDD.unpersist()

      converged = convergedAccum.isZero
      if (!converged) {
        writeDenseVectorBlocks(multiplication, newVectorRDD, tempPath)
      }
      vectorRDD = newVectorRDD

      nIter += 1
    }

    vectorRDD.mapPartitions(iter => {
      val (blockId, block, _) = iter.next()
      val blockIter = block.iterator(false)
      new Iterator[(Long, V)] {
        private var hdDefined = blockIter.next()

        override def hasNext: Boolean = hdDefined

        override def next(): (Long, V) = {
          val record = (PMVUtils.localToGlobal(numBlocks, blockId, blockIter.currentId), blockIter.currentValue)
          hdDefined = blockIter.next()
          record
        }
      }
    })
  }

  private def initializeVectorBlocks[M, V, PV](
    multiplication: Multiplication[M, V, PV],
    problemRDD: RDD[(Int, Seq[SubProblem])],
    numVertices: Long,
    tempPath: String
  ) = {
    val numBlocks = problemRDD.count().toInt
    val dummyContext = new MultiplicationContext(0, numVertices, numBlocks, null)

    // initialize vector blocks
    val vectorRDD = problemRDD
      .mapPartitions(iter => {
        val (blockId, subproblems) = iter.next()
        val denseSetPath = f"${subproblems.head.rootPath}/${Constants.DenseVectorBlock}-$blockId%05d"
        val block = multiplication.createVectorBlock(blockId, numVertices, numBlocks, init = true, dummyContext)
        val denseSet = new BitSet(block.size)
        val dsReader = new UnitVectorReader(denseSetPath)
        try {
          while (dsReader.hasNext) {
            val (id, _) = dsReader.next()
            denseSet.add(id)
          }
        } finally {
          dsReader.close()
        }
        Seq((blockId, block, denseSet)).iterator
      }, preservesPartitioning = true)
      .persist(StorageLevel.MEMORY_AND_DISK)

    // write dense vector blocks
    writeDenseVectorBlocks(multiplication, vectorRDD, tempPath)

    vectorRDD
  }

  private def writeDenseVectorBlocks[M, V, PV](
    multiplication: Multiplication[M, V, PV],
    vectorRDD: RDD[(Int, VectorBlock[V], BitSet)],
    tempPath: String
  ): Unit = {
    vectorRDD.foreachPartition(iter => {
      iter.foreach(v => {
        val (blockId, vecBlock, denseSet) = v
        val writer = multiplication.createVectorWriter(tempPath, blockId)
        try {
          val iter = vecBlock.iterator(onlyNonZero = false, denseSet)
          while (iter.next()) {
            writer.write(iter.currentId, iter.currentValue)
          }
        } finally {
          writer.close()
        }
      })
    })
  }
}
