package pegasus.spark.common

import java.io.File
import java.nio.ByteBuffer
import java.util.concurrent.ConcurrentHashMap

import org.apache.hadoop.conf.Configuration
import org.apache.hadoop.io.{ByteBufferPool, ElasticByteBufferPool}
import org.apache.spark.deploy.SparkHadoopUtil
import org.apache.spark.io.{CompressionCodec, LZ4CompressionCodec, LZFCompressionCodec, SnappyCompressionCodec}
import org.apache.spark.{SparkConf, SparkEnv}

object PMVUtils {
  private val JavaRuntime = Runtime.getRuntime
  private val byteBufferPool: ByteBufferPool = new ElasticByteBufferPool

  def callGCMultipleTimes(): Unit = callGCMultipleTimes(3)

  def callGCMultipleTimes(times: Int): Unit = {
    var i = 0
    while (i < times) {
      JavaRuntime.gc()
      i += 1
    }
  }

  def maxMemoryBytes: Long = JavaRuntime.maxMemory

  def availableMemoryBytes: Long = JavaRuntime.maxMemory - JavaRuntime.totalMemory + JavaRuntime.freeMemory

  def obtainByteBuffer(size: Int): ByteBuffer = byteBufferPool.getBuffer(true, size)

  def returnByteBuffer(buf: ByteBuffer): Unit = byteBufferPool.putBuffer(buf)

  def localDirs(): Array[String] = {
    val conf = sparkConf()
    if (conf == null) {
      Array(System.getProperty("java.io.tmpdir", "/tmp"))
    } else if (System.getenv("CONTAINER_ID") != null) {
      System.getenv("LOCAL_DIRS").split(",")
    } else if (System.getenv("SPARK_EXECUTOR_DIRS") != null) {
      System.getenv("SPARK_EXECUTOR_DIRS").split(File.pathSeparator)
    } else if (System.getenv("SPARK_LOCAL_DIRS") != null) {
      System.getenv("SPARK_LOCAL_DIRS").split(",")
    } else {
      conf.get("spark.local.dir", System.getProperty("java.io.tmpdir")).split(",")
    }
  }

  def sparkConf(): SparkConf = Option(SparkEnv.get).map(_.conf).getOrElse(new SparkConf())

  def hadoopConf(): Configuration = SparkHadoopUtil.get.conf

  private val codecPool = new ConcurrentHashMap[String, CompressionCodec]()

  def compressionCodec(): CompressionCodec = {
    val conf = sparkConf()
    val codecName = conf.get("spark.io.compression.codec", "lz4")
    Option(codecPool.get(codecName)) match {
      case Some(codec) => codec
      case None =>
        val codec = codecName match {
          case "lz4" => new LZ4CompressionCodec(conf)
          case "lzf" => new LZFCompressionCodec(conf)
          case "snappy" => new SnappyCompressionCodec(conf)
        }
        codecPool.put(codecName, codec)

        codec
    }
  }

  def globalToLocal(numBlocks: Int, globalId: Long): Int = (globalId / numBlocks).toInt

  def localToGlobal(numBlocks: Int, blockId: Int, localId: Int): Long = localId.toLong * numBlocks + blockId
}
