package pegasus.spark.common.io

import it.unimi.dsi.fastutil.ints.AbstractIntComparator
import it.unimi.dsi.fastutil.{Arrays, Swapper}
import pegasus.spark.common.PMVUtils

import scala.collection.mutable
import scala.reflect.ClassTag
import scala.reflect.io.File

abstract class ExternalGrouper[@specialized(Int, Long, Double) T: ClassTag](
  taskId: Int, sizePerEntry: Int, numCores: Int
) {
  protected var bufferSize = 0
  protected var current = 0

  // we use hard-corded lazy initialization for values to avoid duplicated initialization.
  // this is a bug of scala compiler. If the bug is fixed, we can change these fields to `protected val`.
  protected var keys: Array[Int] = _
  protected var values: Array[T] = _

  protected def init(specializedValue: T): Unit = {
    PMVUtils.callGCMultipleTimes()
    bufferSize = (PMVUtils.availableMemoryBytes / (sizePerEntry + 8) / numCores * 0.75).toInt
    keys = Array.ofDim[Int](bufferSize)
    values = Array.ofDim[T](bufferSize)
  }

  private var currentFile = 0
  private val files = mutable.Queue[Int]()
  private val tempRoot = {
    val localDirs = PMVUtils.localDirs()
    localDirs(taskId % localDirs.length)
  }

  private val comparator = new AbstractIntComparator {
    override def compare(left: Int, right: Int): Int = {
      if (keys(left) < keys(right)) -1
      else if (keys(left) > keys(right)) 1
      else 0
    }
  }
  protected val swapper = new Swapper {
    override def swap(left: Int, right: Int): Unit = {
      val tempKey = keys(left)
      keys(left) = keys(right)
      keys(right) = tempKey

      // we use an external method for swapping values to avoid unnecessary boxing and unboxing.
      updateValue(left, values(right), right, values(left))
    }
  }

  private def updateValue(left: Int, leftValue: T, right: Int, rightValue: T): Unit = {
    values(left) = leftValue
    values(right) = rightValue
  }

  protected def buildValueReader(path: String): FileReader[T]

  protected def buildValueWriter(path: String): FileWriter[T]

  def append(key: Int, value: T): Unit = {
    if (current == bufferSize) {
      flush()
    }

    keys(current) = key
    values(current) = value
    current += 1
  }

  def flush(): Unit = {
    if (current > 0) {
      // sort indices
      Arrays.quickSort(0, current, comparator, swapper)

      // open files
      val keyWriter = new IntFileWriter(keyFilePath(currentFile))
      val countWriter = new IntFileWriter(countFilePath(currentFile))
      val valueWriter = buildValueWriter(valueFilePath(currentFile))

      // spill to disk
      var i = 0
      var count = 0
      var key = keys(i)
      while (i < current) {
        if (key == keys(i)) {
          count += 1
        } else {
          keyWriter.write(key)
          countWriter.write(count)
          key = keys(i)
          count = 1
        }
        valueWriter.write(values(i))
        i += 1
      }
      keyWriter.write(key)
      countWriter.write(count)

      // close files
      keyWriter.close()
      countWriter.close()
      valueWriter.close()

      files += currentFile
      current = 0
      currentFile += 1
    }
  }

  def groupedIterator(): Iterator[(Int, Int, Array[T])] = {
    if (files.isEmpty) {
      // no spill files
      Arrays.quickSort(0, current, comparator, swapper)
      new MemoryGroupedIterator[T](keys, values, current)
    } else {
      flush()

      while (files.size > 1) {
        val bucket1 = files.dequeue()
        val bucket2 = files.dequeue()
        val newBucket = currentFile
        files += newBucket
        currentFile += 1

        merge(bucket1, bucket2, newBucket)

        File(keyFilePath(bucket1)).delete()
        File(countFilePath(bucket1)).delete()
        File(valueFilePath(bucket1)).delete()
        File(keyFilePath(bucket2)).delete()
        File(countFilePath(bucket2)).delete()
        File(valueFilePath(bucket2)).delete()
      }

      val lastBucket = files.dequeue()
      new FileGroupedIterator[T](
        keyFilePath(lastBucket), countFilePath(lastBucket), buildValueReader(valueFilePath(lastBucket)))
    }
  }

  private def copy(
    keyReader: BufferedReader[Int], countReader: FileReader[Int], valueReader: FileReader[T],
    keyWriter: FileWriter[Int], countWriter: FileWriter[Int], valueWriter: FileWriter[T]) = {
    // write key
    keyWriter.write(keyReader.next())

    // write count
    val count = countReader.next()
    countWriter.write(count)

    // write values
    var i = 0
    while (i < count) {
      valueWriter.write(valueReader.next())
      i += 1
    }
  }

  private def merge(bucket1: Int, bucket2: Int, outBucket: Int) = {
    val key1Reader = new BufferedReader[Int](new IntFileReader(keyFilePath(bucket1)))
    val count1Reader = new IntFileReader(countFilePath(bucket1))
    val value1Reader = buildValueReader(valueFilePath(bucket1))

    val key2Reader = new BufferedReader[Int](new IntFileReader(keyFilePath(bucket2)))
    val count2Reader = new IntFileReader(countFilePath(bucket2))
    val value2Reader = buildValueReader(valueFilePath(bucket2))

    val keyWriter = new IntFileWriter(keyFilePath(outBucket))
    val countWriter = new IntFileWriter(countFilePath(outBucket))
    val valueWriter = buildValueWriter(valueFilePath(outBucket))

    while (key1Reader.hasNext && key2Reader.hasNext) {
      val key1 = key1Reader.head
      val key2 = key2Reader.head

      if (key1 < key2) {
        copy(key1Reader, count1Reader, value1Reader, keyWriter, countWriter, valueWriter)
      } else if (key1 > key2) {
        copy(key2Reader, count2Reader, value2Reader, keyWriter, countWriter, valueWriter)
      } else {
        // merge two keys
        keyWriter.write(key1Reader.next())
        key2Reader.next()

        val count1 = count1Reader.next()
        val count2 = count2Reader.next()
        countWriter.write(count1 + count2)

        var i = 0
        while (i < count1) {
          valueWriter.write(value1Reader.next())
          i += 1
        }
        i = 0
        while (i < count2) {
          valueWriter.write(value2Reader.next())
          i += 1
        }
      }
    }

    // copy remaining records
    while (key1Reader.hasNext) {
      copy(key1Reader, count1Reader, value1Reader, keyWriter, countWriter, valueWriter)
    }
    while (key2Reader.hasNext) {
      copy(key2Reader, count2Reader, value2Reader, keyWriter, countWriter, valueWriter)
    }

    // close readers
    key1Reader.close()
    count1Reader.close()
    value1Reader.close()
    key2Reader.close()
    count2Reader.close()
    value2Reader.close()

    // close writers
    keyWriter.close()
    countWriter.close()
    valueWriter.close()
  }

  private def keyFilePath(id: Int) = (File(tempRoot) / f"pmv-group-key-$taskId-$id%05d").toString()

  private def countFilePath(id: Int) = (File(tempRoot) / f"pmv-group-count-$taskId-$id%05d").toString()

  private def valueFilePath(id: Int) = (File(tempRoot) / f"pmv-group-value-$taskId-$id%05d").toString()
}
