package pegasus.spark.common.types

import com.esotericsoftware.kryo.io.{Input, Output}
import com.esotericsoftware.kryo.{Kryo, Serializer}

class UnweightedMatrixBlock(
  override val blockId: Int,
  override val numCols: Int,
  override val numRows: Int
) extends MatrixBlock[Int] {
  private var lastCol = 0
  private var lastRow = 0
  private val cols = Array.ofDim[Int](numCols)
  private val degs = Array.ofDim[Int](numCols)
  private val rows = Array.ofDim[Int](numRows)
  private val rowSizes = Array.ofDim[Int](numCols)

  private[UnweightedMatrixBlock] def setMetadata(lastCol: Int, lastRow: Int): Unit = {
    this.lastCol = lastCol
    this.lastRow = lastRow
  }

  private[UnweightedMatrixBlock] def getArrays = (cols, degs, rows, rowSizes)

  override def +=(col: Int, deg: Int, rows: Array[Int]): Unit = {
    cols(lastCol) = col
    degs(lastCol) = deg
    Array.copy(rows, 0, this.rows, lastRow, rows.length)
    lastRow += rows.length
    rowSizes(lastCol) = lastRow
    lastCol += 1
  }

  override def iterator(): MatrixIterator[Int] = {
    new MatrixIterator[Int] {
      private var currPos = 0
      private var currCol = 0
      private var currDeg = 0
      private val currRows = new UnweightedMatrixRows()

      override def next(): Boolean = {
        if (currPos < lastCol) {
          currCol = cols(currPos)
          currDeg = degs(currPos)

          val begin = if (currPos == 0) 0 else rowSizes(currPos - 1)
          val end = rowSizes(currPos)
          currRows.setRows(rows, begin, end)

          currPos += 1
          true
        } else {
          false
        }
      }

      override def currentCol(): Int = currCol

      override def currentDeg(): Int = currDeg

      override def currentRows(): MatrixRows[Int] = currRows
    }
  }
}

object UnweightedMatrixBlock {

  class BlockSerializer extends Serializer[UnweightedMatrixBlock] {
    override def read(kryo: Kryo, in: Input, `type`: Class[UnweightedMatrixBlock]): UnweightedMatrixBlock = {
      val (blockId, numCols, numRows) = (in.readInt(), in.readInt(), in.readInt())
      val block = new UnweightedMatrixBlock(blockId, numCols, numRows)
      block.setMetadata(numCols, numRows)
      val (cols, degs, rows, rowSizes) = block.getArrays

      var i = 0
      while (i < numCols) {
        cols(i) = in.readInt()
        degs(i) = in.readInt()
        rowSizes(i) = in.readInt()
        i += 1
      }

      i = 0
      while (i < numRows) {
        rows(i) = in.readInt()
        i += 1
      }

      block
    }

    override def write(kryo: Kryo, out: Output, block: UnweightedMatrixBlock): Unit = {
      out.writeInt(block.blockId)
      out.writeInt(block.numCols)
      out.writeInt(block.numRows)
      val (cols, degs, rows, rowSizes) = block.getArrays

      var i = 0
      while (i < block.numCols) {
        out.writeInt(cols(i))
        out.writeInt(degs(i))
        out.writeInt(rowSizes(i))
        i += 1
      }

      i = 0
      while (i < block.numRows) {
        out.writeInt(rows(i))
        i += 1
      }
    }
  }
}
