package pegasus.spark.degree

import org.apache.spark.{SparkConf, SparkContext}
import pegasus.spark.common.serialization.PMVRegistrator
import pegasus.spark.common.{Computation, GraphLoader}

object DegreeComputation extends Computation {

  object Direction extends Enumeration {
    type Direction = Value

    val In, Out, Both = Value

    def fromString(str: String): Direction = str.toLowerCase match {
      case "in" => In
      case "out" => Out
      case "both" => Both
    }
  }

  import Direction.Direction

  case class Config(
    input: String = "",
    output: String = "",
    direction: Direction = null)

  private val parser = new scopt.OptionParser[Config]("Degree") {
    head("Degree")

    arg[String]("input")
      .action((x, c) => c.copy(input = x))

    arg[String]("output")
      .action((x, c) => c.copy(output = x))

    arg[String]("direction")
      .validate(x => {
        x.toLowerCase match {
          case "in" | "out" | "both" => success
          case _ => failure("direction must be one of in, out, both")
        }
      })
      .action((x, c) => c.copy(direction = Direction.fromString(x)))
  }

  def run(sc: SparkContext, config: Config): Unit = {
    val edges = GraphLoader.edgeListFile(sc, config.input)

    val vertices = config.direction match {
      case Direction.In => edges.map(e => (e._2, 1))
      case Direction.Out => edges.map(e => (e._1, 1))
      case Direction.Both => edges.mapPartitions(iter => {
        iter.flatMap(e => {
          if (e._1 == e._2) Seq((e._1, 1))
          else Seq((e._1, 1), (e._2, 1))
        })
      })
    }

    val degree = vertices.reduceByKey(_ + _)

    degree.map(v => s"${v._1}\t${v._2}").saveAsTextFile(config.output)
  }

  def main(args: Array[String]): Unit = {
    parser.parse(args, Config()) match {
      case Some(config) =>
        val conf = new SparkConf()
          .setAppName("PMV - PageRank")
          .set("spark.serializer", "org.apache.spark.serializer.KryoSerializer")
          .set("spark.kryo.registrator", classOf[PMVRegistrator].getName)
          .set("spark.kryoserializer.buffer", "24m")
        val sc = new SparkContext(conf)

        run(sc, config)

        sc.stop()
      case _ => System.exit(-1)
    }
  }
}
