package pegasus.spark.diameter

import org.apache.spark.{SparkConf, SparkContext}
import pegasus.spark.common._
import pegasus.spark.common.io._
import pegasus.spark.common.serialization.PMVRegistrator
import pegasus.spark.common.types.{BitStrings, BitStringsVectorBlock, MatrixBlock, VectorBlock}

import scala.annotation.tailrec
import scala.util.Random

object DiameterComputation extends Computation {

  case class Config(
    input: String = "",
    output: String = "",
    tempPath: Option[String] = None,
    numBlocks: Int = 40,
    denseThreshold: Int = 150,
    fieldDelimiter: String = "\t",
    maxIterations: Int = -1)

  private val parser = new scopt.OptionParser[Config]("diameter") {
    head("diameter")

    opt[Int]("numBlocks").abbr("b")
      .action((x, c) => c.copy(numBlocks = x))

    opt[Int]("denseThreshold").abbr("t")
      .action((x, c) => c.copy(denseThreshold = x))

    opt[Int]("maxIterations").abbr("i")
      .action((x, c) => c.copy(maxIterations = x))

    opt[String]("fieldDelimiter").abbr("d")
      .action((x, c) => c.copy(fieldDelimiter = x))

    opt[String]("temporaryPath").abbr("tp")
      .action((x, c) => c.copy(tempPath = Some(x)))

    arg[String]("input")
      .action((x, c) => c.copy(input = x))

    arg[String]("output")
      .action((x, c) => c.copy(output = x))
  }

  private def multiplication(cardinalityPath: String) = new Multiplication[Int, BitStrings, BitStrings] {
    private val random = new Random
    private val initValue = new BitStrings

    private def initBitStrings(id: Long): BitStrings = {
      @tailrec
      def setBit(k: Int, i: Int, prob: Double, threshold: Double): Unit = {
        if (i < 64) {
          val nprob = prob + math.pow(2, -i - 1)
          if (prob > threshold) initValue.setBit(k, 1L << (63 - i))
          else setBit(k, i + 1, nprob, threshold)
        }
      }

      random.setSeed(id)

      var k = 0
      while (k < Constants.DefaultNumBitstrings) {
        setBit(k, 0, 0.0, random.nextDouble())
        k += 1
      }

      initValue
    }

    override def createVectorBlock(
      blockId: Int,
      numVertices: Long,
      numBlocks: Int,
      init: Boolean,
      context: MultiplicationContext
    ): VectorBlock[BitStrings] = {
      val block = new BitStringsVectorBlock(numBlocks, numVertices, blockId)
      val size = block.size
      var i = 0
      while (i < size) {
        val globalId = PMVUtils.localToGlobal(numBlocks, blockId, i)
        block.put(i, initBitStrings(globalId))
        i += 1
      }

      block
    }

    override def createVectorReader(vectorPath: String, blockId: Int): VectorReader[BitStrings] = {
      new BitStringsVectorReader(f"$vectorPath/${Constants.DenseVectorBlock}-$blockId%05d")
    }

    override def createVectorWriter(vectorPath: String, blockId: Int): VectorWriter[BitStrings] = {
      new BitStringsVectorWriter(f"$vectorPath/${Constants.DenseVectorBlock}-$blockId%05d")
    }

    override def createIntermediateVectorBlock(
      numVertices: Long,
      numBlocks: Int,
      blockId: Int,
      context: MultiplicationContext
    ): VectorBlock[BitStrings] = {
      new BitStringsVectorBlock(numBlocks, numVertices, blockId)
    }

    override def createIntermediateVectorReader(vectorPath: String, blockId: Int): VectorReader[BitStrings] = {
      new BitStringsVectorReader(f"$vectorPath/${Constants.IntermediateVectorBlock}-$blockId%05d")
    }

    override def createIntermediateVectorWriter(vectorPath: String, blockId: Int): VectorWriter[BitStrings] = {
      new BitStringsVectorWriter(f"$vectorPath/${Constants.IntermediateVectorBlock}-$blockId%05d")
    }

    override def combine2Op(
      matrix: MatrixBlock[Int],
      vector: VectorBlock[BitStrings],
      result: VectorBlock[BitStrings],
      context: MultiplicationContext): Unit = {
      val numBlocks = context.numBlocks
      val srcBlockId = matrix.blockId % numBlocks
      val dstBlockId = matrix.blockId / numBlocks
      val selfBlock = srcBlockId == dstBlockId

      val iter = matrix.iterator()
      while (iter.next()) {
        val col = iter.currentCol()
        val rows = iter.currentRows()
        val value = if (vector.contains(col)) {
          vector.get(col)
        } else {
          initBitStrings(PMVUtils.localToGlobal(numBlocks, srcBlockId, col))
        }

        var i = 0
        val size = rows.size()
        while (i < size) {
          result.addTo(rows.row(i), value)
          i += 1
        }
      }

      if (selfBlock) {
        val vecIter = vector.iterator(false)
        while (vecIter.next()) {
          result.addTo(vecIter.currentId, vecIter.currentValue)
        }
      }
    }

    override def combineAllOp(
      input: VectorReader[BitStrings],
      output: VectorBlock[BitStrings],
      context: MultiplicationContext
    ): Unit = {
      while (input.hasNext) {
        val (id, value) = input.next()
        output.addTo(id, value)
      }
    }

    override def applyOp(
      result: VectorBlock[BitStrings],
      intermediate: VectorBlock[BitStrings],
      context: MultiplicationContext
    ): Unit = {
      val writer = new CardinalityWriter(cardinalityPath, result.blockId)
      val currIter = context.currentIter
      val iter = result.iterator(false)
      while (iter.next()) {
        val id = iter.currentId
        val oldValue = iter.currentValue
        val newValue = intermediate.get(id)

        if (newValue.isEmpty || oldValue == newValue) {
          val globalId = PMVUtils.localToGlobal(context.numBlocks, result.blockId, id)
          writer.write(globalId, currIter, oldValue.cardinality())
        } else {
          context.notifyChanged()
          result.put(id, newValue)
        }
      }
      writer.close()
    }
  }

  def run(sc: SparkContext, config: Config): Unit = {
    val partitionedPath = getPartitionedPath(config.output, config.tempPath)
    val intermediatePath = getIntermediatePath(config.output, config.tempPath)
    val cardinalityPath = s"$intermediatePath/cardinality"
    val edges = GraphLoader.edgeListFile(sc, config.input, config.fieldDelimiter)

    Prepartitioning.unweightedMatrix(edges, config.numBlocks, config.denseThreshold, partitionedPath)

    val (numVertices, problemRDD) = MatrixLoader.loadSubproblems(sc, partitionedPath)
    val (sparseRDD, denseRDD) = MatrixLoader.loadUnweightedMatrix(problemRDD)

    MultiplicationRunner.run(
      multiplication(cardinalityPath),
      problemRDD,
      sparseRDD,
      denseRDD,
      numVertices,
      config.maxIterations,
      intermediatePath
    )

    val cardinalities = sc.parallelize(0 until config.numBlocks, config.numBlocks).flatMap(blockId => {
      val reader = new CardinalityReader(cardinalityPath, blockId)
      new Iterator[(Long, (Int, Double))] {
        private var record: (Long, Int, Double) = _
        private var defined = false
        private var closed = false

        override def hasNext: Boolean = {
          if (!closed && !defined) {
            defined = reader.next()
            if (!defined) {
              reader.close()
            } else {
              record = reader.currentRecord
            }
          }

          defined
        }

        override def next(): (Long, (Int, Double)) = {
          if (!hasNext) throw new NoSuchElementException
          defined = false
          (record._1, (record._2, record._3))
        }
      }
    }).groupByKey().mapPartitions(iter => {
      val cardinalities = Array.ofDim[Double](Constants.MaxIterations)
      val sumCardinalities = Array.ofDim[Double](Constants.MaxIterations)
      iter.map(v => {
        val (id, values) = v
        var maxHop = Int.MinValue

        // collect cardinalities
        cardinalities.indices.foreach(i => cardinalities(i) = 0)
        values.foreach(cd => {
          cardinalities(cd._1 - 1) = cd._2
          maxHop = math.max(maxHop, cd._1)
        })
        (maxHop until Constants.MaxIterations).foreach(i => cardinalities(i) = cardinalities(i - 1))

        // estimate effective radii for each vertex
        val threshold = cardinalities(Constants.MaxIterations - 1)
        val radii = (0 until Constants.MaxIterations).filter(i => cardinalities(i) >= threshold).min

        (0 until Constants.MaxIterations).foreach(i => sumCardinalities(i) += cardinalities(i))

        (id, radii)
      })
    })

    cardinalities.map(v => s"${v._1}\t${v._2}").saveAsTextFile(config.output)
  }

  def main(args: Array[String]): Unit = {
    parser.parse(args, Config()) match {
      case Some(config) =>
        val conf = new SparkConf()
          .setAppName("PMV - PageRank")
          .set("spark.serializer", "org.apache.spark.serializer.KryoSerializer")
          .set("spark.kryo.registrator", classOf[PMVRegistrator].getName)
          .set("spark.kryoserializer.buffer", "24m")
        val sc = new SparkContext(conf)

        run(sc, config)

        sc.stop()
      case _ => System.exit(-1)
    }
  }
}
