/**
  * PegasusN: Peta-Scale Graph Mining System (Pegasus v3.0)
  * Authors: Chiwan Park, Ha-Myung Park, U Kang
  *
  * This software is licensed under Apache License, Version 2.0 (the  "License");
  * you may not use this file except in compliance with the License.
  * You may obtain a copy of the License at
  *
  * http://www.apache.org/licenses/LICENSE-2.0
  *
  * Unless required by applicable law or agreed to in writing, software
  * distributed under the License is distributed on an "AS IS" BASIS,
  * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  * See the License for the specific language governing permissions and
  * limitations under the License.
  * -------------------------------------------------------------------------
  * File: KroneckerGenerator.java
  * - generate a kronecker graph
  * Version: 3.0
  */

package pegasus.spark.graphgen

import java.util.StringTokenizer

import org.apache.log4j.Logger
import org.apache.spark.rdd.RDD
import org.apache.spark.{HashPartitioner, SparkConf, SparkContext}

import scala.collection.mutable
import scala.io.Source

object KroneckerGenerator {

  private val logger = Logger.getLogger(KroneckerGenerator.getClass)

  def main(args: Array[String]): Unit = {

    val level = args(0).toInt
    val numTasks = args(1).toInt
    val seedGraphFile = args(2)
    val outputPath = args(3)

    logger.info(f"lv: $level, numTasks: $numTasks, seedGraphFile: $seedGraphFile")

    val conf = new SparkConf().setAppName(f"[Kronecker] lv: $level, numTasks: $numTasks")

    val sc = new SparkContext(conf)

    val seedGraph = parseSeedGraph(seedGraphFile)

    run(level, seedGraph, numTasks, sc).map{x => x._1 + "\t" + x._2}.saveAsTextFile(outputPath)

    sc.stop()
  }

  /**
    * read seed graph.
    *
    * @param seedGraphPath seed graph file path
    * @return the array of edges in the seed graph
    */
  def parseSeedGraph(seedGraphPath: String): Array[(Int, Int)] ={

    Source.fromFile(seedGraphPath).getLines().map{ line =>
      val st = new StringTokenizer(line)
      val u = st.nextToken().toInt
      val v = st.nextToken().toInt
      (u, v)
    }.toArray

  }

  /**
    * submit the spark job.
    *
    * @param level of recursion
    * @param seedGraph array of edges of the seed graph
    * @param numTasks degree of parallelism (the number of tasks to run)
    * @param sc spark context
    * @return an RDD containing edges of the generated graph
    */
  def run(level: Int, seedGraph: Array[(Int, Int)], numTasks: Int, sc: SparkContext): RDD[(Long, Long)] = {

    val numSeedEdges = seedGraph.length

    val numEdges = Math.pow(numSeedEdges, level).toLong

    val regionQueue = new mutable.PriorityQueue[Region]()(Ordering.by(regionOrder))
    val regionSlotQueue = new mutable.PriorityQueue[RegionSlot]()(Ordering.by(regionSlotOrder))

    for (i <- 0 until numTasks) {
      val numEdgeAssigned = (numEdges * (i + 1) / numTasks) - (numEdges * i / numTasks)
      regionSlotQueue.enqueue(RegionSlot(numEdgeAssigned, mutable.Buffer[Region]()))
    }

    regionQueue.enqueue(Region(0, 0, level - 1, numEdges))

    while (regionQueue.nonEmpty) {
      val region = regionQueue.dequeue
      val slot = regionSlotQueue.dequeue

      if(region.level == 0 || region.numEdges <= slot.edgeCapacity){
        slot.add(region)
      }
      else{
        splitRegion(region, seedGraph).foreach(r => regionQueue.enqueue(r))
      }

      regionSlotQueue.enqueue(slot)
    }

    val regionSlotRDD = sc.parallelize(regionSlotQueue.toSeq.zipWithIndex.map{case (x,y) => (y,x)})
      .partitionBy(new HashPartitioner(numTasks))

    regionSlotRDD.flatMap{ case (_, x) =>
      x.regions.toStream.flatMap(r => generate(r, seedGraph))
    }

  }


  /**
    * generate edges within a given region
    * @param region to process
    * @param seedGraph array of edges of the seed graph
    * @return list of edges in the region
    */
  def generate(region: Region, seedGraph: Array[(Int, Int)]): Stream[(Long, Long)] = {
    if(region.level == 0)
      seedGraph.toStream.map { case (u, v) => (region.srcMin + u, region.dstMin + v) }
    else
      splitRegion(region, seedGraph).toStream.flatMap(r => generate(r, seedGraph))
  }

  /**
    * split a region
    *
    * @param region to split
    * @param seedGraph array of edges of the seed graph
    * @return list of splitted regions
    */
  def splitRegion(region: Region, seedGraph: Array[(Int, Int)]): Array[Region] = {


    val numSeedNodes = seedGraph.flatMap(x=> Seq(x._1, x._2)).max + 1
    val numSeedEdges = seedGraph.length

    val unit = Math.pow(numSeedNodes, region.level).toLong

    seedGraph.map{ case (u, v) =>
      Region(region.srcMin + u * unit, region.dstMin + v * unit, region.level - 1, region.numEdges / numSeedEdges)
    }
  }

  /**
    * a region.
    *
    * @param srcMin the minimum source node id in the region
    * @param dstMin the minimum destination node id in the region
    * @param level of recursion
    * @param numEdges the number of edges in the region
    */
  case class Region(srcMin: Long, dstMin: Long, level: Int, numEdges: Long)

  case class RegionSlot(var edgeCapacity: Long, regions: mutable.Buffer[Region]){

    /**
      * add a region to this region slot
      * @param r region to add
      * @return true if r is added or false
      */
    def add(r: Region){
        regions.append(r)
        edgeCapacity -= r.numEdges
    }
  }

  def regionOrder(r: Region): Long = r.numEdges

  def regionSlotOrder(s: RegionSlot): Long = s.edgeCapacity
}
