/**
  * PegasusN: Peta-Scale Graph Mining System (Pegasus v3.0)
  * Authors: Chiwan Park, Ha-Myung Park, U Kang
  *
  * This software is licensed under Apache License, Version 2.0 (the  "License");
  * you may not use this file except in compliance with the License.
  * You may obtain a copy of the License at
  *
  * http://www.apache.org/licenses/LICENSE-2.0
  *
  * Unless required by applicable law or agreed to in writing, software
  * distributed under the License is distributed on an "AS IS" BASIS,
  * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  * See the License for the specific language governing permissions and
  * limitations under the License.
  * -------------------------------------------------------------------------
  * File: RMATGenerator.java
  * - generate a rmat graph
  * Version: 3.0
  */

package pegasus.spark.graphgen

import org.apache.commons.math3.distribution.MultivariateNormalDistribution
import org.apache.log4j.Logger
import org.apache.spark.rdd.RDD
import org.apache.spark.{HashPartitioner, SparkConf, SparkContext}

import scala.collection.mutable

object RMATGenerator {

  val logger = Logger.getLogger(RMATGenerator.getClass)

  /**
    * The main entry point
    * @param args [0]: level of recursion. [1]: the number of edges. [2]: degree of parallelism (the
    *             number of tasks to run). [3], [4], [5], [6]: the probabilities of quarters. [7]: noise
    *             for smoothing
    */
  def main(args: Array[String]): Unit = {

    val level = args(0).toInt
    val numEdges = args(1).toLong
    val numTasks = args(2).toInt
    val abcd = (args(3).toDouble, args(4).toDouble, args(5).toDouble, args(6).toDouble)
    val noise = args(7).toDouble
    val output = args(8)

    logger.info(f"level: $level, numEdges: $numEdges, numTasks: $numTasks, " +
      f"(a,b,c,d): $abcd")

    val conf = new SparkConf().setAppName(f"[RMAT] lv: $level, E: $numEdges, abcd: $abcd") //.setMaster("local")

    val sc = new SparkContext(conf)

    run(level, numEdges, numTasks, abcd, noise, sc).map{case (u, v) => u + "\t" + v}.saveAsTextFile(output)

    sc.stop()
  }


  /**
    * submit the spark job.
    *
    * @param level of recursion
    * @param numEdges the number of edges
    * @param numTasks degree of parallelism (the number of tasks to run)
    * @param abcd the probabilities of quarters.
    * @param noise for smoothing
    * @param sc spark context
    * @return an RDD containing edges of the generated graph
    */
  def run(level: Int, numEdges: Long, numTasks: Int,
          abcd: (Double, Double, Double, Double), noise: Double, sc: SparkContext): RDD[(Long, Long)] = {

    val regionQueue = new mutable.PriorityQueue[Region]()(Ordering.by(regionOrder))
    val regionSlotQueue = new mutable.PriorityQueue[RegionSlot]()(Ordering.by(regionSlotOrder))

    for (i <- 0 until numTasks) {
      val numEdgeAssigned = (numEdges * (i + 1) / numTasks) - (numEdges * i / numTasks)
      regionSlotQueue.enqueue(RegionSlot(numEdgeAssigned, mutable.Buffer[Region]()))
    }

    val numNodes = 1L << level

    val noiseList = sc.broadcast[Array[Double]]((0L until level).map(_ => Math.random() * 2 * noise - noise).toArray)

    regionQueue.enqueue(Region(0, 0, numNodes - 1, numNodes - 1, numEdges, level-1))

    val rmat = new RMAT(abcd, noiseList.value)

    while (regionQueue.nonEmpty) {
      val region = regionQueue.dequeue
      val slot = regionSlotQueue.dequeue

      if(region.numEdges <= thresholdRegionEdgeNum || region.numEdges <= slot.edgeCapacity){
        slot.add(region)
      }
      else{
        rmat.splitRegion(region).foreach{r => regionQueue.enqueue(r)}
      }

      regionSlotQueue.enqueue(slot)
    }

    val regionSlotRDD = sc.parallelize(regionSlotQueue.toSeq.zipWithIndex.map{case (x,y) => (y,x)})
      .partitionBy(new HashPartitioner(numTasks))

    regionSlotRDD.flatMap{ case (id, x) =>
      val l_rmat = new RMAT(abcd, noiseList.value)
      x.regions.toStream.flatMap(l_rmat.generate)
    }

  }

  // if a region contains less than thresholdRegionEdgeNum, this region will not be splitted.
  val thresholdRegionEdgeNum = 10000

  /**
    * rmat generator
    * @param abcd the probabilities
    * @param noiseList a list of noises applied to each level
    */
  class RMAT (abcd: (Double, Double, Double, Double), noiseList: Array[Double]) {

    private val probs = noiseList.map{ mu =>
      (abcd._1 - 2 * mu * abcd._1 / (abcd._1 + abcd._4),
        abcd._2 + mu,
        abcd._3 + mu,
        abcd._4 - 2 * mu * abcd._4 / (abcd._1 + abcd._4))
    }

    /**
      * generate edges within a given region
      * @param region to process
      * @return list of edges in the region
      */
    def generate(region: Region): Stream[(Long, Long)] = {

      if(region.numEdges < thresholdRegionEdgeNum)
        (0L until region.numEdges).map{_ => makeOneEdge(region)}.distinct.toStream
      else
        splitRegion(region).toStream.flatMap(generate)

    }

    /**
      * make one edge in a region
      * @param region to process
      * @return an edge in the region
      */
    def makeOneEdge(region: Region): (Long, Long) = {

      var x = region.srcMin
      var y = region.dstMin

      var t = probs(region.level)

      for (l <- region.level to 0 by -1) {
        val p = Math.random
        if (p > t._1 + t._2 + t._3) {
          x += (1L << l)
          y += (1L << l)
        }
        else if (p > t._1 + t._2) x += (1L << l)
        else if (p > t._1) y += (1L << l)
      }

      (x, y)

    }

    /**
      * split a region
      * @param region to split
      * @return list of splitted regions
      */
    def splitRegion(region: Region): Array[Region] = {

      val t = probs(region.level)

      val splitEdgeNums = multinomial(region.numEdges, Array(t._1, t._2, t._3))
      val lastSplitEdgeNum = region.numEdges - splitEdgeNums.sum

      val srcMid = (region.srcMin + region.srcMax) >> 1
      val dstMid = (region.dstMin + region.dstMax) >> 1

      val maxSplitNumNodes = (region.srcMax - srcMid) * (region.srcMax - srcMid)

      val res = new mutable.ArrayBuffer[Region]()

      if(splitEdgeNums(0) > 0)
        res += Region(region.srcMin, region.dstMin, srcMid, dstMid,
          Math.min(splitEdgeNums(0), maxSplitNumNodes), region.level - 1)

      if(splitEdgeNums(1) > 0)
        res += Region(region.srcMin, dstMid + 1, srcMid, region.dstMax,
          Math.min(splitEdgeNums(1), maxSplitNumNodes), region.level - 1)

      if(splitEdgeNums(2) > 0)
        res += Region(srcMid + 1, region.dstMin, region.srcMax, dstMid,
          Math.min(splitEdgeNums(2), maxSplitNumNodes), region.level - 1)

      if(lastSplitEdgeNum > 0)
        res += Region(srcMid + 1, dstMid + 1, region.srcMax, region.dstMax,
          Math.min(lastSplitEdgeNum, maxSplitNumNodes), region.level - 1)

      res.toArray
    }

    /**
      * get a sample following a multinomial distribution
      *
      * @param n the number of trials
      * @param p list of probability
      * @return a sample following a multinomial distribution
      */
    def multinomial(n: Long, p: Array[Double]): Array[Long] ={

      val nM = Array.ofDim[Double](p.length, p.length)
      val np = p.map(_*n)

      for (i <- p.indices; j <- p.indices) {
        nM(i)(j) = if (i == j) p(i) else 0
        nM(i)(j) -= p(i) * p(j)
        nM(i)(j) *= n
      }

      val distribution = new MultivariateNormalDistribution(np, nM)
      var sample: Array[Long] = null

      do{
        sample = distribution.sample().map(Math.round)
      }while(sample.exists(_<0))


      sample
    }

  }


  case class Region(srcMin: Long, dstMin: Long, srcMax: Long, dstMax: Long, numEdges: Long, level: Int)

  case class RegionSlot(var edgeCapacity: Long, regions: mutable.Buffer[Region]){

    /**
      * add a region to this region slot
      * @param r region to add
      * @return true if r is added or false
      */
    def add(r: Region): Unit ={
        regions.append(r)
        edgeCapacity -= r.numEdges
    }
  }

  def regionOrder(r: Region): Long = r.numEdges

  def regionSlotOrder(s: RegionSlot): Long = s.edgeCapacity
}
