package pegasus.spark.graphgen

import org.apache.commons.math3.distribution.MultivariateNormalDistribution
import pegasus.spark.graphgen.RMATGenerator.Region

import scala.collection.mutable

/**
  * Created by hmpark on 17. 6. 16.
  */
object RMATSingle {
  var probs: Array[(Double, Double, Double, Double)] = _
  var cumuls: Array[(Double, Double, Double)] = _

  def generateWithLimitLevelMultinomial(level: Int, numEdges: Long, noise: Double, limitLevel: Int,
                                        abcd: (Double, Double, Double, Double)): Stream[(Long, Long)] = {

    val noiseList = (0L until level).map(_ => Math.random() * 2 * noise - noise).toArray

    probs = noiseList.map{noise =>
      (abcd._1 - 2 * noise * abcd._1 / (abcd._1 + abcd._4),
        abcd._2 + noise,
        abcd._3 + noise,
        abcd._4 - 2 * noise * abcd._4 / (abcd._1 + abcd._4))
    }

    cumuls = probs.map { case (a, b, c, d) => (a, a + b, a + b + c) }

    val numNodes = 1L << level
    val region = Region(0, 0, numNodes - 1, numNodes - 1, Math.min(numEdges, numNodes * numNodes), level-1)

    def genManual(region: Region): Stream[(Long, Long)] ={

      if(limitLevel <= region.level)
        splitRegionByMultinomial(region).toStream.flatMap(genManual)
      else
        (0L until region.numEdges).map{_ => makeOneEdge(region)}.distinct.toStream

    }

    genManual(region)

  }

  def generateWithLimitLevelTeG(level: Int, numEdges: Long, noise: Double, limitLevel: Int,
                                        abcd: (Double, Double, Double, Double)): Stream[(Long, Long)] = {

    val noiseList = (0L until level).map(_ => Math.random() * 2 * noise - noise).toArray

    probs = noiseList.map{noise =>
      (abcd._1 - 2 * noise * abcd._1 / (abcd._1 + abcd._4),
        abcd._2 + noise,
        abcd._3 + noise,
        abcd._4 - 2 * noise * abcd._4 / (abcd._1 + abcd._4))
    }

    val numNodes = 1L << level
    val region = Region(0, 0, numNodes - 1, numNodes - 1, Math.min(numEdges, numNodes * numNodes), level-1)

    def genManual(region: Region): Stream[(Long, Long)] ={

      if(limitLevel <= region.level)
        splitRegionFixed(region).toStream.flatMap(genManual)
      else
        (0L until region.numEdges).map{_ => makeOneEdge(region)}.distinct.toStream

    }

    genManual(region)

  }

  def generateWithMultinomial(level: Int, numEdges: Long, noise: Double, thresholdRegionEdgeNum: Long,
                    abcd: (Double, Double, Double, Double)): Stream[(Long, Long)] = {

    val noiseList = (0L until level).map(_ => Math.random() * 2 * noise - noise).toArray

    probs = noiseList.map{noise =>
      (abcd._1 - 2 * noise * abcd._1 / (abcd._1 + abcd._4),
        abcd._2 + noise,
        abcd._3 + noise,
        abcd._4 - 2 * noise * abcd._4 / (abcd._1 + abcd._4))
    }

    def genVer2(region:Region): Stream[(Long, Long)] = {
      if(region.numEdges < thresholdRegionEdgeNum)
        (0L until region.numEdges).map{_ => makeOneEdge(region)}.distinct.toStream
      else
        splitRegionByMultinomial(region).toStream.flatMap(genVer2)
    }

    val numNodes = 1L << level
    val region = Region(0, 0, numNodes - 1, numNodes - 1, Math.min(numEdges, numNodes * numNodes), level-1)
    genVer2(region)

  }

  def genMultSRand(level: Int, numEdges: Long, noise: Double, thresholdRegionEdgeNum: Long,
                              abcd: (Double, Double, Double, Double)): Stream[(Long, Long)] = {

    val noiseList = (0L until level).map(_ => Math.random() * 2 * noise - noise).toArray

    probs = noiseList.map{noise =>
      (abcd._1 - 2 * noise * abcd._1 / (abcd._1 + abcd._4),
        abcd._2 + noise,
        abcd._3 + noise,
        abcd._4 - 2 * noise * abcd._4 / (abcd._1 + abcd._4))
    }

    cumuls = probs.map { case (a, b, c, d) => (a, a + b, a + b + c) }

    def genVer2(region:Region): Stream[(Long, Long)] = {
      if(region.numEdges < thresholdRegionEdgeNum)
        (0L until region.numEdges).map{_ => makeOneEdgeQuick(region)}.distinct.toStream
      else
        splitRegionByMultinomial(region).toStream.flatMap(genVer2)
    }

    val numNodes = 1L << level
    val region = Region(0, 0, numNodes - 1, numNodes - 1, Math.min(numEdges, numNodes * numNodes), level-1)
    genVer2(region)

  }

  def genMultSRandPar(level: Int, numEdges: Long, noise: Double, thresholdRegionEdgeNum: Long,
                   abcd: (Double, Double, Double, Double)): Stream[(Long, Long)] = {

    val noiseList = (0L until level).map(_ => Math.random() * 2 * noise - noise).toArray

    probs = noiseList.map{noise =>
      (abcd._1 - 2 * noise * abcd._1 / (abcd._1 + abcd._4),
        abcd._2 + noise,
        abcd._3 + noise,
        abcd._4 - 2 * noise * abcd._4 / (abcd._1 + abcd._4))
    }

    cumuls = probs.map { case (a, b, c, d) => (a, a + b, a + b + c) }

    def genVer2(region:Region): Stream[(Long, Long)] = {
      if(region.numEdges < thresholdRegionEdgeNum) {
        (0L until region.numEdges).map { _ => makeOneEdgeQuick(region) }.distinct.toStream
      }
      else
        splitRegionByMultinomial(region).par.flatMap(genVer2).toStream
    }

    val numNodes = 1L << level
    val region = Region(0, 0, numNodes - 1, numNodes - 1, Math.min(numEdges, numNodes * numNodes), level-1)
    genVer2(region)

  }


  def generateByTeg(level: Int, numEdges: Long, noise: Double, thresholdRegionEdgeNum: Long,
                              abcd: (Double, Double, Double, Double)): Stream[(Long, Long)] = {

    val noiseList = (0L until level).map(_ => Math.random() * 2 * noise - noise).toArray

    probs = noiseList.map{noise =>
      (abcd._1 - 2 * noise * abcd._1 / (abcd._1 + abcd._4),
        abcd._2 + noise,
        abcd._3 + noise,
        abcd._4 - 2 * noise * abcd._4 / (abcd._1 + abcd._4))
    }

    def genVer2(region:Region): Stream[(Long, Long)] = {
      if(region.numEdges < thresholdRegionEdgeNum)
        (0L until region.numEdges).map{_ => makeOneEdge(region)}.distinct.toStream
      else
        splitRegionFixed(region).toStream.flatMap(genVer2)
    }

    val numNodes = 1L << level
    val region = Region(0, 0, numNodes - 1, numNodes - 1, Math.min(numEdges, numNodes * numNodes), level-1)
    genVer2(region)

  }

  def generateAsOriginal(level: Int, numEdges: Long, noise: Double,
               abcd: (Double, Double, Double, Double)): Stream[(Long, Long)] = {

    generateWithLimitLevelMultinomial(level, numEdges, noise, Int.MaxValue, abcd)

  }

  def makeOneEdge(region: Region): (Long, Long) = {

    var x = region.srcMin
    var y = region.dstMin

    var t = probs(region.level)

    for (l <- region.level to 0 by -1) {
      val p = Math.random
      if (p > t._1 + t._2 + t._3) {
        x += (1L << l)
        y += (1L << l)
      }
      else if (p > t._1 + t._2) x += (1L << l)
      else if (p > t._1) y += (1L << l)
    }

    (x, y)

  }

  def makeOneEdgeQuick(region: Region): (Long, Long) = {

    var x = region.srcMin
    var y = region.dstMin

    var p = Math.random

    for (l <- region.level to 0 by -1) {
      val c = cumuls(l)
      val t = probs(region.level)

      if (p > c._3) {
        x += (1L << l)
        y += (1L << l)
        p = (p - c._3) / t._4
      }
      else if (p > c._2){
        x += (1L << l)
        p = (p - c._2) / t._3
      }
      else if (p > c._1){
        y += (1L << l)
        p = (p - c._1) / t._2
      }
      else{
        p = p / t._1
      }
    }

    (x, y)

  }

  case class Region(srcMin: Long, dstMin: Long, srcMax: Long, dstMax: Long, numEdges: Long, level: Int)

  def splitRegionFixed(region: Region): Array[Region] = {
    val t = probs(region.level)

    val splitEdgeNums = Array(t._1, t._2, t._3).map{x => (x * region.numEdges).toLong}
    val lastSplitEdgeNum = region.numEdges - splitEdgeNums.sum

    val srcMid = (region.srcMin + region.srcMax) >> 1
    val dstMid = (region.dstMin + region.dstMax) >> 1

    val maxSplitNumNodes = (region.srcMax - srcMid) * (region.srcMax - srcMid)

    val res = new mutable.ArrayBuffer[Region]()

    if(splitEdgeNums(0) > 0)
      res += Region(region.srcMin, region.dstMin, srcMid, dstMid,
        Math.min(splitEdgeNums(0), maxSplitNumNodes), region.level - 1)

    if(splitEdgeNums(1) > 0)
      res += Region(region.srcMin, dstMid + 1, srcMid, region.dstMax,
        Math.min(splitEdgeNums(1), maxSplitNumNodes), region.level - 1)

    if(splitEdgeNums(2) > 0)
      res += Region(srcMid + 1, region.dstMin, region.srcMax, dstMid,
        Math.min(splitEdgeNums(2), maxSplitNumNodes), region.level - 1)

    if(lastSplitEdgeNum > 0)
      res += Region(srcMid + 1, dstMid + 1, region.srcMax, region.dstMax,
        Math.min(lastSplitEdgeNum, maxSplitNumNodes), region.level - 1)

    res.toArray
  }

  def splitRegionByMultinomial(region: Region): Array[Region] = {
    val t = probs(region.level)

    val splitEdgeNums = multinomial(region.numEdges, Array(t._1, t._2, t._3))
    val lastSplitEdgeNum = region.numEdges - splitEdgeNums.sum

    val srcMid = (region.srcMin + region.srcMax) >> 1
    val dstMid = (region.dstMin + region.dstMax) >> 1

    val maxSplitNumNodes = (region.srcMax - srcMid) * (region.srcMax - srcMid)

    val res = new mutable.ArrayBuffer[Region]()

    if(splitEdgeNums(0) > 0)
      res += Region(region.srcMin, region.dstMin, srcMid, dstMid,
        Math.min(splitEdgeNums(0), maxSplitNumNodes), region.level - 1)

    if(splitEdgeNums(1) > 0)
      res += Region(region.srcMin, dstMid + 1, srcMid, region.dstMax,
        Math.min(splitEdgeNums(1), maxSplitNumNodes), region.level - 1)

    if(splitEdgeNums(2) > 0)
      res += Region(srcMid + 1, region.dstMin, region.srcMax, dstMid,
        Math.min(splitEdgeNums(2), maxSplitNumNodes), region.level - 1)

    if(lastSplitEdgeNum > 0)
      res += Region(srcMid + 1, dstMid + 1, region.srcMax, region.dstMax,
        Math.min(lastSplitEdgeNum, maxSplitNumNodes), region.level - 1)

    res.toArray
  }

  def multinomial(n: Long, p: Array[Double]): Array[Long] ={

    val nM = Array.ofDim[Double](p.length, p.length)
    val np = p.map(_*n)

    for (i <- p.indices; j <- p.indices) {
      nM(i)(j) = if (i == j) p(i) else 0
      nM(i)(j) -= p(i) * p(j)
      nM(i)(j) *= n
    }

    val distribution = new MultivariateNormalDistribution(np, nM)
    var sample: Array[Long] = null

    do{
      sample = distribution.sample().map(Math.round)
    }while(sample.exists(_<0))


    sample
  }

}
