/**
  * PegasusN: Peta-Scale Graph Mining System (Pegasus v3.0)
  * Authors: Chiwan Park, Ha-Myung Park, U Kang
  *
  * This software is licensed under Apache License, Version 2.0 (the  "License");
  * you may not use this file except in compliance with the License.
  * You may obtain a copy of the License at
  *
  * http://www.apache.org/licenses/LICENSE-2.0
  *
  * Unless required by applicable law or agreed to in writing, software
  * distributed under the License is distributed on an "AS IS" BASIS,
  * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  * See the License for the specific language governing permissions and
  * limitations under the License.
  * -------------------------------------------------------------------------
  * File: RandomGenerator.scala
  * - generate a random graph.
  * Version: 3.0
  */

package pegasus.spark.graphgen

import org.apache.commons.math3.distribution.MultivariateNormalDistribution
import org.apache.log4j.Logger
import org.apache.spark.rdd.RDD
import org.apache.spark.{HashPartitioner, SparkConf, SparkContext}

import scala.collection.mutable

object RandomGenerator {

  private val logger = Logger.getLogger(RandomGenerator.getClass)

  def main(args: Array[String]): Unit = {

    val numNodes = args(0).toLong
    val numEdges = args(1).toLong
    val numTasks = args(2).toInt
    val output = args(3)

    logger.info(f"numNodes: $numNodes, numEdges: $numEdges, numTasks: $numTasks")

    val conf = new SparkConf().setAppName(f"[Random] V: $numNodes, E: $numEdges") //.setMaster("local")

    val sc = new SparkContext(conf)

    run(numNodes, numEdges, numTasks, sc).map{case(u, v) => u + "\t" + v}saveAsTextFile(output)

    sc.stop()
  }

  /**
    * submit the spark job.
    *
    * @param numNodes the number of nodes
    * @param numEdges the number of edges
    * @param numTasks degree of parallelism (the number of tasks to run)
    * @param sc spark context
    * @return an RDD containing edges of the generated graph
    */
  def run(numNodes: Long, numEdges: Long, numTasks: Int, sc: SparkContext): RDD[(Long, Long)] = {

    val regionQueue = new mutable.PriorityQueue[Region]()(Ordering.by(regionOrder))
    val regionSlotQueue = new mutable.PriorityQueue[RegionSlot]()(Ordering.by(regionSlotOrder))

    for (i <- 0 until numTasks) {
      val numEdgeAssigned = (numEdges * (i + 1) / numTasks) - (numEdges * i / numTasks)
      regionSlotQueue.enqueue(RegionSlot(numEdgeAssigned, mutable.Buffer[Region]()))
    }

    regionQueue.enqueue(Region(0, 0, numNodes - 1, numNodes - 1, Math.min(numEdges, numNodes * numNodes)))

    while (regionQueue.nonEmpty) {
      val region = regionQueue.dequeue
      val slot = regionSlotQueue.dequeue

      if(region.numEdges <= thresholdRegionEdgeNum || region.numEdges <= slot.edgeCapacity){
        slot.add(region)
      }
      else{
        splitRegion(region).foreach(r => regionQueue.enqueue(r))
      }

      regionSlotQueue.enqueue(slot)
    }

    val regionSlotRDD = sc.parallelize(regionSlotQueue.toSeq.zipWithIndex.map{case (x,y) => (y,x)})
      .partitionBy(new HashPartitioner(numTasks))

    regionSlotRDD.flatMap{ case (id, x) =>
      x.regions.toStream.flatMap(generate)
    }

  }

  // if a region contains less than thresholdRegionEdgeNum, this region will not be splitted.
  val thresholdRegionEdgeNum = 10000

  /**
    * generate edges within a given region
    * @param region to process
    * @return list of edges in the region
    */
  def generate(region: Region): Stream[(Long, Long)] = {

    if(region.numEdges < thresholdRegionEdgeNum) {
      (0L until region.numEdges).map{_ =>
        ((Math.random * (region.srcMax - region.srcMin)).toLong + region.srcMin,
          (Math.random * (region.dstMax - region.dstMin)).toLong + region.dstMin)
      }.distinct.toStream
    }
    else splitRegion(region).toStream.flatMap(generate)

  }

  /**
    * split a region
    *
    * @param region to split
    * @return list of splitted regions
    */
  def splitRegion(region: Region): Array[Region] = {

    val splitEdgeNums = multinomial(region.numEdges, Array(0.25, 0.25, 0.25))
    val lastSplitEdgeNum = region.numEdges - splitEdgeNums.sum

    val srcMid = (region.srcMin + region.srcMax) >> 1
    val dstMid = (region.dstMin + region.dstMax) >> 1

    val maxSplitNumNodes = (region.srcMax - srcMid) * (region.srcMax - srcMid)

    Array(
      Region(region.srcMin, region.dstMin, srcMid, dstMid,
        Math.min(splitEdgeNums(0), maxSplitNumNodes)),
      Region(region.srcMin, dstMid + 1, srcMid, region.dstMax,
        Math.min(splitEdgeNums(1), maxSplitNumNodes)),
      Region(srcMid + 1, region.dstMin, region.srcMax, dstMid,
        Math.min(splitEdgeNums(2), maxSplitNumNodes)),
      Region(srcMid + 1, dstMid + 1, region.srcMax, region.dstMax,
        Math.min(lastSplitEdgeNum, maxSplitNumNodes))
    )
  }

  /**
    * get a sample following a multinomial distribution
    *
    * @param n the number of trials
    * @param p list of probability
    * @return a sample following a multinomial distribution
    */
  def multinomial(n: Long, p: Array[Double]): Array[Long] ={

    val nM = Array.ofDim[Double](p.length, p.length)
    val np = p.map(_*n)

    for (i <- p.indices; j <- p.indices) {
      nM(i)(j) = if (i == j) p(i) else 0
      nM(i)(j) -= p(i) * p(j)
      nM(i)(j) *= n
    }

    val distribution = new MultivariateNormalDistribution(np, nM)
    var sample: Array[Long] = null

    do{
      sample = distribution.sample().map(Math.round)
    }while(sample.exists(_<0))


    sample
  }

  /**
    * a region.
    *
    * @param srcMin the minimum source node id in the region
    * @param dstMin the minimum destination node id in the region
    * @param srcMax the source node ids in the region are less than srcMax
    * @param dstMax the destination node ids in the region are less than dstMax
    * @param numEdges the number of edges in the region
    */
  case class Region(srcMin: Long, dstMin: Long, srcMax: Long, dstMax: Long, numEdges: Long)

  case class RegionSlot(var edgeCapacity: Long, regions: mutable.Buffer[Region]){

    /**
      * add a region to this region slot
      * @param r region to add
      * @return true if r is added or false
      */
    def add(r: Region){
        regions.append(r)
        edgeCapacity -= r.numEdges
    }
  }

  def regionOrder(r: Region): Long = r.numEdges

  def regionSlotOrder(s: RegionSlot): Long = s.edgeCapacity
}
