/*
 * PegasusG: Peta-Scale Graph Mining System v3.0
 * Authors: Chiwan Park, Ha-Myung Park, U Kang
 *
 * This software is licensed under Apache License, Version 2.0 (the  "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 * -------------------------------------------------------------------------
 * File: NetRayScatter.java
 * - Draw a scatter plot of the given data
 * Version: 3.0
 */
package pegasus.spark.netray

import org.apache.spark.{SparkConf, SparkContext}

object NetRayScatter {

  case class Options(inputPath: String, outputPath: String, targetWidth: Int, fieldIds: Seq[Int], logScale: Boolean)

  private def parseOptions(args: Array[String]): Options = {
    val parser = new scopt.OptionParser[Options]("netray-scatter") {
      opt[String]("inputPath").abbr("i")
        .action((x, c) => c.copy(inputPath = x))
        .required()
      opt[String]("outputPath").abbr("o")
        .action((x, c) => c.copy(outputPath = x))
        .required()
      opt[Int]("targetWidth").abbr("w")
        .action((x, c) => c.copy(targetWidth = x))
        .required()
      opt[Seq[Int]]("fieldIds").abbr("f")
        .action((x, c) => c.copy(fieldIds = x))
        .validate(x => if (x.length == 2) success else failure("The number of specified fields must be 2."))
      opt[Unit]("logScale").abbr("l")
        .action((_, c) => c.copy(logScale = true))
    }

    parser.parse(args, Options("", "", -1, Seq(0, 1), logScale = false)) match {
      case Some(options) => options
      case None => sys.exit(-1)
    }
  }

  def main(args: Array[String]): Unit = {
    val options = parseOptions(args)
    val conf = new SparkConf()
      .setAppName(s"NetRayScatter - ${options.inputPath}")
    val sc = new SparkContext(conf)

    // accumulators for min/max values
    val (xMin, xMax, yMin, yMax) = (new MinAccumulator, new MaxAccumulator, new MinAccumulator, new MaxAccumulator)
    sc.register(xMin, "xMin")
    sc.register(xMax, "xMax")
    sc.register(yMin, "yMin")
    sc.register(yMax, "yMax")

    // load matrix
    val Seq(xField, yField) = options.fieldIds
    val initMat = sc.textFile(options.inputPath).mapPartitions(iter => {
      val delimiter = "\\s+".r
      iter.filter(!_.startsWith("#")).map(line => {
        val split = delimiter.split(line)
        val x = split(xField).toLong
        val y = split(yField).toLong
        xMin.add(x)
        xMax.add(x)
        yMin.add(y)
        yMax.add(y)
        (x, y)
      })
    }, preservesPartitioning = true).cache()

    // trigger empty action to compute min/max values
    initMat.foreachPartition(_ => ())

    // re-scale matrix
    val scaledMat = if (options.logScale) {
      // log scale
      val (diffX, diffY) = (math.log(xMin.value), math.log(yMin.value))
      val convertXRatio = (options.targetWidth - 1) / (math.log(xMax.value) - math.log(xMin.value))
      val convertYRatio = (options.targetWidth - 1) / (math.log(yMax.value) - math.log(yMin.value))
      initMat.mapPartitions(iter => {
        iter.map(mat => {
          val (x, y) = mat
          val xBlk = (math.log(x) - diffX) * convertXRatio + 0.5
          val yBlk = (math.log(y) - diffY) * convertYRatio + 0.5
          ((xBlk, yBlk), 1L)
        })
      }, preservesPartitioning = true)
    } else {
      // linear scale
      val (diffX, diffY) = (xMin.value, yMin.value)
      val convertXRatio = (options.targetWidth - 1) / (xMax.value - xMin.value)
      val convertYRatio = (options.targetWidth - 1) / (yMax.value - yMin.value)
      initMat.mapPartitions(iter => {
        iter.map(mat => {
          val (x, y) = mat
          val xBlk = (x - diffX) * convertXRatio + 0.5
          val yBlk = (y - diffY) * convertYRatio + 0.5
          ((xBlk, yBlk), 1L)
        })
      }, preservesPartitioning = true)
    }

    // aggregate
    val aggFunc = (v1: Long, v2: Long) => v1 + v2
    val aggregatedMat = scaledMat.aggregateByKey(0L)(aggFunc, aggFunc)

    // output
    val outputRDD = aggregatedMat.map(mat => s"${mat._1._1}\t${mat._1._2}\t${mat._2}")
    outputRDD.saveAsTextFile(options.outputPath)

    sc.stop()
  }
}
