/*
 * PegasusG: Peta-Scale Graph Mining System v3.0
 * Authors: Chiwan Park, Ha-Myung Park, U Kang
 *
 * This software is licensed under Apache License, Version 2.0 (the  "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 * -------------------------------------------------------------------------
 * File: NetRaySpy.java
 * - Draw a spy plot of the given matrix
 * Version: 3.0
 */
package pegasus.spark.netray

import org.apache.spark.{SparkConf, SparkContext}

object NetRaySpy {

  case class Options(inputPath: String, outputPath: String, numNodes: Long, targetWidth: Int, logScale: Boolean)

  private def parseOptions(args: Array[String]) = {
    val parser = new scopt.OptionParser[Options]("netray-spy") {
      opt[String]("inputPath").abbr("i")
        .action((x, c) => c.copy(inputPath = x))
        .required()
      opt[String]("outputPath").abbr("o")
        .action((x, c) => c.copy(outputPath = x))
        .required()
      opt[Long]("numNodes").abbr("n")
        .action((x, c) => c.copy(numNodes = x))
        .required()
      opt[Int]("targetWidth").abbr("w")
        .action((x, c) => c.copy(targetWidth = x))
        .required()
      opt[Unit]("logScale").abbr("l")
        .action((_, c) => c.copy(logScale = true))
    }

    parser.parse(args, Options("", "", -1L, -1, logScale = false)) match {
      case Some(options) => options
      case None => sys.exit(-1)
    }
  }

  def main(args: Array[String]): Unit = {
    val options = parseOptions(args)
    val conf = new SparkConf()
      .setAppName(s"NetRaySpy - ${options.inputPath}")
    val sc = new SparkContext(conf)

    // load adjacency matrix
    val adjMat = sc.textFile(options.inputPath).mapPartitions(iter => {
      val delimiter = "\\s+".r
      iter.filter(!_.startsWith("#")).map(line => {
        val split = delimiter.split(line)
        if (split.length == 2) {
          (split(0).toLong, split(1).toLong, 1L)
        } else {
          (split(0).toLong, split(1).toLong, split(2).toLong)
        }
      })
    }, preservesPartitioning = true)

    // re-scale matrix
    val scaledMat = if (options.logScale) {
      // log scale
      val convertRatio = (options.targetWidth - 1) / math.log10(options.numNodes)
      adjMat.mapPartitions(iter => {
        iter.map(mat => {
          val (x, y, value) = mat
          val xBlk = (math.log10(x + 1) * convertRatio + 0.5).toInt
          val yBlk = (math.log10(y + 1) * convertRatio + 0.5).toInt
          ((xBlk, yBlk), value)
        })
      }, preservesPartitioning = true)
    } else {
      // linear scale
      val convertRatio = options.targetWidth / options.numNodes.toDouble
      adjMat.mapPartitions(iter => {
        iter.map(mat => {
          val (x, y, value) = mat
          val xBlk = ((x + 1) * convertRatio).toInt
          val yBlk = ((y + 1) * convertRatio).toInt
          ((xBlk, yBlk), value)
        })
      })
    }

    // aggregate
    val aggFunc = (v1: Long, v2: Long) => v1 + v2
    val aggregatedMat = scaledMat.aggregateByKey(0L)(aggFunc, aggFunc)

    // output
    val outputRDD = aggregatedMat.map(mat => s"${mat._1._1}\t${mat._1._2}\t${mat._2}")
    outputRDD.saveAsTextFile(options.outputPath)

    sc.stop()
  }
}
