package pegasus.spark.sssp

import org.apache.spark.{SparkConf, SparkContext}
import pegasus.spark.common._
import pegasus.spark.common.io.{DoubleVectorReader, DoubleVectorWriter, VectorReader, VectorWriter}
import pegasus.spark.common.serialization.PMVRegistrator
import pegasus.spark.common.types.{DoubleVectorBlock, MatrixBlock, VectorBlock}

object SSSPComputation extends Computation {

  case class Config(
    input: String = "",
    output: String = "",
    tempPath: Option[String] = None,
    sourceVertex: Long = -1L,
    numBlocks: Int = 40,
    denseThreshold: Int = 150,
    fieldDelimiter: String = "\t",
    maxIterations: Int = -1)

  private val parser = new scopt.OptionParser[Config]("SSSP") {
    head("SSSP")

    opt[Int]("numBlocks").abbr("b")
      .action((x, c) => c.copy(numBlocks = x))

    opt[Int]("denseThreshold").abbr("t")
      .action((x, c) => c.copy(denseThreshold = x))

    opt[Int]("maxIterations").abbr("i")
      .action((x, c) => c.copy(maxIterations = x))

    opt[String]("fieldDelimiter").abbr("d")
      .action((x, c) => c.copy(fieldDelimiter = x))

    opt[String]("temporaryPath").abbr("tp")
      .action((x, c) => c.copy(tempPath = Some(x)))

    arg[String]("input")
      .action((x, c) => c.copy(input = x))

    arg[String]("output")
      .action((x, c) => c.copy(output = x))

    arg[Long]("sourceVertex")
      .action((x, c) => c.copy(sourceVertex = x))
  }

  private def multiplication(
    sourceVertex: Long,
    threshold: Double
  ) = new Multiplication[(Int, Double), Double, Double] {
    override def createVectorBlock(
      blockId: Int,
      numVertices: Long,
      numBlocks: Int,
      init: Boolean,
      context: MultiplicationContext
    ): VectorBlock[Double] = {
      val block = new DoubleVectorBlock(numBlocks, numVertices, blockId, Constants.Infinite)
      if (init && sourceVertex % numBlocks == blockId) {
        block.put(PMVUtils.globalToLocal(numBlocks, sourceVertex), 0.0)
      }
      block
    }

    override def createVectorReader(vectorPath: String, blockId: Int): VectorReader[Double] = {
      new DoubleVectorReader(f"$vectorPath/${Constants.DenseVectorBlock}-$blockId%05d")
    }

    override def createVectorWriter(vectorPath: String, blockId: Int): VectorWriter[Double] = {
      new DoubleVectorWriter(f"$vectorPath/${Constants.DenseVectorBlock}-$blockId%05d")
    }

    override def createIntermediateVectorBlock(
      numVertices: Long,
      numBlocks: Int,
      blockId: Int,
      context: MultiplicationContext
    ): VectorBlock[Double] = {
      new DoubleVectorBlock(numBlocks, numVertices, blockId)
    }

    override def createIntermediateVectorReader(vectorPath: String, blockId: Int): VectorReader[Double] = {
      new DoubleVectorReader(f"$vectorPath/${Constants.IntermediateVectorBlock}-$blockId%05d")
    }

    override def createIntermediateVectorWriter(vectorPath: String, blockId: Int): VectorWriter[Double] = {
      new DoubleVectorWriter(f"$vectorPath/${Constants.IntermediateVectorBlock}-$blockId%05d")
    }

    override def combine2Op(
      matrix: MatrixBlock[(Int, Double)],
      vector: VectorBlock[Double],
      result: VectorBlock[Double],
      context: MultiplicationContext
    ): Unit = {
      val iter = matrix.iterator()
      while (iter.next()) {
        val col = iter.currentCol()
        val rows = iter.currentRows()
        val srcDist = if (vector.contains(col)) vector.get(col) else Constants.Infinite

        if (srcDist < Constants.Infinite) {
          var i = 0
          val size = rows.size()
          while (i < size) {
            val (row, value) = rows.row(i)
            if (!result.contains(row) || result.get(row) > srcDist + value) {
              result.put(row, srcDist + value)
            }
            i += 1
          }
        }
      }
    }

    override def combineAllOp(
      input: VectorReader[Double],
      output: VectorBlock[Double],
      context: MultiplicationContext
    ): Unit = {
      while (input.hasNext) {
        val (id, dist) = input.next()
        if (!output.contains(id) || output.get(id) > dist) {
          output.put(id, dist)
        }
      }
    }

    override def applyOp(
      result: VectorBlock[Double],
      intermediate: VectorBlock[Double],
      context: MultiplicationContext
    ): Unit = {
      val iter = result.iterator(false)
      while (iter.next()) {
        val id = iter.currentId
        val dist = iter.currentValue
        if (intermediate.contains(id) && intermediate.get(id) < dist) {
          context.notifyChanged()
          result.put(id, intermediate.get(id))
        }
      }
    }
  }

  def run(sc: SparkContext, config: Config): Unit = {
    val partitionedPath = getPartitionedPath(config.output, config.tempPath)
    val intermediatePath = getIntermediatePath(config.output, config.tempPath)
    val edges = GraphLoader.weightedEdgeListFile(sc, config.input, config.fieldDelimiter)

    Prepartitioning.weightedMatrix(edges, config.numBlocks, config.denseThreshold, partitionedPath)

    val (numVertices, problemRDD) = MatrixLoader.loadSubproblems(sc, partitionedPath)
    val (sparseRDD, denseRDD) = MatrixLoader.loadWeightedMatrix(problemRDD)

    val result = MultiplicationRunner.run(
      multiplication(config.sourceVertex, 1e-9),
      problemRDD,
      sparseRDD,
      denseRDD,
      numVertices,
      config.maxIterations,
      intermediatePath
    )

    result.map(v => s"${v._1}\t${v._2}").saveAsTextFile(config.output)
  }

  def main(args: Array[String]): Unit = {
    parser.parse(args, Config()) match {
      case Some(config) =>
        val conf = new SparkConf()
          .setAppName("PMV - PageRank")
          .set("spark.serializer", "org.apache.spark.serializer.KryoSerializer")
          .set("spark.kryo.registrator", classOf[PMVRegistrator].getName)
          .set("spark.kryoserializer.buffer", "24m")
        val sc = new SparkContext(conf)

        run(sc, config)

        sc.stop()
      case _ => System.exit(-1)
    }
  }
}
