package pegasus.spark.subgraph

import java.util
import java.util.Arrays

class VF2(adj: Array[Array[Int]], patternString: String, filters: Array[Long], colors: Array[Int],
          numColors: Int){

  /* pattern graph variables */
  //pattern[i][j] = pattern[j][i] = true if i and j are linked.
  private[subgraph] var pattern: Array[Array[Boolean]] = {
    var patternSize = 1
    while(patternSize*(patternSize-1)/2 < patternString.length()) patternSize += 1

    val pattern = Array.ofDim[Boolean](patternSize, patternSize)

    var k = 0
    for(j <- 0 until patternSize){
      for(i <- j + 1 until patternSize){
        pattern(j)(i) = patternString.charAt(k) == '1'
        pattern(i)(j) = pattern(j)(i)
        k += 1
      }
    }

    pattern
  }

  // partialOrder[i][j] = true if i < j.
  private[subgraph] var partialOrder: Array[Array[Boolean]] = {

    val partialOrder = Array.ofDim[Boolean](pattern.length, pattern.length)
    val path = Array.ofDim[Int](pattern.length)

    def dfs(p: Int): Unit ={
      if(p == path.length){
        for(i <- pattern.indices; j <- pattern.indices){
          if(pattern(i)(j) != pattern(path(i))(path(j))) return
        }

        //from here, the path is isomorphic to the original node sequence.
        for(i <- pattern.indices){
          if(path(i) != i){
            partialOrder(i)(path(i)) = true
            return
          }
        }
      }
      else{
        for(i <- pattern.indices){
          var j = 0
          while(j < p && path(j) != i) j += 1
          if(j == p){
            path(p) = i
            dfs(p + 1)
          }
        }
      }
    }

    dfs(0)
    partialOrder

  }

  //for checking feasibility rule
  //the k-th bit of fr(n)(0) is set if a pattern node n has a large neighbor k
  private[subgraph] var fr: Array[Array[Int]] = {

    val fr = Array.ofDim[Int](pattern.length, 3)

    for(i <- pattern.indices){
      for(j <- 0 until i){
        if(pattern(i)(j)){
          fr(i)(0) |= 1 << j
        }
      }

      for(j <- i + 1 until pattern.length){
        if(pattern(i)(j)){
          var k = 0
          while(k < i && !pattern(k)(j)) k += 1
          if(k < i) fr(i)(1) += 1
          else fr(i)(2) += 1
        }
      }
    }

    fr
  }

  /* match state variables */
  private[subgraph] var state: Array[Int] = Array.ofDim[Int](adj.length)
  private[subgraph] var matched: Array[Int] = Array.ofDim[Int](pattern.length)
  private[subgraph] var cand_size: Array[Int] = Array.ofDim[Int](pattern.length)
  private[subgraph] var candidate: Array[Int] = Array.ofDim[Int](adj.length)
  private[subgraph] var dependency: Array[Array[Long]] = Array.ofDim[Long](pattern.length, pattern.length)
  // dependency[i][j] = the minimum possible value of j in i-th level = the maximum node having
  // a dependency relation with j in i-th level.
  // high 32 bits = degree, low 32 bits = node id.

  private[subgraph] var count: Long = 0L
  private[subgraph] var colorbit: Long = 0L
  private[subgraph] var color_matched: Array[Int] = Array.ofDim[Int](pattern.length)


  def subgraphs(): Iterator[Array[Long]] ={

    new Iterator[Array[Long]]{

      var u = 0
      var it_now: Iterator[Array[Long]] = Iterator.empty
      var isRestored: Boolean = true

      override def hasNext: Boolean = it_now.hasNext || {

        if(!isRestored){
          restore(0, u, 0, 0)
          isRestored = true
          u += 1
        }

        if(u >= adj.length) false
        else{
          if(adj(u) == null) {
            it_now = Iterator.empty
          }
          else {
            it_now = if(prepare(0, u)){
              vf2(1)
            }
            else Iterator.empty
            isRestored = false
          }
          hasNext
        }
      }

      override def next(): Array[Long] = if(hasNext) it_now.next() else Iterator.empty.next()
    }

  }

  def prepare(p: Int, u: Int): Boolean ={

    matched(p) = u
    state(u) = ~p
    cand_size(p) = if(p == 0) 0 else cand_size(p-1)
    val uColor = u % colors.length
    colorbit |= 1 << uColor
    color_matched(p) = colors(uColor)

    var m = Long.MaxValue
    for(j <- p + 1 until pattern.length){
      dependency(p)(j) = if(p == 0) -1 else dependency(p-1)(j)

      val order_u = (adj(u).length.toLong << 32) | u

      if (partialOrder(p)(j) && dependency(p)(j) < order_u)
        dependency(p)(j) = order_u
      if (m > dependency(p)(j)) m = dependency(p)(j)
    }

    var existCand_cnt = 0
    var newCand_cnt = 0

    (adj(u).length-1 to 0 by -1).takeWhile{j=>
      val v = adj(u)(j)
      ((adj(v).length.toLong << 32) | v) > m
    }.foreach{ j =>
      val v = adj(u)(j)
      if(state(v) > 0){ // v is candidate but not matched
        existCand_cnt += 1
        state(v) |= 1 << p
      }
      else if(state(v) == 0){ // v is neither a candidate nor a match
        newCand_cnt += 1
        state(v) |= 1 << p
        candidate(cand_size(p)) = v
        cand_size(p) += 1
      }
    }

    val feasibility: Boolean = existCand_cnt + newCand_cnt >= fr(p)(1) + fr(p)(2)

    feasibility
  }

  def restore(p: Int, u: Int, t: Int, t_colorbit: Long): Unit = {

    colorbit = t_colorbit
    state(u) = t

    var m = Long.MaxValue
    for(j <- p + 1 until pattern.length){
      if(m > dependency(p)(j)) m = dependency(p)(j)
    }

    (adj(u).length-1 to 0 by -1).takeWhile{ j =>
      val v = adj(u)(j)
      ((adj(v).length.toLong << 32) | v) > m
    }.foreach{ j =>
      val v = adj(u)(j)
      if(state(v) > 0){
        state(v) &= ~(1 << p)
      }
    }

  }

  def vf2(p: Int): Iterator[Array[Long]] = {

    new Iterator[Array[Long]] {

      var lv = 0
      var start = 0
      var u = 0
      var it_now: Iterator[Array[Long]] = Iterator.empty
      var isRestored: Boolean = true
      var t_state: Int = _
      var t_colorbit: Long = _

      override def hasNext: Boolean = it_now.hasNext || {

        if(!isRestored){
          restore(p, u, t_state, t_colorbit)
          isRestored = true
        }

        (lv until p).find{ lv =>

          if(start >= cand_size(lv)) false
          else {
            u = candidate(start)
            val order_u = (adj(u).length.toLong << 32) | u

            if (order_u > dependency(p - 1)(p)) true
            else {
              start = cand_size(lv)
              false
            }
          }
        } match {
          case Some(x) =>
            lv = x
            val u = candidate(start)

            // state[u] > 0 means that u is not matched candidate.
            // fr[p][0] == status[u] --> following feasibility rule 1
            if(state(u) > 0 && (fr(p)(0) & state(u)) == fr(p)(0)){
              if(p + 1 >= pattern.length){

                val uColor = u % colors.length

                if (filters.contains(colorbit | (1 << uColor))) {
                  matched(p) = u
                  color_matched(p) = colors(uColor)

                  val subgraph = matched zip color_matched map { case (n, c) =>
                    (n/colors.length).toLong * numColors + c
                  }

                  it_now = Iterator.single[Array[Long]](subgraph)
                }
              }
              else{
                t_state = state(u)
                t_colorbit = colorbit

                it_now = if (prepare(p, u)) vf2(p + 1)
                         else Iterator.empty

                isRestored = false
              }
            }

            start += 1
            hasNext
          case None => false
        }
      }

      override def next(): Array[Long] = if(hasNext) it_now.next() else Iterator.empty.next()
    }

  }
}
